/*
 * linux/fs/hfs/balloc.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * hfs_bnode_alloc() and hfs_bnode_bitop() are based on GPLed code
 * Copyright (C) 1995  Michael Dreher
 *
 * This file contains the code to create and destroy nodes
 * in the B-tree structure.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 *
 * The code in this file initilizes some structures by calling
 * memset(&foo, 0, sizeof(foo)).  This produces the desired behavior
 * only due to the non-ANSI assumption that the machine representation
 * of NULL is all zeros.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/fs.h>
#include <asm/bitops.h>
#include "hfs_fs.h"
#include "hfs_btree.h"

/*================ File-local functions ================*/

/*
 * hfs_bnode_bitop()
 *
 * Description:
 *   Allocate/free the requested node of a B-tree of the hfs filesystem
 *   by setting/clearing the corresponding bit in the B-tree bitmap.
 *   The size of the B-tree will not be changed.
 * Input Variable(s):
 *   struct hfs_btree *tree: Pointer to a B-tree
 *   __u32 bitnr: The node number to free
 *   int set: 0 to clear the bit, non-zero to set it.
 * Output Variable(s):
 *   None
 * Returns:
 *    0: no error
 *   -1: The node was already allocated/free, nothing has been done.
 *   -2: The node is out of range of the B-tree.
 *   -3: hfs_bnode_find() failed
 *   -4: note enough map nodes to hold all the bits
 * Preconditions:
 *   'tree' points to a "valid" (struct hfs_btree)
 *   'bitnr' is a node number within the range of the btree, which is
 *   currently free/allocated.
 * Postconditions:
 *   The bit number 'bitnr' of the node bitmap is set/cleared and the
 *   number of free nodes in the btree is decremented/incremented by one.
 */
int hfs_bnode_bitop(struct hfs_btree *tree, __u32 bitnr, int set)
{
	struct hfs_bnode_ref *bnr = NULL; /* pointer to the current bnode */
	int retval = 0;		/* the return value of the function */
	__u32 next_node = 0;	/* next bnode in chain (start in header)*/
	__u32 bmap_size;	/* number of bits in the current bitmap */
	int map_record;		/* record number of the bitmap in this bnode */
	unsigned long *lword;	/* address of the long containing the bit */

	if (bitnr >= from_l(btree_head(tree)->bthNNodes)) {
		printk("hfs_bnode_bitop: node number out of range.\n");
		retval = -2;
		goto bail2;
	}

	for (;;) {
		if (!(bnr = hfs_bnode_find(tree, next_node, HFS_LOCK_BLIND))) {
			printk("hfs_bnode_bitop: map node not found.\n");
			retval = -3;
			goto bail2;
		}

		map_record = from_w(bnode_head(bnr->bn)->ndNRecs);
		bmap_size = 8*bnode_rsize(bnr->bn, map_record);

		if (bitnr < bmap_size) {
			break;
		}

		/* continue on to next map node if available */
		if (!(next_node = from_l(bnode_head(bnr->bn)->ndFLink))) {
			printk("hfs_bnode_bitop: too few map nodes.\n");
			retval = -4;
			goto bail1;
		}
		hfs_bnode_relse(bnr);
		bitnr -= bmap_size;
	}

	lword = (unsigned long *)bnode_address(bnr->bn, map_record)+(bitnr>>5);
	bitnr %= 32;
	if ((set && set_bit(HFS_BE_BIT(bitnr), lword)) ||
	    (!set && !clear_bit(HFS_BE_BIT(bitnr), lword))) {
		printk("hfs_bnode_bitop: bitmap corruption.\n");
		retval = -1;
		goto bail1;
	}
	bnode_dirty(bnr->bn);

	/* adjust the free count */
	add_l(&btree_head(tree)->bthFree, (set ? -1 : 1));
	btree_dirty(tree);

bail1:
	hfs_bnode_relse(bnr);
bail2:
	return retval;
}

/*
 * hfs_bnode_init()
 *
 * Description:
 *   Initialize a newly allocated bnode.
 * Input Variable(s):
 *   struct hfs_btree *tree: Pointer to a B-tree
 *   __u32 node: the node number to allocate
 * Output Variable(s):
 *   NONE
 * Returns:
 *   (struct hfs_bnode *) for the new bnode or NULL
 * Preconditions:
 *   'tree' points to a "valid" (struct hfs_btree)
 *   'node' existes and has been allocated in the bitmap of bnodes.
 * Postconditions:
 *   On success:
 *    The node is not read from disk, nor added to the bnode cache.
 *    The 'sticky' and locking-related fields are all zero/NULL.
 *    The bnode's nd{[FB]Link, Type, NHeight} fields are uninitialized.
 *    The bnode's ndNRecs field and offsets table indicate an empty bnode.
 *   On failure:
 *    The node is deallocated.
 */
struct hfs_bnode * hfs_bnode_init(struct hfs_btree * tree, __u32 node)
{
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	extern int bnode_count;
#endif
	struct hfs_bnode *retval;

	if (!(retval = KMALLOC(sizeof(*retval), GFP_KERNEL))) {
		printk("hfs_bnode_init: out of memory.\n");
		goto bail2;
	}

	/* Partially initialize the in-core structure */
	memset(retval, 0, sizeof(*retval));
	retval->magic = HFS_BNODE_MAGIC;
	retval->tree = tree;
	retval->node = node;
	if (!(retval->bh = hfs_getblk(&tree->file, node, 0))) {
		printk("hfs_bnode_init: unable to read node from dev "
		       "0x%04x\n", tree->file.dev);
		goto bail1;
	}

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	++bnode_count;
#endif

	/* Partially initialize the on-disk structure */
	bnode_head(retval)->ndNRecs = c_to_w(0);
	bnode_head(retval)->ndResv2 = c_to_w(0);
        RECTBL(retval, 1) = c_to_w(sizeof(struct NodeDescriptor));
	mark_buffer_uptodate(retval->bh, 1);
	bnode_dirty(retval);

	return retval;

bail1:
	KFREE_S(retval, sizeof(*retval));
bail2:
	/* clear the bit in the bitmap */
	hfs_bnode_bitop(tree, node, 0);
	return NULL;
}

/*
 * hfs_bnode_alloc()
 *
 * Description:
 *   Find a cleared bit in the B-tree node bitmap of the hfs filesystem,
 *   set it and return the corresponding bnode, with its contents zeroed.
 *   When there is no free bnode in the tree, an error is returned, no
 *   new nodes will be added by this function!
 * Input Variable(s):
 *   struct hfs_btree *tree: Pointer to a B-tree
 * Output Variable(s):
 *   NONE
 * Returns:
 *   (struct hfs_bnode *) for the new bnode or NULL
 * Preconditions:
 *   'tree' points to a "valid" (struct hfs_btree)
 *   There is at least one free bnode.
 * Postconditions:
 *   On success:
 *     The corresponding bit in the btree bitmap is set.
 *     The number of free nodes in the btree is decremented by one.
 *   The node is not read from disk, nor added to the bnode cache.
 *   The 'sticky' field is uninitialized.
 */
struct hfs_bnode *hfs_bnode_alloc(struct hfs_btree *tree)
{
	struct hfs_bnode_ref *bnr = NULL; /* the pointer to the current bnode */
	__u32 bitnr = 0;	/* which bit are we examining */
	__u32 bmap_size;	/* number of bits in current bitmap */
	__u16 map_record;	/* record number of the bitmap in this bnode */
	__u16 first;		/* the first clear bit in this bnode */
	unsigned long *bitmap;	/* address of the bitmap in this bnode */
	__u32 next_node = 0;	/* next bnode in chain (start in header) */
	
	for (;;) {
		if (!(bnr = hfs_bnode_find(tree, next_node, HFS_LOCK_BLIND))) {
			printk("hfs_bnode_alloc: map node not found.\n");
			goto bail2;
		}

		map_record = from_w(bnode_head(bnr->bn)->ndNRecs);
		bmap_size = 8*bnode_rsize(bnr->bn, map_record);
		bitmap = (unsigned long *)bnode_address(bnr->bn, map_record);
		
		/* search the current node */
		first = hfs_find_first_zero_bit(bitmap,	bmap_size);
		if (first < bmap_size) {
			break;
		}

		/* continue search in next map node */
		next_node = from_l(bnode_head(bnr->bn)->ndFLink);
		hfs_bnode_relse(bnr);

		if (!next_node) {
			printk("hfs_bnode_alloc: too few map nodes.\n");
			goto bail2;
		}
		bitnr += bmap_size;
	}

	if ((bitnr += first) >= from_l(btree_head(tree)->bthNNodes)) {
		printk("hfs_bnode_alloc: no free nodes found, count wrong?\n");
		goto bail1;
	}

	if (set_bit(HFS_BE_BIT(first % 32), bitmap + (first>>5))) {
		printk("hfs_bnode_alloc: bitmap corruption.\n");
		goto bail1;
	}
	bnode_dirty(bnr->bn);
	hfs_bnode_relse(bnr);

	/* decrement the free count */
	add_l(&btree_head(tree)->bthFree, -1);
	btree_dirty(tree);

	return hfs_bnode_init(tree, bitnr);

bail1:
	hfs_bnode_relse(bnr);
bail2:
	return NULL;
}


/*
 * init_mapnode()
 *
 * Description:
 *   Initializes a given node as a mapnode in the given tree.
 * Input Variable(s):
 *   struct hfs_btree *tree: the tree to add a map node to.
 *   __u32: the node to use as a mapnode.
 * Output Variable(s):
 *   NONE
 * Returns:
 *   struct hfs_bnode_ref *: the new mapnode or NULL
 * Preconditions:
 *   'tree' is a valid (struct hfs_btree).
 *   'node' is the number of the first node in 'tree' that is not
 *    represented by a bit in the existing mapnodes.
 * Postconditions:
 *   On failure 'tree' is unchanged and NULL is returned.
 *   On success the node given by 'node' has been added to the linked
 *    list of mapnodes attached to 'tree', and has been initialized as
 *    a valid mapnode with its first bit set to indicate it's self as
 *    allocated.
 */
static struct hfs_bnode_ref * init_mapnode(struct hfs_btree *tree, __u32 node)
{
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	extern int bnode_count;
#endif
	struct hfs_bnode_ref *retval;

	if (!(retval = KMALLOC(sizeof(*retval), GFP_KERNEL))) {
		printk("hfs_bnode_add: out of memory.\n");
		goto bail3;
	}
	if (!(retval->bn = KMALLOC(sizeof(*retval->bn), GFP_KERNEL))) {
		printk("hfs_bnode_add: out of memory.\n");
		goto bail2;
	}

	memset(retval->bn, 0, sizeof(*retval->bn));
	retval->bn->magic = HFS_BNODE_MAGIC;
	retval->bn->tree = tree;
	retval->bn->node = node;
        retval->bn->sticky = HFS_NOT_STICKY;
	if (!(retval->bn->bh = hfs_getblk(&tree->file, node, 0))) {
		printk("hfs_bnode_add: unable to read node %d from dev "
		       "0x%04x\n", node, tree->file.dev);
		goto bail1;
	}

        retval->lock_type = HFS_LOCK_NONE;
        hfs_bnode_lock(retval, HFS_LOCK_WRITE);
        if ((retval->bn->next = bhash(tree, retval->bn->node))) {
                retval->bn->next->prev = retval->bn;
        }
        bhash(tree, retval->bn->node) = retval->bn;

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	++bnode_count;
#endif

	/* Partially initialize the on-disk structure */
	bnode_head(retval->bn)->ndType = ndMapNode;
	bnode_head(retval->bn)->ndNRecs = c_to_w(1);
        RECTBL(retval->bn, 1) = c_to_w(sizeof(struct NodeDescriptor));
        RECTBL(retval->bn, 2) = c_to_w(0x1fa);
	set_bit(HFS_BE_BIT(0), bnode_address(retval->bn, 1));
	mark_buffer_uptodate(retval->bn->bh, 1);
	return retval;

bail1:
	KFREE_S(retval->bn, sizeof(*retval->bn));
bail2:
	KFREE_S(retval, sizeof(*retval));
bail3:
	return NULL;
}

/*
 * hfs_btree_extend()
 *
 * Description:
 *   Adds nodes to a B*-tree if possible.
 * Input Variable(s):
 *   struct hfs_btree *tree: the btree to add nodes to.
 * Output Variable(s):
 *   NONE
 * Returns:
 *   void
 * Preconditions:
 *   'tree' is a valid (struct hfs_btree *).
 * Postconditions:
 *   If possible the number of nodes indicated by the tree's clumpsize
 *    have been added to the tree, updating all in-core and on-disk
 *    allocation information.
 *   If insufficient disk-space was avaialble then fewer nodes may have
 *    been added than would be expected based on the clumpsize.
 *   In the case of the extents B*-tree this function will add fewer
 *    nodes than expected if adding more would result in an extent
 *    record for the extents tree being added to the extents tree.
 *    The situation could be dealt with, but doing so confuses Macs.
 */
void hfs_btree_extend(struct hfs_btree *tree)
{
	struct hfs_bnode_ref *bnr = NULL;
	struct hfs_bnode_ref *tmp = NULL;
	__u32 old_nodes, new_nodes, total_nodes, new_mapnodes, seen;
	__u32 next_node;
	__u16 ablksz, ablock;
	int map_record;

        ablksz = HFS_SB(tree->file.inode->i_sb)->s_alloc_blksz;
	old_nodes = tree->file.inode->i_blocks;
	ablock = ((tree->file.clumpsize >> HFS_SECTOR_SIZE_BITS) +
		   old_nodes + ablksz - 1) / ablksz - 1;

	hfs_file_extend(&tree->file, ablock);
	total_nodes = tree->file.inode->i_blocks;
	new_nodes = total_nodes- old_nodes;
	if (!new_nodes) {
		return;
	}

	if (!(bnr = hfs_bnode_find(tree, 0, HFS_LOCK_BLIND))) {
		printk("hfs_btree_extend: header node not found.\n");
		return;
	}

	seen = 0;
	new_mapnodes = 0;
	for (;;) {
		map_record = from_w(bnode_head(bnr->bn)->ndNRecs);
		seen += 8*bnode_rsize(bnr->bn, map_record);
		next_node = from_l(bnode_head(bnr->bn)->ndFLink);

		if (seen >= total_nodes) {
			hfs_bnode_relse(bnr);
			break;
		}

		if (!next_node) {
			if (!(tmp = init_mapnode(tree, seen))) {
				printk("hfs_btree_extend: "
				       "can't build mapnode.\n");
				hfs_bnode_relse(bnr);
				return;
			}
			bnode_head(tmp->bn)->ndBLink = c_to_l(bnr->bn->node);
			bnode_dirty(tmp->bn);
			bnode_head(bnr->bn)->ndFLink = to_l(seen);
			bnode_dirty(bnr->bn);
			++new_mapnodes;
		} else {
		 	if (!(tmp = hfs_bnode_find(tree, next_node,
						   HFS_LOCK_BLIND))) {
				printk("hfs_btree_extend: "
				       "mapnode not found.\n");
				hfs_bnode_relse(bnr);
				return;
			}
		}
		hfs_bnode_relse(bnr);
		bnr = tmp;
	}

	btree_head(tree)->bthNNodes = to_l(total_nodes);
	add_l(&btree_head(tree)->bthFree, new_nodes - new_mapnodes);
	btree_dirty(tree);
	hfs_extent_out(&tree->file, tree->raw_extent);
	*(tree->size) = to_l(total_nodes << HFS_SECTOR_SIZE_BITS);
	mark_buffer_dirty(HFS_SB(tree->file.inode->i_sb)->s_sbbh, 1);

	/* XXX should write the backup MDB here */

	return;
}
