/*
 * linux/fs/hfs/bfind.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the code to find records in the B-tree structure.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 *
 * The code in this file initilizes some structures by calling
 * memset(&foo, 0, sizeof(foo)).  This produces the desired behavior
 * only due to the non-ANSI assumption that the machine representation
 * of NULL is all zeros.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include "hfs_fs.h"
#include "hfs_btree.h"

/*================ Global functions ================*/

/*
 * hfs_bfind()
 *
 * Description:
 *   This is the wrapper for hfs_brec_find() that allows access from outside
 *   this source file.	Its main purpose is to construct an allowable
 *   'flags' argument for hfs_brec_find().
 * Input Variable(s):
 *   struct hfs_btree *tree: pointer to the (struct hfs_btree) to search
 *   struct hfs_brec_key *target: pointer to the (struct hfs_brec_key) to find
 *   struct hfs_brec **brec: location to store a pointer to the located
 *    (struct hfs_brec) or NULL.
 *   int mode: the search mode which is converted to a valid set of
 *    flags for the call to hfs_brec_find()
 * Output Variable(s):
 *   struct hfs_brec **brec: the value hfs_brec_find() produces in the
 *    argument by the same name, or NULL
 * Returns:
 *   int: -EINVAL if 'mode' is invalid, otherwise the return value from hfs_brec_find()
 * Preconditions:
 *   The three pointer arguments satisfy the preconditions of hfs_brec_find()
 * Postconditions:
 *   If 'mode' is not a recognized mode, then '*brec' is NULL and
 *   -EINVAL is returned.  Otherwise both the return value and '*brec'
 *   are determined by calling hfs_brec_find() with the 'flags' argument
 *   deterimined by 'mode.'
 *   See below for the mode-to-flags correspondence.
 */
int hfs_bfind(struct hfs_btree *tree, const struct hfs_brec_key *target,
	      struct hfs_brec **brec, int mode)
{
	int flags;

	if (brec) {
		*brec = NULL;
	}

	if (mode == HFS_BFIND_READ_EQ) {
		flags = HFS_BFIND_RDONLY | HFS_BFIND_EXACT;
	} else if (mode == HFS_BFIND_READ_LE) {
		flags = HFS_BFIND_RDONLY;
	} else if (mode == HFS_BFIND_WRITE) {
		flags = HFS_BFIND_EXACT | HFS_BFIND_LOCK;
	} else {
		printk("hfs_bfind: invalid mode %d\n", mode);
		return -EINVAL;
	}
	return hfs_brec_find(tree, target, brec, flags);
}

/*
 * hfs_bsucc()
 *
 * Description:
 *   This function overwrites '**brec' with its successor in the B-tree,
 *   obtaing the same type of access.
 * Input Variable(s):
 *   struct hfs_brec **brec: address of the address of the (struct hfs_brec)
 *     to overwrite with its successor
 * Output Variable(s):
 *   struct hfs_brec **brec: the address of the address of the successor of
 *    the original '**brec' or address of NULL
 * Returns:
 *   int: 0 on success, or one of -EINVAL, -EIO, or -EINVAL on failure
 * Preconditions:
 *   'brec' is NULL, points to NULL, or points to a pointer to a "valid"
 *    (struct hfs_brec)
 * Postconditions:
 *   If the given 'brec' is not "valid" -EINVAL is returned and
 *    '*brec' and '**brec' are unchanged.
 *   If the given 'brec' is "valid" but has no successor then -ENOENT
 *    is returned and '*brec' is NULL.
 *   If a call to hfs_bnode_find() is neccessary to find the successor,
 *    but fails then -EIO is returned and '*brec' is NULL.
 *   If none of the three previous conditions prevents finding the
 *    successor of '**brec', then 0 is returned, '*brec' is unchanged
 *    and '**brec' is overwritten with the (struct hfs_brec) for its
 *    successor.
 *   In the two cases where '*brec' is set to NULL, the (struct
 *    hfs_brec) previously indicated by '*brec' is freed first.
 */
int hfs_bsucc(struct hfs_brec **brec)
{
	struct hfs_brec *tmp;
	struct hfs_bnode_ref *old;
	struct hfs_bnode *bn;
	int node;

	if (brec && (tmp = *brec) && (tmp->magic == HFS_BREC_MAGIC) && 
	    (bn = tmp->bnode->bn) && (bn->magic == HFS_BNODE_MAGIC) &&
	    bn->tree && (bn->tree->magic == HFS_BTREE_MAGIC) && bn->bh) {
		if (tmp->record == from_w(bnode_head(bn)->ndNRecs)) {
			/* current record is the last one in its node,
			   so we must get next node */
			if (!(node = from_l(bnode_head(bn)->ndFLink))) {
				hfs_brec_relse(brec);
				return -ENOENT;
			}
			old = tmp->bnode;
			tmp->bnode = hfs_bnode_find(bn->tree, node,
						tmp->bnode->lock_type);
			hfs_bnode_relse(old);
			if (!tmp->bnode) {
				printk("hfs_bsucc: unable to read node from "
				       "dev 0x%04x\n",
				       tmp->bnode->bn->tree->file.dev);
				hfs_brec_relse(brec);
				return -EIO;
			}
			tmp->record = 1;
		} else {
			tmp->record++;
		}
		tmp->key = brec_key(tmp);
		tmp->data.generic = bkey_address(tmp->key);

		return 0;
	}
	printk("hfs_bsucc: invalid/corrupt arguments.\n");
	return -EINVAL;
}
