/*
 * linux/fs/hfs/bins_del.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the code to insert and delete records in a B-tree.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include "hfs_fs.h"
#include "hfs_btree.h"

/*================ File-local functions ================*/

#if 0
static void dump_node(struct hfs_bnode *bn)
{
	__u8 *ptr;
	int i, j, nrecs;
	printk("node %d:\n", bn->node);
	printk(" ndFLink: %08lx\n", from_l(bnode_head(bn)->ndFLink));
	printk(" ndBLink: %08lx\n", from_l(bnode_head(bn)->ndBLink));
	printk(" ndType: %02x\n", bnode_head(bn)->ndType);
	printk(" ndNHeight: %02x\n", bnode_head(bn)->ndNHeight);
	printk(" ndRecs: %04x\n", nrecs=from_w(bnode_head(bn)->ndNRecs));
	for (i=1; i<=nrecs; ++i) {
		printk(" record %d:\n", i);
		printk("  key:");
		ptr = (__u8 *)bnode_key(bn,i);
		for (j=0; j<bnode_key(bn,i)->KeyLen; ++j, ++ptr) {
			printk("%02x ", *ptr);
		}
		printk("\n");
		printk("  data:");
		ptr = (__u8 *)bnode_key(bn,i) + ROUND(1+ bnode_key(bn,i)->KeyLen);
		for (j=0; j<bnode_rsize(bn,i) - ROUND(1+ bnode_key(bn,i)->KeyLen); ++j, ++ptr) {
			printk("%02x ", *ptr);
		}
		printk("\n");
	}
	printk(" freeptr: %04x\n", bnode_offset(bn,nrecs+1));
}
#endif

/*
 * binsert_nonfull()
 *
 * Description:
 *   Inserts a record in a given bnode known to have sufficient space.
 * Input Variable(s):
 *   struct hfs_btree* tree: pointer to the tree to insert in
 *   struct hfs_brec** brec: path to node to insert in
 *   struct hfs_nrec_key* key: pointer to the key for the record to insert
 *   void* data: pointer to the record to insert
 *   __u16 datasize: size of the record to insert
 * Output Variable(s):
 *   NONE
 * Returns:
 *   NONE
 * Preconditions:
 *   'tree' is a pointer to a valid (struct hfs_btree).
 *   'brec' is a pointer to a poiner to a valid linked list of
 *    (struct hfs_brec)s, with sufficient free space in its bottom node
 *    to insert the given record.
 *   'key' is a pointer to a valid (struct hfs_brec_key) which, in sorted
 *    order, belongs at the location indicated by 'brec'.
 *   'data' is non-NULL an points to appropriate data of length 'datasize'
 * Postconditions:
 *   The record has been inserted in the position indicated by 'brec'.
 *   If the new record is in the first position in its bnode, then the
 *   key for this bnode in its parent is updated and this condition is
 *   applied recursivly to the parent and its ancestors.
 */
static void binsert_nonfull(struct hfs_btree *tree, struct hfs_brec **brec,
			    const struct hfs_brec_key *key, const void *data,
			    __u8 keysize, __u16 datasize)
{
	int i, rec, nrecs, size, tomove;
	__u8 *start;
	struct hfs_brec *tmp;
	struct hfs_bnode *bnode = (*brec)->bnode->bn;

	rec = ++((*brec)->record);
	size = ROUND(keysize+1) + datasize;
	nrecs = add_w(&bnode_head(bnode)->ndNRecs, 1);
	tomove = bnode_offset(bnode, nrecs) - bnode_offset(bnode, rec);
	
	/* adjust the record table */
	for (i = nrecs; i >= rec; --i) {
		RECTBL(bnode,i+1) = to_w(bnode_offset(bnode,i) + size);
	}

	/* make room */
	start = bnode_address(bnode, rec);
	memmove(start + size, start, tomove);

	/* copy in the key and the data*/
	*start = keysize;
	keysize = ROUND(keysize+1);
	memcpy(start + 1, (__u8 *)key + 1, keysize-1);
	memcpy(start + keysize, data, datasize);

	/* update keys in ancestors as needed */
	if (!((*brec)->flags & HFS_BREC_RIGHT)) {
		tmp = (*brec);
		keysize = tree->bthKeyLen;
		while ((tmp->record == 1) && (tmp = tmp->up)) {
			memcpy(brec_address(tmp)+1, (char *)key+1, keysize);
		}
	}
}

/*
 * shift_right()
 *
 * Description:
 *   Shifts some records from a node to its right neighbor.
 * Input Variable(s):
 *   struct hfs_bnode* left: the node to shift records from
 *   struct hfs_bnode* right: the node to shift records to
 *   __u16 first: the number of the first record in 'left' to move to 'right'
 * Output Variable(s):
 *   NONE
 * Returns:
 *   NONE
 * Preconditions:
 *   'left' and 'right' point to valid (struct hfs_bnode)s.
 *   'left' contaitns at least 'first' records.
 *   'right' has enough free space to hold the records to be moved from 'left'
 * Postconditions:
 *   The record numbered 'first' and all records after it in 'left' are
 *   placed at the beginning of 'right'.
 *   The key corresponding to 'right' in its parent is NOT updated.
 */
static void shift_right(struct hfs_bnode *left, struct hfs_bnode *right,
			__u16 first)
{
	int i, adjust, nrecs;
	unsigned size;
	HFS_WORD *to, *from;

	if (first) {
		/* initialize variables */
		nrecs = from_w(bnode_head(left)->ndNRecs) + 1 - first;
		size = bnode_end(left) - bnode_offset(left, first);
		/* move (possibly empty) contents of right node forward */
		memmove(bnode_datastart(right) + size,
			bnode_datastart(right), 
			bnode_end(right) - sizeof(struct NodeDescriptor));

		/* copy in new records */
		memcpy(bnode_datastart(right), bnode_address(left,first), size);

		/* fix up offsets in right node */
		i = from_w(bnode_head(right)->ndNRecs) + 1;
		from = &RECTBL(right, i);
		to = from - nrecs;
		while (i--) {
			*(to++) = to_w(size + from_w(*(from++)));
		}
		adjust = sizeof(struct NodeDescriptor) - 
			 bnode_offset(left, first);
		i = nrecs-1;
		from = &RECTBL(left, first+i);
		while (i--) {
			*(to++) = to_w(adjust + from_w(*(from++)));
		}

		/* fix record counts */
		add_w(&bnode_head(left)->ndNRecs, -nrecs);
		add_w(&bnode_head(right)->ndNRecs, nrecs);
	}
}

/*
 * NAME()
 *
 * Description:
 *   
 * Input Variable(s):
 *   
 * Output Variable(s):
 *   
 * Returns:
 *   
 * Preconditions:
 *   
 * Postconditions:
 *   
 */
static void new_root(struct hfs_btree *tree,
		     struct hfs_bnode_ref *left,
		     struct hfs_bnode_ref *right)
{
	struct hfs_bnode *root;
	struct hfs_brec_key *key;
	int keylen = tree->bthKeyLen;

	if (left && !right) {
		printk("new_root: LEFT but no RIGHT\n");
		return;
	}

	if (!(root = hfs_bnode_alloc(tree))) {
		return;
	}

	root->sticky = HFS_STICKY;
	tree->root = root;
	tree->bthRoot = root->node;
	btree_head(tree)->bthRoot = to_l(root->node);
	add_w(&btree_head(tree)->bthDepth, 1);

	bnode_head(root)->ndNHeight = from_w(btree_head(tree)->bthDepth);
	bnode_head(root)->ndFLink = c_to_l(0);
	bnode_head(root)->ndBLink = c_to_l(0);
	if (!left) {
		/* tree was empty */
		btree_head(tree)->bthFNode = to_l(root->node);
		btree_head(tree)->bthLNode = to_l(root->node);
		bnode_head(root)->ndType = ndLeafNode;
	} else {
		bnode_head(root)->ndType = ndIndxNode;

		RECTBL(root, 2) = to_w(sizeof(struct NodeDescriptor) +
				       ROUND(1+keylen) + sizeof(HFS_LWORD));
		key = bnode_key(root, 1);
		key->KeyLen = keylen;
		memcpy(&key->u, &bnode_key(left->bn, 1)->u, keylen);
		*((HFS_LWORD *)bkey_address(key)) = to_l(left->bn->node);
		
		RECTBL(root, 3) = to_w(sizeof(struct NodeDescriptor) +
				       2*ROUND(1+keylen) + 2*sizeof(HFS_LWORD));
		key = bnode_key(root, 2);
		key->KeyLen = keylen;
		memcpy(&key->u, &bnode_key(right->bn, 1)->u, keylen);
		*((HFS_LWORD *)bkey_address(key)) = to_l(right->bn->node);

		bnode_head(root)->ndNRecs = c_to_w(2);

		/* the former root (left) is now just a normal node */
		left->bn->sticky = HFS_NOT_STICKY;
		if ((left->bn->next = bhash(tree, left->bn->node))) {
			left->bn->next->prev = left->bn;
		}
		bhash(tree, left->bn->node) = left->bn;
	}
	btree_dirty(tree);
	bnode_dirty(root);
}

/*
 * insert_empty_bnode()
 *
 * Description:
 *   Adds an empty node to the right of 'left'.
 * Input Variable(s):
 *   struct hfs_btree *tree: the tree to add a node to
 *   struct hfs_bnode *left: the node to add a node after
 * Output Variable(s):
 *   NONE
 * Returns:
 *   struct hfs_bnode_ref *: reference to the new bnode.
 * Preconditions:
 *   'tree' points to a valid (struct hfs_btree) with at least 1 free node.
 *   'left' points to a valid (struct hfs_bnode) belonging to 'tree'.
 * Postconditions:
 *   If NULL is returned then 'tree' and 'left' are unchanged.
 *   Otherwise a node with 0 records is inderted in the tree to the right
 *   of the node 'left'.  The 'ndFLink' of 'left' and the 'ndBLink' of
 *   the former right-negihbor of 'left' (if one existed) point to the
 *   new node.  If 'left' had no right neighbor and is a leaf node the
 *   the 'bthLNode' of 'tree' point to the new node.  The free-count and
 *   bitmap for 'tree' are kept current by hfs_bnode_alloc() with supplies
 *   the required node.
 */
static struct hfs_bnode_ref * insert_empty_bnode(struct hfs_btree *tree,
						 const struct hfs_bnode *left)
{
	struct hfs_bnode_ref *retval;
	struct hfs_bnode_ref *right;
	__u32 node;

	if (!(retval = KMALLOC(sizeof(*retval), GFP_KERNEL))) {
		printk("hfs_binsert: out of memory.\n");
		goto bail3;
	}

	retval->lock_type = HFS_LOCK_NONE;
	if (!(retval->bn = hfs_bnode_alloc(tree))) {
		printk("hfs_binsert: out of bnodes?.\n");
		goto bail2;
	}
	hfs_bnode_lock(retval, HFS_LOCK_WRITE);
	retval->bn->sticky = HFS_NOT_STICKY;
	if ((retval->bn->next = bhash(tree, retval->bn->node))) {
		retval->bn->next->prev = retval->bn;
	}
	bhash(tree, retval->bn->node) = retval->bn;

        if ((node = from_l(bnode_head(left)->ndFLink))) {
		/* XXX if the tree is corrupt this could block: */
 		if (!(right = hfs_bnode_find(tree, node, HFS_LOCK_WRITE))) {
			printk("hfs_binsert: corrupt btree.\n");
			goto bail1;
		}
                bnode_head(right->bn)->ndBLink = to_l(retval->bn->node);
                hfs_bnode_relse(right);
        } else if (bnode_head(left)->ndType == ndLeafNode) {
                btree_head(tree)->bthLNode = to_l(retval->bn->node);
                btree_dirty(tree);
	}

        bnode_head(retval->bn)->ndFLink = bnode_head(left)->ndFLink;
        bnode_head(retval->bn)->ndBLink = to_l(left->node);
        bnode_head(retval->bn)->ndType = bnode_head(left)->ndType;
        bnode_head(retval->bn)->ndNHeight = bnode_head(left)->ndNHeight;

        bnode_head(left)->ndFLink = to_l(retval->bn->node);

	return retval;

bail1:
	hfs_bnode_bitop(tree, retval->bn->node, 0);
bail2:
	hfs_bnode_relse(retval);
bail3:
	return NULL;
}

/*
 * split()
 *
 * Description:
 *   Splits an overfull node durring insertion.
 *   Picks the split point that results in the most-nearly equal
 *   space usage in the new and old nodes.
 * Input Variable(s):
 *   struct hfs_bnode_ref *left: points to the left (overfull) node.
 *   struct hfs_bnode_ref *right: points to the right (empty) node.
 *   struct hfs_brec *brec: indicates where the new record will be inserted.
 *   int size: the number of bytes to be used by the new record and its key.
 * Output Variable(s):
 *   struct hfs_brec *brec: chnaged to indicate when the new record goes now.
 * Returns:
 *   int: 0 on success, non-zero on failure  XXX OUT OF DATE 
 * Preconditions:
 *   'left' and 'right' point to valid (struct hfs_bnode_ref)s.
 *   'brec' points to a valid (struct hfs_brec).
 * Postconditions:
 *   The records in 'left' are redistributed across 'left' and 'right' so
 *   that after inserting the new record, the space usage in these two
 *   bnodes is as equal as possible.
 *   'brec' is updated so that a call to binsert_nonfull() will insert the
 *   new record in the correct location.
 */
static struct hfs_bnode_ref * split(struct hfs_bnode_ref *left,
		 		    struct hfs_brec *brec, int size)
{
	struct hfs_bnode *bnode = left->bn;
	int nrecs, cutoff, index, tmp, used, in_right;
	struct hfs_bnode_ref *right;

	if ((right = insert_empty_bnode(bnode->tree, bnode))) {
		nrecs = from_w(bnode_head(bnode)->ndNRecs);
		cutoff = (size + bnode_offset(bnode, nrecs+1) -
			      sizeof(struct NodeDescriptor) +
			      (nrecs+1)*sizeof(HFS_WORD))/2;
		used = 0;
		in_right = 1;
		/* note that this only works because records sizes are even */
		for (index=1; index <= brec->record; ++index) {
			tmp = bnode_rsize(bnode, index)/2;
			used += tmp + sizeof(HFS_WORD);
			if (used >= cutoff) {
				goto found;
			}
			used += tmp;
		}
		size /= 2;
		used += size + sizeof(HFS_WORD);
		if (used >= cutoff) {
			goto found;
		}
		in_right = 0;
		used += size;
		for (; index <= nrecs; ++index) {
			tmp = bnode_rsize(bnode, index)/2;
			used += tmp + sizeof(HFS_WORD);
			if (used >= cutoff) {
				goto found;
			}
			used += tmp;
		}
		/* couldn't find the split point! */
		hfs_bnode_relse(right);
	}
	return NULL;

found:
	if (in_right) {
		brec->bnode = right;
		brec->record -= index-1;
		brec->flags |= HFS_BREC_RIGHT;
	}
	shift_right(bnode, right->bn, index);

	return right;
}

/*
 * binsert()
 *
 * Description:
 *   Inserts a record in a tree known to have enough room, even if the
 *   insertion requires the splitting of nodes.
 * Input Variable(s):
 *    struct hfs_btree *tree: the tree to insert in
 *    struct hfs_brec **brec: partial path to the node to insert in
 *    const struct hfs_brec_key *key: key for the new record
 *    const void *data: data for the new record
 *    __u8 keysize: size of the key
 *    __u16 datasize: size of the data
 *    int reserve: number of nodes reserved in case of splits
 * Output Variable(s):
 *    *brec = NULL
 * Returns:
 *    int: 0 on sucess, error code on failure
 * Preconditions:
 *    'tree' points to a valid (struct hfs_btree) which has enough free
 *     nodes to complete this insertion in the event it requires splitting
 *     of nodes, and those nodes a reserved using the 'reserved' field of
 *     tree.
 *    'brec' points to a pointer to a valid (struct hfs_brec) corresponding
 *     to a record in a leaf node, after which a record is to be inserted,
 *     or to "record 0" of the leaf node if the record is to be inserted
 *     before all existing records in the node.  The linked list of
 *     (struct hfs_brec)'s includes all ancestors of the leaf node that
 *     are needed to complete the insertion including the parents of any
 *     nodes that will be split.
 *    'key' points to a valid (struct hfs_brec_key) which is appropriate
 *     to this tree, and which belongs at the insertion point.
 *    'data' points data appropriate for the indicated node.
 *    'keysize' gives the size in bytes of the key.
 *    'datasize' gives the size in bytes of the data.
 *    'reserve' gives the number of nodes that have been reserved in the
 *     tree to allow for splitting of nodes.
 * Postconditions:
 *    All 'reserve'd nodes have been either used or released.
 *    *brec = NULL
 *    On success the key and data have been inserted at the indicated
 *    location in the tree, all appropriate fields of the in-core data
 *    structures have been changed and updated versions of the on-disk
 *    data structures have been scheduled for write-back to disk.
 *    On failure the B*-tree is probably invalid both on disk and in-core.
 *
 *    XXX: Some attempt at repair might be made in the event of failure,
 *    or the fs should be remounted read-only so things don't get worse.
 */
static int binsert(struct hfs_btree *tree, struct hfs_brec **brec,
		   const struct hfs_brec_key *key, const void *data,
		   __u8 keysize, __u16 datasize, int reserve)
{
	struct hfs_bnode_ref *right = NULL;
	struct hfs_bnode_ref *left = NULL;
	struct hfs_bnode_ref *old_right, *old_left;
	HFS_LWORD node;
	
	while ((*brec) != NULL) {
		old_right = right;
		old_left = left;
		right = NULL;
		left = NULL;
		if ((*brec)->flags & HFS_BREC_OVERFLOW) {
			left = (*brec)->bnode;
			right = split(left, *brec,
				      ROUND(keysize+1) + ROUND(datasize));
			--reserve;
			--tree->reserved;
			if (right == NULL) {
				printk("hfs_binsert: unable to split node!\n");
				tree->reserved -= reserve;
				hfs_brec_relse(brec);
				return -ENOSPC;
			}
			binsert_nonfull(tree, brec, key, data,
					keysize, datasize);
			hfs_bnode_lock(left, HFS_LOCK_READ);
			hfs_bnode_lock(right, HFS_LOCK_READ);
			(*brec)->bnode = NULL;
		
			if (left->bn->node == tree->root->node) {
				new_root(tree, left, right);
				tree->reserved -= reserve;
				hfs_brec_relse(brec);
			} else if ((*brec)->up) {
				*brec = (*brec)->up;
				hfs_brec_relse_one(&(*brec)->down);
				node = to_l(right->bn->node);
				data = &node;
				datasize = sizeof(node);
				key = bnode_key(right->bn, 1);
				keysize = tree->bthKeyLen;
			} else {
				printk("hfs_binsert: Missing parent.\n");
				tree->reserved -= reserve;
				hfs_brec_relse(brec);
				return -EIO;
			}
		} else {
			tree->reserved -= reserve;
			binsert_nonfull(tree, brec, key, data,
					keysize, datasize);
			hfs_brec_relse(brec);
		}
		hfs_bnode_relse(old_left);
		hfs_bnode_relse(old_right);
	}
	hfs_bnode_relse(left);
	hfs_bnode_relse(right);
	return 0;
}

/*
 * available()
 *
 * short-hand for calculating the available nodes in a tree
 */
static inline int available(const struct hfs_btree *tree)
{
	return from_l(btree_head(tree)->bthFree) - tree->reserved;
}

/*================ Global functions ================*/

/*
 * hfs_binsert()
 *
 * Description:
 *   This function inserts a new record into a b-tree.
 * Input Variable(s):
 *   struct hfs_btree *tree: pointer to the (struct hfs_btree) to insert in
 *   struct hfs_brec_key *key: pointer to the (struct hfs_brec_key) to insert
 *   void *data: pointer to the data to associate with 'key' in the b-tree
 *   unsigned int datasize: the size of the data
 * Output Variable(s):
 *   NONE
 * Returns:
 *   int: 0 on success, error code on failure
 * Preconditions:
 *   'tree' points to a "valid" (struct hfs_btree)
 *   'key' points to a "valid" (struct hfs_brec_key)
 *   'data' points to "valid" memory of length 'datasize'
 * Postconditions:
 *    XXX: finish this comment.
 */
int hfs_binsert(struct hfs_btree *tree, const struct hfs_brec_key *key,
		const void *data, __u16 datasize)
{ 
	struct hfs_brec *brec, *tmp;
	int err, reserve, retval;
	__u8 keysize;

	if (!tree || (tree->magic != HFS_BTREE_MAGIC) || !key || !data) {
		printk("hfs_binsert: invalid arguments.\n");
		return -EINVAL;
	}

restart:
	if (btree_head(tree)->bthNRecs == c_to_l(0)) {
		/* create the root bnode */
		new_root(tree, NULL, NULL);
		if (!(brec = hfs_brec_get_root(tree, 0, HFS_LOCK_RESRV))) {
			printk("hfs_binsert: failed to create root.\n");
			return -ENOSPC;
		}
		brec->record = 0;
	} else {
		err = hfs_brec_find(tree, key, &brec, HFS_BFIND_INSERT);
		if (err < 0) {
			printk("hfs_binsert: hfs_brec_find failed.\n");
			return err;
		} else if (err == 0) {
			hfs_brec_relse(&brec);
			return -EEXIST;
		}
	}

        keysize = key->KeyLen;
        datasize = ROUND(datasize);
        brec->flags = 0;
        if (bnode_freespace(brec->bnode->bn) <
                            (sizeof(HFS_WORD) + ROUND(keysize+1) + datasize)) {
                brec->flags |= HFS_BREC_OVERFLOW;
        }
	if (brec->record == 0) {
		brec->flags |= HFS_BREC_FIRST;
	}

	tmp = brec;
	if (!brec->flags) {
		hfs_brec_relse(&brec->up);
		reserve = 0;
	} else {
		reserve = 1;
		while (tmp->up) {
			tmp = tmp->up;
			++reserve;
		}
		if (tmp->bnode->bn->node != 0) {
			--reserve;
		}
		/* make certain we have enough nodes to proceed */
		if (available(tree) < reserve) {
			hfs_brec_relse(&brec);
			if (key->u.xkr.FNum == c_to_l(HFS_EXT_FID)) {
				/* don't recurse on hfs_btree_extend() */
				return -ENOSPC;
			}
			down(&tree->file.inode->i_sem);
			if (available(tree) < reserve) {
				hfs_btree_extend(tree);
			}
			up(&tree->file.inode->i_sem);
			if (available(tree) < reserve) {
				return -ENOSPC;
			} else {
				goto restart;
			}
		}
		tree->reserved += reserve;
	}

	hfs_brec_lock(tmp);
	retval = binsert(tree, &brec, key, data, keysize, datasize, reserve);
	if (!retval) {
		add_l(&btree_head(tree)->bthNRecs, 1);
		btree_dirty(tree);
	}
	return retval;
}
