/*
 *  linux/fs/hfs/bitmap.c
 *
 *  By Paul H. Hargrove
 *  Based on GPLed code Copyright (C) 1995  Michael Dreher
 *  This file may be distributed under the terms of the GPL
 *
 *  Functions to modify the volume bitmap: search/set/clear bits
 *
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/sched.h>
#include <linux/locks.h>
#include <asm/bitops.h>

/*
 * hfs_vbm_count_free()
 *
 * Description:
 *   Count the number of consecutive cleared bits in the bitmap blocks of
 *   the hfs superblock starting at bit number 'start'.  'sb' had better
 *   be locked or the indicated number of blocks may be no longer free,
 *   when this functions returns!
 * Input Variable(s):
 *   struct super_block *sb: Pointer to the hfs superblock
 *   __u16 start: bit number to start at
 * Output Variable(s):
 *   NONE
 * Returns:
 *   The number of consecutive cleared bits starting at bit 'start'
 * Preconditions:
 *   'sb' points to a "valid" (struct super_block), locked by lock_super().
 * Postconditions:
 *   NONE
 */
__u16 hfs_vbm_count_free(const struct super_block *sb, __u16 start)
{
	__u16 block_nr; /* index of the current bitmap block */
	__u16 bit_nr;   /* index of the current bit in block */
	__u16 count;    /* number of bits found so far */
	__u16 len;      /* number of bits found in this block */
	__u16 max_block;/* index of last bitmap block */
	__u16 max_bits; /* index of last bit in block */
	struct hfs_sb_info *hsb = HFS_SB(sb); /* shortcut to hfs superblock */

	/* is this a valid HFS superblock? */
	if (!sb || sb->s_magic != HFS_SUPER_MAGIC) {
		return 0;
	}

	block_nr = start / HFS_BM_BPB;
	bit_nr   = start % HFS_BM_BPB;
	max_block = (hsb->s_fs_ablocks+HFS_BM_BPB-1) / HFS_BM_BPB - 1;

	count = 0;
	while (block_nr <= max_block) {
		if (block_nr != max_block) {
			max_bits = HFS_BM_BPB;
		} else {
			max_bits = hsb->s_fs_ablocks % HFS_BM_BPB;
		}

		len = hfs_count_zero_bits(hsb->s_bitmap[block_nr]->b_data,
						max_bits, bit_nr, max_bits);
		count += len;

		/* see if we fell short of the end of this block */
		if ((len + bit_nr) < max_bits) {
			break;
		}

		++block_nr;
		bit_nr = 0;
	}
	return count;
}

/*
 * hfs_vbm_search_free()
 *
 * Description:
 *   Search for 'num_bits' consecutive cleared bits in the bitmap blocks of
 *   the hfs superblock. 'sb' had better be locked or the returned range
 *   may be no longer free, when this functions returns!
 *   XXX Currently the search starts from bit 0, but it should start with
 *   the bit number stored in 's_alloc_ptr' of the MDB.
 * Input Variable(s):
 *   struct super_block *sb: Pointer to the hfs superblock
 *   __u16 *num_bits: Pointer to the number of cleared bits
 *     to search for
 * Output Variable(s):
 *   __u16 *num_bits: The number of consecutive unset bits of the
 *     returned range. If the bitmap is fragmented, this will be less than
 *     requested and it will be zero, when the disk is full.
 * Returns:
 *   The number of the first bit of the range of cleared bits which has been
 *   found. When 'num_bits' is zero, this is invalid!
 * Preconditions:
 *   'sb' points to a "valid" (struct super_block), locked by lock_super().
 *   'num_bits' points to a variable of type (__u16), which contains
 *      the number of cleared bits to find.
 * Postconditions:
 *   'num_bits' is set to the length of the found sequence.
 */
__u16 hfs_vbm_search_free(const struct super_block *sb, __u16 *num_bits)
{
	__u16 block_nr; /* index of the current bitmap block */

	/* position and length of current portion of a run */
	__u16 cur_pos, cur_len;

	/* position and length of current complete run */
	__u16 pos=0, len=0;
	
	/* position and length of longest complete run */
	__u16 longest_pos=0, longest_len=0;

	void *bitmap; /* contents of the current bitmap block */
	__u16 max_block; /* upper limit of outer loop */
	__u16 max_bits; /* upper limit of inner loop */
	struct hfs_sb_info *hsb = HFS_SB(sb); /* shortcut to hfs superblock */

	/* is this a valid HFS superblock? */
	if (!sb || sb->s_magic != HFS_SUPER_MAGIC) {
		*num_bits = 0;
		printk("hfs_vbm_search_free: not a valid super block\n");
		return 0;
	}
	
	/* make sure we have actual work to perform */
	if (!(*num_bits)) {
		return 0;
	}

	max_block = (hsb->s_fs_ablocks+HFS_BM_BPB-1) / HFS_BM_BPB - 1;
	
	/* search all bitmap blocks */
	for (block_nr = 0; block_nr <= max_block; block_nr++) {
		bitmap = hsb->s_bitmap[block_nr]->b_data;

		if (block_nr != max_block) {
			max_bits = HFS_BM_BPB;
		} else {
			max_bits = hsb->s_fs_ablocks % HFS_BM_BPB;
		}

		cur_pos = 0;
		do {
			cur_len = hfs_count_zero_bits(bitmap, max_bits,
						  cur_pos, *num_bits);
			len += cur_len;
			if (len > longest_len) {
				longest_pos = pos;
				longest_len = len;
				if (len >= *num_bits) {
					goto search_end;
				}
			}
			if ((cur_pos + cur_len) == max_bits) {
				break; /* zeros may continue into next block */
			}

			/* find start of next run of zeros */
			cur_pos = hfs_find_next_zero_bit(bitmap, max_bits,
							 cur_pos + cur_len);
			pos = cur_pos + HFS_BM_BPB*block_nr;
			len = 0;
		} while (cur_pos < max_bits);
	}

search_end:
	*num_bits = longest_len;
	return longest_pos;
}


/*
 * hfs_set_vbm_bits()
 *
 * Description:
 *   Set the requested bits in the volume bitmap of the hfs filesystem
 * Input Variable(s):
 *   struct super_block *sb: Pointer to the hfs superblock
 *   __u16 start: The offset of the first bit
 *   __u16 count: The number of bits
 * Output Variable(s):
 *   None
 * Returns:
 *    0: no error
 *   -1: One of the bits was already set.  This is a strange
 *       error and when it happens, the filesystem must be repaired!
 *   -2: One or more of the bits are out of range of the bitmap.
 *   -3: The 's_magic' field of the superblock does not match
 * Preconditions:
 *   'sb' points to a "valid" (struct super_block), locked by lock_super().
 * Postconditions:
 *   Starting with bit number 'start', 'count' bits in the volume bitmap
 *   are set. The affected bitmap blocks are marked "dirty", the free
 *   block count of the superblock is updated and the superblock is 
 *   marked dirty.
 */
int hfs_set_vbm_bits(struct super_block *sb, __u16 start, __u16 count)
{
	__u16 block_nr; /* index of the current bitmap block */
	__u16 long_nr;  /* index of the current longword in block */
	__u16 bit_nr;   /* index of the current bit in longword */
	__u16 left = count; /* number of bits left to be set */
	__u32 *bitmap;   /* the current bitmap block's contents */
	struct hfs_sb_info *hsb = HFS_SB(sb); /* shortcut to hfs superblock */

	/* is this a valid HFS superblock? */
	if (!sb || sb->s_magic != HFS_SUPER_MAGIC) {
		return -3;
	}

	/* are all of the bits in range ? */
	if ((start + count) > hsb->s_fs_ablocks) {
		return -2;
	}

	block_nr = start / HFS_BM_BPB;
	long_nr = (start % HFS_BM_BPB) / 32;
	bit_nr = start % 32;

	bitmap = (__u32 *)(hsb->s_bitmap[block_nr]->b_data);

	/* do any partial word at the start */
	if (bit_nr != 0) {
		while ((bit_nr < 32) && left) {
			if (set_bit(HFS_BE_BIT(bit_nr), bitmap + long_nr)) {
				mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
				return -1;
			}
			++bit_nr;
			--left;
		}
		bit_nr=0;

		/* advance long_nr and check for end of this block */
		if (++long_nr > 127) {
			long_nr = 0;
			mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
			++block_nr;
			bitmap = (__u32 *)hsb->s_bitmap[block_nr]->b_data;
		}
	}

	/* do full words */
	while (left > 31) {
		if (bitmap[long_nr] != 0UL) {
			mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
			return -1;
		}
		bitmap[long_nr] = ~0UL;
		left -= 32;

		/* advance long_nr and check for end of this block */
		if (++long_nr > 127) {
			long_nr = 0;
			mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
			++block_nr;
			bitmap = (__u32 *)hsb->s_bitmap[block_nr]->b_data;
		}
	}

			
	/* do any partial word at end */
	while (left) {
		if (set_bit(HFS_BE_BIT(bit_nr), bitmap + long_nr)) {
			mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
			return -1;
		}
		++bit_nr;
		--left;
	}

	mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
	hsb->s_free_ablocks -= count;
	sb->s_dirt = 1;

	/* successfull completion */
	return 0;
}

/*
 * hfs_clear_vbm_bits()
 *
 * Description:
 *   Clear the requested bits in the volume bitmap of the hfs filesystem
 * Input Variable(s):
 *   struct super_block *sb: Pointer to the hfs superblock
 *   __u16 start: The offset of the first bit
 *   __u16 count: The number of bits
 * Output Variable(s):
 *   None
 * Returns:
 *    0: no error
 *   -1: One of the bits was already clear.  This is a strange
 *       error and when it happens, the filesystem must be repaired!
 *   -2: One or more of the bits are out of range of the bitmap.
 *   -3: The 's_magic' field of the superblock does not match
 * Preconditions:
 *   'sb' points to a "valid" (struct super_block), locked by lock_super().
 * Postconditions:
 *   Starting with bit number 'start', 'count' bits in the volume bitmap
 *   are cleared. The affected bitmap blocks are marked "dirty", the free
 *   block count of the superblock is updated and the superblock is 
 *   marked dirty.
 */
int hfs_clear_vbm_bits(struct super_block *sb, __u16 start, __u16 count)
{
	__u16 block_nr; /* index of the current bitmap block */
	__u16 long_nr;  /* index of the current longword in block */
	__u16 bit_nr;   /* index of the current bit in longword */
	__u16 left = count; /* number of bits left to be set */
	__u32 *bitmap;   /* the current bitmap block's contents */
	struct hfs_sb_info *hsb = HFS_SB(sb); /* shortcut to hfs superblock */

	/* is this a valid HFS superblock? */
	if (!sb || sb->s_magic != HFS_SUPER_MAGIC) {
		return -3;
	}

	/* are all of the bits in range ? */
	if ((start + count) > hsb->s_fs_ablocks) {
		return -2;
	}

	block_nr = start / HFS_BM_BPB;
	long_nr = (start % HFS_BM_BPB) / 32;
	bit_nr = start % 32;

	bitmap = (__u32 *)hsb->s_bitmap[block_nr]->b_data;

	/* do any partial word at the start */
	if (bit_nr != 0) {
		while ((bit_nr < 32) && left) {
			if (!clear_bit(HFS_BE_BIT(bit_nr), bitmap + long_nr)) {
				mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
				return -1;
			}
			++bit_nr;
			--left;
		}
		bit_nr=0;

		/* advance long_nr and check for end of this block */
		if (++long_nr > 127) {
			long_nr = 0;
			mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
			++block_nr;
			bitmap = (__u32 *)hsb->s_bitmap[block_nr]->b_data;
		}
	}

	/* do full words */
	while (left > 31) {
		if (bitmap[long_nr] != ~0UL) {
			mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
			return -1;
		}
		bitmap[long_nr] = 0UL;
		left -= 32;

		/* advance long_nr and check for end of this block */
		if (++long_nr > 127) {
			long_nr = 0;
			mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
			++block_nr;
			bitmap = (__u32 *)hsb->s_bitmap[block_nr]->b_data;
		}
	}

			
	/* do any partial word at end */
	while (left) {
		if (!clear_bit(HFS_BE_BIT(bit_nr), bitmap + long_nr)) {
			mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
			return -1;
		}
		++bit_nr;
		--left;
	}

	mark_buffer_dirty(hsb->s_bitmap[block_nr], 1);
	hsb->s_free_ablocks += count;
	sb->s_dirt = 1;

	/* successfull completion */
	return 0;
}
