/*
 *  linux/fs/hfs/bitops.c
 *
 *  Copyright (C) 1996   Paul H. Hargrove
 *  This file may be distributed under the terms of the GNU Public License
 *
 *  This file contains functions to handle bitmaps in big-endian order
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/kernel.h>
#include <asm/bitops.h>

/*
 * hfs_find_first_zero_bit()
 *
 * Description:
 *   Given the address of a block of memory find the first (by
 *   big-endian bit-ordering) zero bit in the first 'size' bits.
 */
__u32 hfs_find_first_zero_bit(const void * addr, __u32 size)
{
	const __u32 *start = (const __u32 *)addr;
	const __u32 *end   = start + ((size+31) >> 5);
	const __u32 *curr  = start;
	int bit = 0;
	
	while (curr < end) {
		if (*curr == ~0UL) {
			++curr;
		} else {
			while (test_bit(HFS_BE_BIT(bit), curr)) {
			       ++bit;
			}
			break;
		}
	}
	return ((curr-start)<<5) | bit;
}

/*
 * hfs_find_next_zero_bit()
 *
 * Description:
 *   Given the address of a block of memory find the next (by
 *   big-endian bit-ordering) zero bit in the first 'size' bits.
 */
__u32 hfs_find_next_zero_bit(const void * addr, __u32 size, __u32 offset)
{
	const __u32 *start = (const __u32 *)addr;
	const __u32 *end   = start + ((size+31) >> 5);
	const __u32 *curr  = start + (offset>>5);
	int bit = offset % 32;
	
	if (size < offset) {
		return size;
	}

	/* finish up first partial long */
	if (bit != 0) {
		do {
			if (!test_bit(HFS_BE_BIT(bit), curr)) {
				goto done;
			}
			++bit;
		} while (bit < 32);
		bit = 0;
		++curr;
	}
	
	/* skip fully set longs */
	while (*curr == ~0UL) {
		if (curr == end) {
			goto done;
		} else {
			++curr;
		}
	}

	/* check final partial long */
	do {
		if (!test_bit(HFS_BE_BIT(bit), curr)) {
			goto done;
		}
		++bit;
	} while (bit<32);

done:
	return ((curr-start)<<5) | bit;
}

/*
 * hfs_count_zero_bits()
 *
 * Description:
 *  Given a block of memory, its length in bits, and a starting bit number,
 *  determine the number of consecutive zero bits (in big-endian ordering)
 *  starting at that location, upto 'max'.
 */
__u32 hfs_count_zero_bits(const void *addr, __u32 size, __u32 offset, __u32 max)
{
	const __u32 *curr = (const __u32 *)addr + (offset>>5);
	const __u32 *end = (const __u32 *)addr + ((size+31)>>5);
	int bit = offset % 32;
	__u32 count = 0;

	/* fix up max to avoid running off end */
	if (max > (size - offset)) {
		max = size - offset;
	}

	/* finish up first partial long */
	if (bit != 0) {
		do {
			if (test_bit(HFS_BE_BIT(bit), curr)) {
				goto done;
			}
			++count;
			++bit;
		} while (bit<32);
		++curr;
	}
	
	/* eat up any full words of zeros */
	if (size > (offset + max)) {
		size = offset + max;
	}
	while (*curr == 0UL) {
		if (curr == end) {
			goto done;
		} else {
			++curr;
			count += 32;
		}
	}
	
	/* count bits in the last word */
	bit = 0;
	do {
		if (test_bit(HFS_BE_BIT(bit), curr)) {
			goto done;
		}
		++count;
		++bit;
	} while (bit<32);

done:
        if (count > max) {
	        count = max;
	}
	return count;
}
