/*
 * linux/fs/hfs/bnode.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the code to access nodes in the B-tree structure.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 *
 * The code in this file initilizes some structures by calling
 * memset(&foo, 0, sizeof(foo)).  This produces the desired behavior
 * only due to the non-ANSI assumption that the machine representation
 * of NULL is all zeros.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include "hfs_fs.h"
#include "hfs_btree.h"

/*================ File-local variables ================*/
 
/* debugging statistics */
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
static int hit = 0;
static int miss = 0;
int bnode_count = 0;
#endif

/*================ File-local functions ================*/

/*
 * hfs_bnode_delete()
 *
 * Description:
 *   This function is called to remove a bnode from the cache and
 *   release its resources.
 * Input Variable(s):
 *   struct hfs_bnode *bn: Pointer to the (struct hfs_bnode) to be
 *   removed from the cache.
 * Output Variable(s):
 *   NONE
 * Returns:
 *   void
 * Preconditions:
 *   'bn' points to a "valid" (struct hfs_bnode).
 * Postconditions:
 *   The node 'bn' is removed from the cache, its memory freed and its
 *   buffer_head (if any) released.
 */
static void hfs_bnode_delete(struct hfs_bnode *bn)
{
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	--bnode_count;
#endif
	/* join neighbors */
	if (bn->next) {
		bn->next->prev = bn->prev;
	}
	if (bn->prev) {
		bn->prev->next = bn->next;
	}
	/* fix cache slot if necessary */
	if (bhash(bn->tree, bn->node) == bn) {
		bhash(bn->tree, bn->node) = bn->next;
	}
	/* release resources */
	brelse(bn->bh); /* safe: brelse() checks for NULL argument */
	KFREE_S(bn, sizeof(struct hfs_bnode));
}


/*
 * hfs_bnode_read()
 *
 * Description: 
 *   This function creates a (struct hfs_bnode) and, if appropriate,
 *   inserts it in the cache.
 * Input Variable(s):
 *   struct hfs_btree *tree: pointer to the (struct hfs_btree)
 *    containing the desired node
 *   int node: the number of the desired node.
 *   int sticky: the value to assign to the 'sticky' field.
 * Output Variable(s):
 *   NONE
 * Returns:
 *   (struct hfs_bnode *) pointing to the newly created bnode or NULL.
 * Preconditions:
 *   'tree' points to a "valid" (struct hfs_btree).
 *   'node' is an existing node number in the B-tree.
 * Postconditions:
 *   If KMALLOC() fails then NULL is returned, otherwise the following
 *   are true of the returned (struct hfs_bnode *):
 *    The 'magic' field is set to indicate a valid (struct hfs_bnode). 
 *    The 'sticky', 'tree' and 'node' fields are initialized to the
 *    values of the of the corresponding arguments.
 *    If the 'sticky' argument is zero then the fields 'prev' and
 *    'next' are initialized by inserting the (struct hfs_bnode) in the
 *    linked list of the appropriate cache slot; otherwise they are
 *    initialized to NULL.
 *    The data is read from disk (or buffer cache) and the 'bh' field
 *    points to the buffer_head for that data.
 *    If no other processes tried to access this node while this
 *    process was waiting on disk I/O (if necessary) then the
 *    remaining fields are zero ('count', 'resrv', 'lock') or NULL
 *    ('wqueue', 'rqueue') corresponding to no accesses.
 *    If there were access attempts durring I/O then they were blocked
 *    until the I/O was complete, and the fields 'count', 'resrv',
 *    'lock', 'wqueue' and 'rqueue' reflect the results of unblocking
 *    those processes when the I/O was completed.
 */
struct hfs_bnode * hfs_bnode_read(struct hfs_btree *tree, int node, int sticky)
{
	struct hfs_bnode *bnode;

	if ((bnode=KMALLOC(sizeof(struct hfs_bnode), GFP_KERNEL))) {
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
		++bnode_count;
#endif
		/* Initialize the structure */
		memset(bnode, 0, sizeof(struct hfs_bnode));
		bnode->magic = HFS_BNODE_MAGIC;
		bnode->tree = tree;
		bnode->node = node;
		bnode->sticky = sticky;

		if (sticky == HFS_NOT_STICKY) {
			/* Insert it in the cache if appropriate */
			if ((bnode->next = bhash(tree, node))) {
				bnode->next->prev = bnode;
			}
			bhash(tree, node) = bnode;

			/* Make the bnode look like it is being
			   modified so other processes will wait for
			   the I/O to complete */
			bnode->count = bnode->resrv = bnode->lock = 1;
		}

		/* Read in the node, possibly causing a schedule()
		   call.  If the I/O fails then emit a warning.	 Each
		   process that was waiting on the bnode (including
		   the current one) will notice the failure and
		   hfs_bnode_relse() the node.  The last hfs_bnode_relse()
		   will call hfs_bnode_delete() and discard the bnode.  */


		if (!(bnode->bh = bread(tree->file.dev,
					hfs_block_map(&tree->file, node, 0),
					HFS_SECTOR_SIZE))) {
			printk("hfs_bnode_read: unable to read node from dev"
			       " 0x%04x\n", tree->file.dev);
		}
		if (sticky==HFS_NOT_STICKY) {
			/* Undo our fakery with the lock state and
			   wake_up() anyone who we managed to trick */
			--bnode->count;
			bnode->resrv = bnode->lock = 0;
			wake_up(&bnode->rqueue);
		} else if (!bnode->bh) {
			/* If the read fails on the root or header
			   node then free the bnode and return NULL */
			KFREE_S(bnode, sizeof(struct hfs_bnode));
			bnode = NULL;
		}
	}
	return bnode;
}

/*
 * hfs_bnode_lock()
 *
 * Description:
 *   This function does the locking of a bnode.
 * Input Variable(s):
 *   struct hfs_bnode *bn: pointer to the (struct hfs_bnode) to lock
 *   int lock_type: the type of lock desired
 * Output Variable(s):
 *   NONE
 * Returns:
 *   void
 * Preconditions:
 *   'bn' points to a "valid" (struct hfs_bnode).
 *   'lock_type' is a valid hfs_lock_t
 * Postconditions:
 *   The 'count' field of 'bn' is incremented by one.  If 'lock_type'
 *   is HFS_LOCK_RESRV the 'resrv' field is also incremented.
 */
void hfs_bnode_lock(struct hfs_bnode_ref *bnr, int lock_type)
{
	struct hfs_bnode *bn = bnr->bn;

	if (lock_type == bnr->lock_type) {
		return;
	}

	if (bnr->lock_type == HFS_LOCK_NONE) {
		++bn->count;
	} else if (lock_type == HFS_LOCK_NONE) {
		--bn->count;
	}

	switch (lock_type) {
	default:
		goto bail;
		break;

	case HFS_LOCK_READ:
		/* We may not obtain read access if any process is
		   currently modifying or waiting to modify this node.
		   If we can't obtain access we wait on the rqueue
		   wait queue to be woken up by the modifying process
		   when it relinquishes its lock. */
		switch (bnr->lock_type) {
		default:
			goto bail;
			break;

		case HFS_LOCK_WRITE:
			if (bn->bh) {
				mark_buffer_dirty(bn->bh, 1);
			}
			bn->lock = 0;
		case HFS_LOCK_RESRV:
			bn->resrv = 0;
			wake_up(&bn->rqueue);
			break;

		case HFS_LOCK_NONE:
			while (bn->lock || bn->wqueue) {
				sleep_on(&bn->rqueue);
			}
			break;
		}
		break;
			
	case HFS_LOCK_RESRV:
		/* We may not obtain a reservation (read access with
		   an option to write later), if any process currently
		   holds a reservation on this node.  That includes
		   any process which is currently modifying this node.
		   If we can't obtain access, then we wait on the
		   rqueue wait queue to e woken up by the
		   reservation-holder when it calls hfs_bnode_relse. */
		switch (bnr->lock_type) {
		default:
			goto bail;
			break;

		case HFS_LOCK_NONE:
			while (bn->resrv) {
				sleep_on(&bn->rqueue);
			}
			bn->resrv = 1;
			break;

		case HFS_LOCK_WRITE:
			if (bn->bh) {
				mark_buffer_dirty(bn->bh, 1);
			}
			bn->lock = 0;
			wake_up(&bn->rqueue);
			break;
		}
		break;
		
	case HFS_LOCK_WRITE:
		switch (bnr->lock_type) {
		default:
			goto bail;
			break;

		case HFS_LOCK_NONE:
			while (bn->resrv) {
				sleep_on(&bn->rqueue);
			}
			bn->resrv = 1;
		case HFS_LOCK_RESRV:
			while (bn->count > 1) {
				sleep_on(&bn->wqueue);
			}
			bn->lock = 1;
			break;
		}
		break;

	case HFS_LOCK_BLIND:
		switch (bnr->lock_type) {
		default:
			goto bail;
			break;

		case HFS_LOCK_NONE:
			break;
		}
		break;

	case HFS_LOCK_NONE:
		switch (bnr->lock_type) {
		default:
			goto bail;
			break;

		case HFS_LOCK_READ:
			/* This process was reading this node.	If
			   there now exactly one process using the
			   node then wake_up() a (potentially
			   nonexistent) waiting process.  Note that I
			   refer to "a" process since the reservation
			   system ensures that only one process can
			   get itself on the wait queue.  */
			if (bn->count == 1) {
				wake_up(&bn->wqueue);
			}
			break;

		case HFS_LOCK_WRITE:
			/* This process was modifying this node.
			   Unlock the node and fall-through to the
			   HFS_LOCK_RESRV case, since a 'reservation'
			   is a prerequisite for HFS_LOCK_WRITE.  */
			if (bn->bh) {
				mark_buffer_dirty(bn->bh, 1);
			}
			bn->lock = 0;
		case HFS_LOCK_RESRV:
			/* This process had placed a 'reservation' on
			   this node, indicating an intention to
			   possibly modify the node.  We can get to
			   this spot directly (if the 'reservation'
			   not converted to a HFS_LOCK_WRITE), or by
			   falling through from the above case if the
			   reservation was converted.
			   Since HFS_LOCK_RESRV and HFS_LOCK_WRITE
			   both block processes that want access
			   (HFS_LOCK_RESRV blocks other processes that
			   want reservations but allow HFS_LOCK_READ
			   accesses, while HFS_LOCK_WRITE must have
			   exclusive access and thus blocks both
			   types) we wake_up() any processes that
			   might be waiting for access.	 If multiple
			   processes are waiting for a reservation
			   then the magic of process scheduling will
			   settle the dispute. */
			bn->resrv = 0;
			wake_up(&bn->rqueue);
			break;

		case HFS_LOCK_BLIND:
			break;
		}
		break;
	}
	bnr->lock_type = lock_type;
	return;

bail:
	printk("hfs_bnode_lock: invalid lock change: %d->%d.\n",
		bnr->lock_type, lock_type);
	return;
}

/*
 * hfs_bnode_relse()
 *
 * Description:
 *   This function is called when a process is done using a bnode.  If
 *   the proper conditions are met then we call hfs_bnode_delete() to remove
 *   it from the cache.	 If it is not deleted then we update its state
 *   to reflect one less process using it.
 * Input Variable(s):
 *   struct hfs_bnode *bn: pointer to the (struct hfs_bnode) to release.
 *   int lock_type: The type of lock held by the process releasing this node.
 * Output Variable(s):
 *   NONE
 * Returns:
 *   void
 * Preconditions:
 *   'bn' is NULL or points to a "valid" (struct hfs_bnode).
 * Postconditions:
 *   If 'bn' meets the appropriate conditions (see below) then it is
 *   kept in the cache and all fields are set to consistent values
 *   which reflect one less process using the node than upon entry.
 *   If 'bn' does not meet the conditions then it is deleted (see
 *   hfs_bnode_delete() for postconditions).
 *   In either case, if 'lock_type' is HFS_LOCK_WRITE
 *   then the corresponding buffer_head is dirtied.
 */
void hfs_bnode_relse(struct hfs_bnode_ref *bnr)
{
	struct hfs_bnode *bn;

	if (!bnr) {
		return;
	}
	bn = bnr->bn;

	/* dirty buffer if we (might) have modified it */
	if (bn->bh && (bnr->lock_type == HFS_LOCK_WRITE)) {
		mark_buffer_dirty(bn->bh, 1);
	}

	/* We update the lock state of the node if it is still in use
	   or if it is "sticky" (such as the B-tree head and root).
	   Otherwise we just delete it.	 */
	if ((bn->count > 1) || (bn->sticky != HFS_NOT_STICKY)) {
		hfs_bnode_lock(bnr, HFS_LOCK_NONE);
	} else {
		hfs_bnode_delete(bn);
	}
	KFREE_S(bnr, sizeof(*bnr));
}

/*
 * hfs_bnode_find()
 *
 * Description:
 *   This function is called with 'lock_type' equal HFS_LOCK_READ, 
 *   HFS_LOCK_RESERV or HFSSS_LOCK_BLIND to obtain a bnode.  The cache
 *   is searched for the node.  If it not found there it is added to
 *   the cache by hfs_bnode_read().  There are two special cases node=0
 *   (the header node) and node='tree'->bthRoot (the root node), in
 *   which the nodes are obtained from fields of 'tree' without
 *   consulting or modifying the cache.
 * Input Variable(s):
 *   struct hfs_tree *tree: pointer to the (struct hfs_btree) from
 *    which to get a node.
 *   int node: the node number to get from 'tree'.
 *   int lock_type: The kind of access (HFS_LOCK_READ, HFS_LOCK_BLIND, or
 *    HFS_LOCK_RESRV) to obtain to the node
 * Output Variable(s):
 *   NONE
 * Returns:
 *   (struct hfs_bnode *) The requested node or NULL.
 * Preconditions:
 *   'tree' points to a "valid" (struct hfs_btree).
 * Postconditions:
 *   If 'node' refers to a valid node in 'tree' and 'lock_type' has
 *   one of the values listed above and no I/O errors occur then the
 *   value returned points to a valid (struct hfs_bnode) corresponding
 *   to the requested node with the requested access type.  The node
 *   is also added to the cache if not previously present and not the
 *   root or header.
 *   If the conditions given above are not met, NULL is returned.
 */
struct hfs_bnode_ref * hfs_bnode_find(struct hfs_btree *tree,
				      __u32 node, int lock_type)
{
	struct hfs_bnode *bn;
	struct hfs_bnode_ref *bnr;

	if (!(bnr=KMALLOC(sizeof(*bnr), GFP_KERNEL))) {
		printk("hfs_bnode_find: out of memory\n");
		return NULL;
	}
	bnr->lock_type = HFS_LOCK_NONE;

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	printk("hfs_bnode_find: %c %d:%d ", ((lock_type==HFS_LOCK_READ)?'R':'V'),
	       HFS_IFID(tree->file.inode->i_ino), node);
#endif
	/* check special cases and then look for the node in the cache. */
	if (!node) {
		bn = tree->head;
	} else if (node == tree->bthRoot) {
		bn = tree->root;
	} else {
		bn = bhash(tree, node);
		while (bn && (bn->magic == HFS_BNODE_MAGIC) &&
		       (bn->node != node)) {
			bn = bn->next;
		}
	}

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	if (bn) {
		++hit;
	} else {
		++miss;
	}
	printk("%c (%d%%) ", ((bn)?'+':'-'), ((hit*100)/(hit+miss)));
#endif

	/* If we are not dealing with a special case (header or root)
	   and the node is not in the cache, we call hfs_bnode_read() to
	   get it. */
	if (bn || (bn=hfs_bnode_read(tree, node, HFS_NOT_STICKY))) {
		/* check validity */
		if (bn->magic != HFS_BNODE_MAGIC) {
			/* If we find a corrupt bnode then we return
			   NULL.  However, we don't try to remove it
			   from the cache or release its resources
			   since we have no idea what kind of trouble
			   we could get into that way. */
			printk("hfs_bnode_find: bnode cache is corrupt.\n");
			bn = NULL;
		} else {
			/* Wait our turn */
			bnr->bn = bn;
			hfs_bnode_lock(bnr, lock_type);
			/* Check for failure to read the node from disk */
			if (!bn->bh) {
				hfs_bnode_relse(bnr);
				bn = NULL;
			}
		}
	}

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	if (!bn) {
		printk("failed\n");
	} else {
		printk("use %d(%d) lvl %d [%d]\n", bn->count, bn->bh->b_count,
		       bnode_head(bn)->ndNHeight, bnode_count);
	}
#endif
	return bnr;
}
