/*
 * linux/fs/hfs/brec.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the code to access records in a btree.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 *
 * The code in this file initilizes some structures by calling
 * memset(&foo, 0, sizeof(foo)).  This produces the desired behavior
 * only due to the non-ANSI assumption that the machine representation
 * of NULL is all zeros.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/fs.h>
#include "hfs_fs.h"
#include "hfs_btree.h"

/*================ File-local functions ================*/

/* 
 * dump_brec()
 *
 * prints out a brec for debugging purposes.
 */
static void dump_brec(struct hfs_brec *brec) {
	int i;
	printk("dump_brec: begin\n");
	while (brec) {
		printk("dump_brec: record %d of node %d\n",
			brec->record, brec->bnode->bn->node);
		if (brec->key) {
			printk("dump_brec: key=");
			for (i=0; i<brec->key->KeyLen; ++i) {
				printk("%02x ", ((__u8 *)brec->key)[i]);
			}
			printk("\n");
		}
		brec = brec->up;
	}
	printk("dump_brec: end\n");
}

/*
 * brec_alloc()
 *
 * Description:
 *   This function allocates an hfs_brec structure.  This is the
 *   structure used to comunicate with the catalog and extents codes
 *   about records in b-trees.
 * Input Variable(s):
 *   NONE
 * Output Variable(s):
 *   NONE
 * Returns:
 *   (struct hfs_brec *): NULL or pointer to an empty hfs_brec
 * Preconditions:
 *   NONE
 * Postconditions:
 *   The returned pointer is NULL if KMALLOC() returned NULL.
 *   Otherwise the return value points to a (struct hfs_brec) with all
 *   fields set to zero or NULL (as appropriate), except for the 'magic'
 *   field which is initialized to indicate a valid (struct hfs_brec).
 */
static inline struct hfs_brec *brec_alloc(void)
{
	struct hfs_brec *brec;
#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
	extern int hfs_bcount;
	++hfs_bcount;
#endif
	if ((brec=KMALLOC(sizeof(struct hfs_brec), GFP_KERNEL))) {
		memset(brec, 0, sizeof(struct hfs_brec));
		brec->magic = HFS_BREC_MAGIC;
		brec->flags = HFS_BREC_UNINITIALIZED;
	}
	return brec;
}

/*
 * hfs_brec_lock()
 *
 * Description:
 *   This function obtains HFS_LOCK_WRITE access to the bnode
 *   containing this hfs_brec.	All decendents in the path from this
 *   record to the leaf are given HFS_LOCK_WRITE access and all
 *   ancestors in the path from the root to here are released.	Called
 *   only by hfs_brec_find().
 * Input Variable(s):
 *   struct hfs_brec *brec: pointer to the record in the B-tree to
 *    obtain HFS_LOCK_WRITE access to.
 * Output Variable(s):
 *   NONE
 * Returns:
 *   void
 * Preconditions:
 *   'brec' points to a "valid" (struct hfs_brec)
 * Postconditions: 
 *   If the (struct hfs_brec) is initially held with HFS_LOCK_RESRV
 *   access then upon return all records from the root upto but not
 *   including 'brec' are released, and 'brec' and all records along
 *   the path to the leaf upto (but excuding) the first record not
 *   held with HFS_LOCK_RESRV access are held with HFS_LOCK_WRITE
 *   access.  If there are no records along the path to the leaf that
 *   are not held with HFS_LOCK_RESRV access (the normal case) then
 *   all records along the path (including the leaf) are held with
 *   HFS_LOCK_WRITE access upon return.	 Note that if there is a
 *   record in the path not held with HFS_LOCK_RESRV access then an
 *   error message is printed.	If 'brec' is not held with
 *   HFS_LOCK_RESRV access then no changes to access types are made,
 *   no records are released and an error message is printed.
 */
void hfs_brec_lock(struct hfs_brec *brec) 
{
	if (brec->bnode->lock_type == HFS_LOCK_RESRV) {
		hfs_brec_relse(&brec->up);
		do {
			hfs_bnode_lock(brec->bnode, HFS_LOCK_WRITE);
		} while ((brec = brec->down) &&
			 (brec->bnode->lock_type == HFS_LOCK_RESRV));
		if (brec) {
			printk("hfs_hfs_brec_lock: unreseved record in path "
			       "to leaf\n");
		}
	} else
		printk("hfs_hfs_brec_lock: record not reserved\n");

}

/*
 * first()
 *
 * returns HFS_BREC_FIRST if brec->record == 1, 0 otherwise
 */
static inline int first(const struct hfs_brec *brec)
{
	return (brec->record == 1) ? HFS_BREC_FIRST : 0;
}

/*
 * overflow()
 *
 * return HFS_BREC_OVERFLOW if the node has no room for an 
 * additional pointer record, 0 otherwise.
 */
static inline int overflow(const struct hfs_btree *tree,
			   const struct hfs_bnode *bnode)
{
	/* there is some algebra involved in getting this form */
	return ((HFS_SECTOR_SIZE - bnode_end(bnode) -
                 (2+from_w(bnode_head(bnode)->ndNRecs))*sizeof(HFS_WORD)) <
	        (ROUND(tree->bthKeyLen+1) + sizeof(HFS_LWORD))) ?
		HFS_BREC_OVERFLOW : 0;
}

/*
 * underflow()
 *
 * return HFS_BREC_UNDERFLOW if the node will be less that 1/2 full
 * upon removal of a pointer record, 0 otherwise.
 */
static inline int underflow(const struct hfs_btree *tree,
			    const struct hfs_bnode *bnode)
{
	/* there is some algebra involved in getting this form */
	return ((from_w(bnode_head(bnode)->ndNRecs)*sizeof(HFS_WORD) +
		 bnode_offset(bnode, from_w(bnode_head(bnode)->ndNRecs))) >
	        (HFS_SECTOR_SIZE + sizeof(struct NodeDescriptor))/2) ?
		HFS_BREC_UNDERFLOW : 0;
}

/*
 * hfs_brec_get_root()
 *
 * Description:
 *   Obtain access to the root node of a B-tree.  Note that this first
 *   must obtain access to the header node.
 * Input Variable(s):
 *   struct hfs_tree *tree: pointer to the (struct hfs_btree) of which
 *    we want the root.
 *   int flags: bitwise or of flags passed to hfs_brec_find()
 * Output Variable(s):
 *   NONE
 * Returns:
 *   pointer to a partially initialized (struct hfs_brec) for the root node.
 * Preconditions:
 *   'tree' points to a "valid" (struct hfs_btree) with "valid" 'root' field
 * Postconditions:
 *   If the two calls to brec_alloc() succeed then the return value
 *   points to a (struct hfs_brec) which corresponds to the root node
 *   of 'tree'.	 If the HFS_BFIND_RDONLY bit of 'flags' is set then both
 *   the root and header nodes are obtained with HFS_LOCK_READ access,
 *   otherwise they are obtained HFS_LOCK_RESRV access.	 The fields
 *   'record', 'key' and 'data' are the only uninitialized fields.
 *   Thus the returned (struct hfs_brec) refers only to a bnode and
 *   not to a particular record in it.  If 'keep_mask' is non-zero
 *   then the 'up' field of the returned (struct hfs_brec) will indicate
 *   the header node if it is needed to complete the appropriate
 *   operation (insertion or deletion), otherwise the header node is
 *   released before this function returns and 'up' is NULL.
 */
struct hfs_brec *hfs_brec_get_root(struct hfs_btree *tree,
				   int keep_mask, int lock_type)
{
	struct hfs_brec *head, *root=NULL;
	
	if ((head = brec_alloc()) && (root = brec_alloc())) {
		head->bnode = hfs_bnode_find(tree, 0, lock_type);
		root->bnode = hfs_bnode_find(tree, tree->bthRoot, lock_type);
	}
	
	if (root) {
		/* HFS_BREC_FIRST not applicable for root */
		root->flags = overflow(tree, root->bnode->bn) |
			      underflow(tree, root->bnode->bn);
	}

	if (root && (root->flags & keep_mask)) {
		root->up = head;
		head->down = root;
	} else {
		hfs_brec_relse(&head);
	}
	return root;
}

/*
 * hfs_brec_get_child()
 *
 * Description:
 *   Obtain access to a child of an internal node in a B-tree.
 * Input Variable(s):
 *   struct hfs_brec *brec: pointer to the (struct hfs_brec) of which
 *    to obtain a child.
 *   int flags: bitwise OR of flags passed to hfs_brec_find().
 * Output Variable(s):
 *   NONE
 * Returns:
 *   pointer to a partially initialized (struct hfs_brec) for the
 *    desired node, or NULL.
 * Preconditions:
 *   'brec' points to a "valid" (struct hfs_brec) corresponding to a
 *   record in a bnode of type ndIndxNode and the record field
 *   indicates the (non-deleted) index record for the desired child.
 * Postconditions:
 *   If the call to hfs_bnode_find() (and to brec_alloc() if appropriate)
 *   succeeds then the return value
 *   points to a (struct hfs_brec) corresponding to the requested
 *   bnode, otherwise 'brec' is released and NULL is returned.	The
 *   child is obtaind with the same 'lock_type' field as 'brec'.
 *   'record', 'key' and 'data' are the only uninitialized fields.
 *   Thus the returned (struct hfs_brec) refers only to a bnode
 *   not to a particular record in it.  If 'keep_mask' is non-zero
 *   then the 'up' field of the returned (struct hfs_brec) will indicate
 *   the ancestors that may be needed to complete the appropriate
 *   operation (insertion or deletion), otherwise the header node is
 *   released before this function returns and 'up' is NULL.
 */
struct hfs_brec *hfs_brec_get_child(struct hfs_btree *tree,
				    struct hfs_brec *brec, int keep_mask)
{
	int lock_type = brec->bnode->lock_type;
	struct hfs_bnode_ref *tmp, *bn;

	if ((bn=hfs_bnode_find(tree, from_l(*brec->data.child), lock_type))) {

		brec->flags = first(brec) |
		  	      overflow(tree, brec->bnode->bn) |
		  	      underflow(tree, brec->bnode->bn);

		if (keep_mask) {
			if ((brec->down = brec_alloc())) {
				if (!(keep_mask & brec->flags)) {
					hfs_brec_relse(&brec->up);
				} else if ((brec->up) &&
					   !(brec->flags & brec->up->flags)) {
					hfs_brec_relse(&brec->up->up);
				}
				brec->down->up = brec;
				brec = brec->down;
				brec->bnode = bn;
			} else {
				hfs_brec_relse(&brec);
				hfs_bnode_relse(bn);
			}
		} else {
			tmp = brec->bnode;
			brec->bnode = bn;
			hfs_bnode_relse(tmp);
		}
	} else
		hfs_brec_relse(&brec);

	return brec;
}

/*
 * hfs_brec_find()
 *
 * Description:
 *   This function has sole responsibility for locating existing
 *   records in a B-tree.  Given a B-tree and a key it locates the
 *   "greatest" record "less than or equal to" the given key.  The
 *   exact behavior is determined by the bits of the flags variable as
 *   follows:
 *     HFS_BFIND_RDONLY: request HFS_LOCK_READ access to the accessed
 *	nodes, otherwise request HFS_LOCK_RESRV access
 *     HFS_BFIND_EXACT: only accept an exact match, otherwise take the
 *	"largest" record less than 'target' as a "match"
 *     HFS_BFIND_LOCK: request HFS_LOCK_WRITE access to the node containing
 *	the "matching" record when it is located
 *     HFS_BFIND_INSERT: keep access to internal nodes only so long as
 *	insertion of a record following the "matching" record could
 *	potentially modify them.
 *     HFS_BFIND_DELETE: keep access to internal nodes only so long as
 *	deletion of the "matching" record could potentially modify them.
 * Input Variable(s):
 *   struct hfs_btree *tree: pointer to the (struct hfs_btree) to search
 *   struct hfs_brec_key *target: pointer to the (strucy hfs_brec_key)
 *    to search for
 *   struct hfs_brec **brec: the location to store a pointer to the
 *    located record or NULL
 *   int flags: bitwise OR of flags which determine the function's behavior
 * Output Variable(s):
 *   struct hfs_brec **brec: '*brec' is NULL on failure or a pointer
 *    to the located record on success.
 * Returns:
 *   int: 0 on success or an error code on failure:
 *     -EINVAL: one of the input variables was NULL.
 *     -ENOMEM: unable to allocate memory for the header and root nodes
 *     -ENOENT: tree is valid but empty or no "matching" record was located.
 *	 If the HFS_BFIND_EXACT bit of 'flags' is not set then the case of no
 *	 matching record will give a '*brec' with a 'record' field of zero
 *	 rather thn returning this error.
 *     -EIO: an assertion about the structure of a vaild B-tree failed
 *	 indicating corruption of either the B-tree structure on the
 *	 disk or one of the in-core structures representing the B-tree.
 *	 (This could also be returned if a KMALLOC() call failed in a
 *	 subordinate routine that is intended to get the data from the
 *	 disk of the buffer cache.)
 * Preconditions:
 *   'tree' is NULL or points to a "valid" (struct hfs_btree)
 *   'target' is NULL or points to a "valid" (struct hfs_brec_key)
 * Postconditions:
 *   If 'brec', 'tree' or 'target' is NULL then -EINVAL is returned
 *   and nothing is changed.
 *   If 'brec', 'tree' and 'target' are non-NULL but the tree is empty
 *   then '*brec' is NULL and -ENOENT is returned.
 *   If 'brec', 'tree' and 'target' are non-NULL but the call to
 *   hfs_brec_get_root() fails then '*brec' is NULL and -ENOMEM is
 *   returned.
 *   If 'brec', 'tree' and 'target' are non-NULL and the tree is
 *   non-empty then the tree is searched as follows:
 *    If any call to hfs_brec_get_child() fails or returns a node that is
 *     neither an index node nor a leaf node then '*brec' is NULL and
 *     -EIO is returned to indicate that the B-tree or buffer-cache
 *     are corrupted.
 *    If every record in the tree is "greater than" the given key
 *     and the HFS_BFIND_EXACT bit of 'flags' is set then '*brec' is NULL
 *     and -ENOENT is returned.
 *    If every record in the tree is "greater than" the given key
 *     and the HFS_BFIND_EXACT bit of 'flags' is clear then '*brec' refers
 *     to the first leaf node in the tree and has a 'record' field of
 *     zero, and zero is returned.
 *    If a "matching" record is located with key "equal to" 'target'
 *     then the return value is 0 and '*brec' points to the record.
 *    If a "matching" record is located with key "greater than" 'target'
 *     then the behavior is determined as follows:
 *	If the HFS_BFIND_EXACT bit of 'flags' is not set then a positive
 *	 integer is returned and '*brec' is a pointer to the located brec.
 *	If the HFS_BFIND_EXACT bit of 'flags' is set then -ENOENT is returned
 *	 and '*brec' is NULL.
 *    If the return value is 0 and the HFS_BFIND_LOCK bit of 'flags' is set
 *    then hfs_brec_lock() is called on '*brec' before returning.
 */
int hfs_brec_find(struct hfs_btree *tree,
		  const struct hfs_brec_key *target,
		  struct hfs_brec **brec, int flags)
{
	struct hfs_bnode *bn;
	int result, ntype, keep_mask;

	/* check for invalid arguments */
	if (!brec || !tree || (tree->magic != HFS_BTREE_MAGIC) || !target) {
		return -EINVAL;
	}

	/* check for empty tree */
	if (!tree->root || (btree_head(tree)->bthNRecs == c_to_l(0))) {
		return -ENOENT;
	}

	/* build keep_mask from flags */
	if (flags & HFS_BFIND_INSERT) {
		keep_mask = HFS_BREC_FIRST | HFS_BREC_OVERFLOW;
	} else if (flags & HFS_BFIND_DELETE) {
		keep_mask = HFS_BREC_FIRST | HFS_BREC_UNDERFLOW;
	} else {
		keep_mask = 0;
	}

	/* start search at root of tree */
	if (!(*brec = hfs_brec_get_root(tree, keep_mask,
					(flags & HFS_BFIND_RDONLY) ?
					HFS_LOCK_READ : HFS_LOCK_RESRV))) {
		return -ENOMEM;
	}

	/* traverse the tree */
	do {
		bn = (*brec)->bnode->bn;

		(*brec)->record = from_w(bnode_head(bn)->ndNRecs);
		if (!(*brec)->record) {
			printk("hfs_bfind: empty bnode\n");
			dump_brec(*brec);
			hfs_brec_relse(brec);
			return -EIO;
		}

		/* reverse linear search yielding largest key "less
		   than or equal to" 'target'.
		   It is questionable whether a binary search would be
		   significantly faster */
		do {
			(*brec)->key = brec_key(*brec);
			if (!(*brec)->key->KeyLen) {
				printk("hfs_bfind: empty key\n");
				dump_brec(*brec);
				hfs_brec_relse(brec);
				return -EIO;
			}
			result = (tree->compare)(target, (*brec)->key);
		} while ((result<0) && (--(*brec)->record));

		ntype = bnode_head(bn)->ndType;

		/* see if all keys > target */
		if (!(*brec)->record) {
			if (bnode_head(bn)->ndBLink) {
				/* at a node other than the left-most at a
				   given level it means the parent had an
				   incorrect key for this child */
				hfs_brec_relse(brec);
				printk("hfs_bfind: corrupted b-tree %ld.\n",
				       HFS_IFID(tree->file.inode->i_ino));
				return -EIO;
			}
			if (flags & HFS_BFIND_EXACT) {
				/* we're not going to find it */
				hfs_brec_relse(brec);
				return -ENOENT;
			}
			if (ntype == ndIndxNode) {
				/* since we are at the left-most node at
				   the current level and looking for the
				   predecessor of 'target' keep going down */
				(*brec)->record = 1;
			} else {
				/* we're at first leaf so fall through */
			}
		}

		(*brec)->data.generic = bkey_address((*brec)->key);

		/* get next node if necessary */
		if ((ntype == ndIndxNode) &&
		    !(*brec = hfs_brec_get_child(tree, *brec, keep_mask))) {
			return -EIO;
		}
	} while (ntype == ndIndxNode);

	if (ntype != ndLeafNode) {
		printk("hfs_bfind: invalid node type %02x in node %d of "
		       "btree %lu\n", bnode_head(bn)->ndType, bn->node,
		       HFS_IFID(tree->file.inode->i_ino));
		dump_brec(*brec);
		hfs_brec_relse(brec);
		return -EIO;
	}

	if ((flags & HFS_BFIND_EXACT) && result) {
		hfs_brec_relse(brec);
		return -ENOENT;
	}

	if (flags & HFS_BFIND_LOCK) {
		hfs_brec_lock(*brec);
	}

	return result ? 1 : 0;
}

/*================ Global functions ================*/

/*
 * hfs_brec_relse_one() 
 *
 * Description:
 *   Releases a (struct hfs_brec), freeing the associated memory and
 *   calling hfs_bnode_relse() on the associated bnode.
 * Input Variable(s):
 *   struct hfs_brec **brec: the address of the address of the 
 *    (struct hfs_brec) to free.
 * Output Variable(s):
 *   struct hfs_brec**brec: '*brec' is set to NULL on success,
 *    or unchanged on failure.
 * Returns:
 *   void
 * Preconditions:
 *   'brec' is NULL, points to NULL, or points to a pointer to to a "valid"
 *    (struct hfs_brec).
 * Postconditions:
 *   If '*brec' and all the (struct hfs_brec)s is "valid" then the
 *    associatedwmemory is freed, the associated (struct hfs_bnode)
 *    if any is freed and '*brec' is NULL.
 *   If '*brec' or any (struct hfs_brec) is not "valid" then
 *    nothing is changed.
 */
void hfs_brec_relse_one(struct hfs_brec **brec) {
#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
	extern int hfs_bcount;
#endif

	if (!brec) {
		return;
	}

	if ((*brec)->magic == HFS_BREC_MAGIC) {
		hfs_bnode_relse((*brec)->bnode);
		KFREE_S((*brec), sizeof(**brec));
		(*brec) = NULL;
#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
		--hfs_bcount;
#endif
	} else {
		printk("hfs_brec_relse_one: corrupted hfs_brec.\n");
	}
}

/*
 * hfs_brec_relse() 
 *
 * Description:
 *   Releases a (struct hfs_brec), freeing the associated memory and
 *   calling hfs_bnode_relse() on the associated bnode.  All (struct
 *   hfs_brec)s in the 'up' list are deleted in the same manner. 
 * Input Variable(s):
 *   struct hfs_brec **brec: the address of the address of the 
 *    (struct hfs_brec) to free.
 * Output Variable(s):
 *   struct hfs_brec**brec: '*brec' is set to NULL on success,
 *    or to the address of the first invalid (struct hfs_brec) on failure.
 * Returns:
 *   void
 * Preconditions:
 *   'brec' is NULL, points to NULL, or points to a pointer to to a "valid"
 *    (struct hfs_brec).
 * Postconditions:
 *   If '*brec' and all the (struct hfs_brec)s in the 'up' linked list
 *    are "valid" then the memory associated with each is freed, the
 *    (struct hfs_bnode) associated with each is hfs_bnode_relse()ed and
 *    '*brec' is NULL.
 *   If '*brec' or any (struct hfs_brec) in the 'up' linked list is
 *    not "valid" then only the (struct hfs_brec)s before the invalid
 *    one (if any) are freed as above,'*brec' points to the
 *    invalid one, and an error is also printed.
 */
void hfs_brec_relse(struct hfs_brec **brec) {
	struct hfs_brec *tmp;
#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
	extern int hfs_bcount;
#endif

	if (!brec) {
		return;
	}

	while ((tmp = *brec) && (tmp->magic == HFS_BREC_MAGIC)) {
		*brec = tmp->up;
		hfs_bnode_relse(tmp->bnode);
		KFREE_S(tmp, sizeof(*tmp));
#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
		--hfs_bcount;
#endif
	}
	if (tmp) {
		printk("hfs_brec_relse: corrupted hfs_brec.\n");
	}
}
