/*
 * linux/fs/hfs/btree.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the code to manipulate the B-tree structure.
 * The catalog and extents files are both B-trees.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 *
 * The code in this file initilizes some structures by calling
 * memset(&foo, 0, sizeof(foo)).  This produces the desired behavior
 * only due to the non-ANSI assumption that the machine representation
 * of NULL is all zeros.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/locks.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include "hfs_fs.h"
#include "hfs_btree.h"

/*================ File-local functions ================*/

/*
 * cat_compare()
 *
 * Description:
 *   This is the comparison function used for the catalog B-tree.  In
 *   comparing catalog B-tree entries, the parent id is the most
 *   significant field (compared as unsigned ints).  The name field is
 *   the least significant (compared in "Macintosh lexical order",
 *   see hfs_strcmp() in string.c)
 * Input Variable(s):
 *   struct hfs_brec_key *key1: pointer to the first key to compare
 *   struct hfs_brec_key *key2: pointer to the second key to compare
 * Output Variable(s):
 *   NONE
 * Returns:
 *   int: negative if key1<key2, positive if key1>key2, and 0 if key1==key2
 * Preconditions:
 *   key1 and key2 point to "valid" (struct hfs_brek_key)s.
 * Postconditions:
 *   This function has no side-effects
 */
static int cat_compare(const struct hfs_brec_key *key1,
		       const struct hfs_brec_key *key2)
{
	unsigned int parents;
	int retval;

	parents = from_l(key1->u.ckr.ParID) - from_l(key2->u.ckr.ParID);
	if (parents != 0) {
		retval = (int)parents;
	} else {
		retval = hfs_strcmp(key1->u.ckr.CName, key2->u.ckr.CName);
	}
	return retval;
}

/*
 * ext_compare()
 *
 * Description:
 *   This is the comparison function used for the extents B-tree.  In
 *   comparing extent B-tree entries, the file id is the most
 *   significant field (compared as unsigned ints); the fork type is
 *   the second most significant field (compared as unsigned chars);
 *   and the allocation block number field is the least significant
 *   (compared as unsigned ints).
 * Input Variable(s):
 *   struct hfs_brec_key *key1: pointer to the first key to compare
 *   struct hfs_brec_key *key2: pointer to the second key to compare
 * Output Variable(s):
 *   NONE
 * Returns:
 *   int: negative if key1<key2, positiove if key1>key2, and 0 if key1==key2
 * Preconditions:
 *   key1 and key2 point to "valid" (struct hfs_brec_key)s.
 * Postconditions:
 *   This function has no side-effects */
static int ext_compare(const struct hfs_brec_key *key1,
		       const struct hfs_brec_key *key2)
{
	unsigned int tmp;
	int retval;

	tmp = from_l(key1->u.xkr.FNum) - from_l(key2->u.xkr.FNum);
	if (tmp != 0) {
		retval = (int)tmp;
	} else {
		tmp = (unsigned char)key1->u.xkr.FkType
			- (unsigned char)key2->u.xkr.FkType;
		if (tmp != 0) {
			retval = (int)tmp;
		} else {
			retval = (int)(from_w(key1->u.xkr.FABN)
				       - from_w(key2->u.xkr.FABN));
		}
	}
	return retval;
}

/*
 * hfs_bnode_ditch() 
 *
 * Description:
 *   This function deletes an entire linked list of bnodes, so it
 *   does not need to keep the linked list consistent as
 *   hfs_bnode_delete() does.
 *   Called by hfs_btree_init() for error cleanup and by hfs_btree_free().
 * Input Variable(s):
 *   struct hfs_bnode *bn: pointer to the first (struct hfs_bnode) in
 *    the linked list to be deleted.
 * Output Variable(s):
 *   NONE
 * Returns:
 *   void
 * Preconditions:
 *   'bn' is NULL or points to a "valid" (struct hfs_bnode) with a 'prev'
 *    field of NULL.
 * Postconditions:
 *   'bn' and all (struct hfs_bnode)s in the chain of 'next' pointers
 *   are deleted, freeing the associated memory and brelse()ing the
 *   associated buffer_heads.
 */
static void hfs_bnode_ditch(struct hfs_bnode *bn) {
	struct hfs_bnode *tmp;
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
	extern int bnode_count;
#endif

	while (bn != NULL) {
		tmp = bn->next;
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
		printk("deleting node %d from tree %d with count %d\n",
		       bn->node, HFS_IFID(bn->tree->file.inode->i_ino),
		       bn->count);
		--bnode_count;
#endif
		brelse(bn->bh); /* safe: brelse() checks for NULL argument */
		KFREE_S(bn, sizeof(struct hfs_bnode));
		bn = tmp;
	}
}

/*================ Global functions ================*/

/*
 * hfs_btree_free()
 *
 * Description:
 *   This function frees a (struct hfs_btree) obtained from hfs_btree_init().
 *   Called by hfs_put_super().
 * Input Variable(s):
 *   struct hfs_btree *bt: pointer to the (struct hfs_btree) to free
 * Output Variable(s):
 *   NONE
 * Returns:
 *   void
 * Preconditions:
 *   'bt' is NULL or points to a "valid" (struct hfs_btree)
 * Postconditions:
 *   If 'bt' points to a "valid" (struct hfs_btree) then all (struct
 *    hfs_bnode)s associated with 'bt' are freed by calling
 *    hfs_bnode_ditch() and the memory associated with the (struct
 *    hfs_btree) is freed.
 *   If 'bt' is NULL or not "valid" an error is printk()'d and nothing
 *    is changed.
 */
void hfs_btree_free(struct hfs_btree *bt)
{
	int lcv;

	if (bt && (bt->magic == HFS_BTREE_MAGIC)) {
		bt->file.inode->i_dirt = 0;
		iput(bt->file.inode);

		hfs_extent_trim(&bt->file, &bt->file.first);

		for (lcv=0; lcv<HFS_CACHELEN; ++lcv) {
#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
			printk("deleting nodes from bucket %d:\n", lcv);
#endif
			hfs_bnode_ditch(bt->cache[lcv]);
		}

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
		printk("deleting header and bitmap nodes\n");
#endif
		hfs_bnode_ditch(bt->head);

#if defined(DEBUG_BNODES) || defined(DEBUG_ALL)
		printk("deleting root node\n");
#endif
		hfs_bnode_ditch(bt->root);

		KFREE_S(bt, sizeof(struct hfs_btree));
	} else if (bt) {
		printk("hfs_btree_free: corrupted hfs_btree.\n");
	} else {
		printk("hfs_btree_free: NULL argument.\n");
	}
}

/*
 * hfs_btree_init()
 *
 * Description:
 *   Given some vital information from the MDB (HFS superblock),
 *   initializes the fields of a (struct hfs_btree).
 * Input Variable(s):
 *   struct super_block *sb: pointer to the superblock
 *   ino_t fid: the file ID (HFS_CAT_FID or HFS_EXT_FID) of the B-tree
 *   const struct hfs_raw_extent *ext: pointer to the raw extent
 *    record for the B-tree
 *   HFS_LWORD tsize: the size, in bytes, of the B-tree
 *   HFS_LWORD csize: the size, in bytes, of the clump size for the B-tree
 * Output Variable(s):
 *   NONE
 * Returns:
 *   (struct hfs_btree *): pointer to the initialized hfs_btree on sucess,
 *    or NULL on failure
 * Preconditions:
 *   'sb' points to a "valid" (struct super_block)
 *   'ext' points to a "valid" (struct hfs_raw_inode)
 * Postconditions:
 *   Assuming the inputs are what they claim to be, no errors occur
 *   reading from disk, and no inconsistencies are noticed in the data
 *   read from disk, the return value is a pointer to a "valid"
 *   (struct hfs_btree).  If there are errors reading from disk or
 *   inconsistencies are noticed in the data read from disk, then and
 *   all resources that were allocated are released and NULL is
 *   returned.	If the inputs are not what they claim to be or if they
 *   are unnoticed inconsistencies in the data read from disk then the
 *   returned hfs_btree is probably going to lead to errors when it is
 *   used in a non-trivial way.
 */
struct hfs_btree * hfs_btree_init(struct super_block *sb, ino_t fid,
				  struct hfs_raw_extent *ext,
				  HFS_LWORD *tsize, HFS_LWORD csize)
{
	struct hfs_btree * bt;
	struct BTHdrRec * th;
	struct hfs_bnode * tmp;
	unsigned int next;
#if defined(DEBUG_HEADER) || defined(DEBUG_ALL)
	int i;
#endif

	if (!sb || !ext ||
			!(bt=KMALLOC(sizeof(struct hfs_btree), GFP_KERNEL))) {
		goto bail3;
	}

	bt->magic = HFS_BTREE_MAGIC;
	bt->raw_extent = ext;
	bt->size = tsize;
	bt->reserved = 0;
	memset(bt->cache, 0, HFS_CACHELEN*sizeof(struct hfs_bnode *));
	bt->file.magic = HFS_FILE_MAGIC;
	bt->file.dev = sb->s_dev;
	bt->file.clumpsize = from_l(csize);
	bt->file.fork = HFS_FK_DATA;
	bt->file.inode = get_empty_inode();
	bt->file.inode->i_blocks = from_l(*tsize) >> HFS_SECTOR_SIZE_BITS;
	bt->file.inode->i_ino = fid;
	bt->file.inode->i_sb = sb;
	/* bt->file.inode->i_dev = sb->s_dev; If we did we couldn't umount. */
	hfs_extent_in(&bt->file, ext);

	if (!(bt->head=hfs_bnode_read(bt, 0, HFS_STICKY))) {
		goto bail2;
	}
	th = btree_head(bt);

	/* read in the bitmap nodes (if any) */
	tmp = bt->head;
	while ((next = from_l(bnode_head(tmp)->ndFLink))) {
		if (!(tmp->next=hfs_bnode_read(bt, next, HFS_STICKY))) {
			goto bail1;
		}
		tmp->next->prev = tmp;
		tmp = tmp->next;
	}

#if defined(DEBUG_HEADER) || defined(DEBUG_ALL)
	printk("bthDepth %d\n", (int)from_w(th->bthDepth));
	printk("bthRoot %d\n", (int)from_l(th->bthRoot));
	printk("bthNRecs %d\n", (int)from_l(th->bthNRecs));
	printk("bthFNode %d\n", (int)from_l(th->bthFNode));
	printk("bthLNode %d\n", (int)from_l(th->bthLNode));
	printk("bthKeyLen %d\n", (int)from_w(th->bthKeyLen));
	printk("bthNNodes %d\n", (int)from_l(th->bthNNodes));
	printk("bthFree %d\n", (int)from_l(th->bthFree));
	for (i=0; i<HFS_SECTOR_SIZE; i+=16) {
		printk("%02x %02x %02x %02x %02x %02x %02x %02x %02x "
		       "%02x %02x %02x %02x %02x %02x %02x\n",
			((__u8 *)bt->head->bh->b_data)[i],
			((__u8 *)bt->head->bh->b_data)[i+1],
			((__u8 *)bt->head->bh->b_data)[i+2],
			((__u8 *)bt->head->bh->b_data)[i+3],
			((__u8 *)bt->head->bh->b_data)[i+4],
			((__u8 *)bt->head->bh->b_data)[i+5],
			((__u8 *)bt->head->bh->b_data)[i+6],
			((__u8 *)bt->head->bh->b_data)[i+7],
			((__u8 *)bt->head->bh->b_data)[i+8],
			((__u8 *)bt->head->bh->b_data)[i+9],
			((__u8 *)bt->head->bh->b_data)[i+10],
			((__u8 *)bt->head->bh->b_data)[i+11],
			((__u8 *)bt->head->bh->b_data)[i+12],
			((__u8 *)bt->head->bh->b_data)[i+13],
			((__u8 *)bt->head->bh->b_data)[i+14],
			((__u8 *)bt->head->bh->b_data)[i+15]);
	}
#endif

	if (from_w(th->bthNodeSize) != HFS_SECTOR_SIZE) {
		printk("hfs_btree_init: bthNodeSize!=512 not supported\n");
		goto bail1;
	}

	if (fid == HFS_CAT_FID) {
		bt->compare = cat_compare;
	} else if (fid == HFS_EXT_FID) {
		bt->compare = ext_compare;
	} else {
		goto bail1;
	}
	bt->bthRoot = from_l(th->bthRoot);
	bt->bthKeyLen = from_w(th->bthKeyLen);

	/* Read in the root if it exists.
	   The header always exists, but the root exists only if the
	   tree is non-empty */
	if ((th->bthDepth != c_to_w(0)) && bt->bthRoot) {
		if (!(bt->root=hfs_bnode_read(bt, bt->bthRoot, HFS_STICKY))) {
			goto bail1;
		}
	} else {
		bt->root = NULL;
	}

	return bt;

 bail1:
	hfs_bnode_ditch(bt->head);
 bail2:
	KFREE_S(bt, sizeof(struct hfs_btree));
 bail3:
	return NULL;
}
