/*
 * linux/fs/hfs/catalog.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the functions related to the catalog B-tree.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/mm.h>
#include <linux/fs.h>

/*================ Function-like macros ================*/

/* get cdrType of a catalog **brec */
#define TYPE(brec) ((*brec)->data.raw_inode->cdrType)

/* get catalog B-tree of filesystem from an inode */
#define CAT_TREE(sb) (HFS_SB(sb)->s_cat_tree)

/*================ Global functions ================*/

/*
 * hfs_build_cat_key()
 *
 * Given the ID of the parent and the name build a search key.
 */
void hfs_build_cat_key(ino_t parent, __u8 len, const char *name,
		       struct hfs_brec_key *key)
{
	/* silently truncate long names */
	if (len > HFS_NAMELEN) {
		len = HFS_NAMELEN;
	}

	memset(key, 0, sizeof(*key));
	key->KeyLen = 6 + len;
	key->u.ckr.ParID = to_l(parent);
	key->u.ckr.CName[0] = len;
	memcpy(key->u.ckr.CName+1, name, len);
}

/*
 * hfs_cat_find()
 *
 * Given a search key find the matching catalog brec, returning 0 on
 * success, or an error code on failure.  Note that read only access
 * is obtained for the record in question if write=0, and write access
 * otherwise.
 */
int hfs_cat_find(struct super_block *sb, const struct hfs_brec_key *target,
		 struct hfs_brec **brec, int write)
{
	if (!sb) {
		return -EINVAL;
	}

	return hfs_bfind(CAT_TREE(sb), target, brec,
			 write ? HFS_BFIND_WRITE : HFS_BFIND_READ_EQ);
}

/*
 * hfs_cat_next()
 *
 * Given a catalog brec structure, replace it with the count'th next brec.
 * If (skip & HFS_SKIP_DIR), then don't count directories.
 * Return an error code if there is a problem, 0 if OK.
 * Note that an error code of -ENOENT means there are no more entries
 * in this directory.
 */
int hfs_cat_next(struct hfs_brec **brec, __u32 count, int skip)
{
	int error;

	if (!brec) {
		return -EINVAL;
	}

	while (count--) {
		do {
			/* Get the next catalog tree entry */
			if ((error = hfs_bsucc(brec))) {
				return error;
			}
			/* Check for end of directory */
			if (TYPE(brec)==HFS_CDR_THD) {
				hfs_brec_relse(brec);
				return -ENOENT;
			}
		} while (((skip&HFS_SKIP_DIR) && (TYPE(brec)==HFS_CDR_DIR)) ||
			 ((skip&HFS_SKIP_FTH) && (TYPE(brec)==HFS_CDR_FTH)) ||
			 ((skip&HFS_SKIP_FIL) && (TYPE(brec)==HFS_CDR_FIL)));
	}
	return 0;
}

/*
 * hfs_cat_nth()
 *
 * Given the super_block of an HFS filesystem, the file id of a
 * directory in that filesystem, and an integer n return the nth entry
 * in that directory in *brec.	Return 0 on success or an error code
 * on failure.	If (skip & HFS_SKIP_DIR) then entries for directories
 * are not counted toward 'n'.
 */
int hfs_cat_nth(struct inode *dir, struct hfs_brec **brec, __u32 n, int skip)
{
	struct hfs_brec_key key;
	int error;

	if (!n) {
		return -EINVAL;
	}
	
	/* Find the directory */
	hfs_build_cat_key(HFS_IFID(dir->i_ino), 0, NULL, &key);
	if ((error = hfs_bfind(CAT_TREE(dir->i_sb), &key, brec,
			       HFS_BFIND_READ_EQ))) {
		return error;
	}

	/* Jump to the 'n'th entry or the end of the directory */
	error = hfs_cat_next(brec, n, skip);
	
	return error;
}
