/*
 * linux/fs/hfs/dir.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains directory-related functions independent of which
 * scheme is being used to represent forks.
 *
 * Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/stat.h>

/*================ Function-like macros ================*/

/* Given a (struct inode *) produce a pointer to the search key for
   the corresponding HFS file or directory */
#define KEY(x)			((struct hfs_brec_key *)HFS_I(x)->key)

/* Given a (struct inode *) call the s_nameout() function of its filesystem */
#define NAMEOUT(x,a,b,c)        (HFS_SB(x->i_sb)->s_nameout(a,b,c))

/*================ File-local functions ================*/

/*
 * hfs_dupdir()
 *
 * The code common to cap_dupdir() and nat_dupdir().
 */
int hfs_dupdir(struct inode * dir, ino_t type, struct inode ** result)
{
	struct inode * tmp = iget(dir->i_sb, type | HFS_IFID(dir->i_ino));

	(*result) = tmp;
	if (tmp && !tmp->i_mode) {
		memcpy(HFS_I(tmp), HFS_I(dir), sizeof(*HFS_I(dir)));
		tmp->i_mtime = dir->i_mtime;
		tmp->i_ctime = dir->i_ctime;
		tmp->i_atime = dir->i_atime;
		tmp->i_mode = dir->i_mode;
		tmp->i_blocks = 0;
		tmp->i_uid = dir->i_uid;
		tmp->i_gid = dir->i_gid;
		return 1;
	}
	return 0;
}

/*
 * hfs_lookup_parent()
 *
 * Called when lookup() is called on ".." in a regular directory.
 * Uses the thread record for the current directory to locate its
 * parent.
 * Note that this function does an iput(dir) before returning.
 */
struct inode * hfs_lookup_parent(struct inode *dir, int *error)
{
	struct hfs_brec_key key;
	struct hfs_brec *entry;
	struct inode *retval = NULL;

	key.u.ckr.ParID = KEY(dir)->u.ckr.ParID;
	key.u.ckr.CName[0] = 0;
	if (!(*error = hfs_cat_find(dir->i_sb, &key, &entry, 0))) {
		retval = hfs_iget(dir->i_sb,
				  from_l(entry->data.raw_inode->u.thd.ParID),
				  entry->data.raw_inode->u.thd.CName[0],
				  entry->data.raw_inode->u.thd.CName+1,
				  HFS_ITYPE_NORM);
		hfs_brec_relse(&entry);
	}
	iput(dir);
	return retval;
}

/*================ Global functions ================*/

/*
 * hfs_dir_read()
 *
 * This is the read() entry in the file_operations structure for HFS
 * directories.	 It simply returns an error code, since reading is not
 * supported.
 */
int hfs_dir_read(struct inode * inode, struct file * filp,
			char *buf, int count)
{
	return -EISDIR;
}

/*
 * hfs_do_create()
 *
 * This function does the main work of adding a new file to a directory
 * by adding a file entry to the catalog tree and creating an inode
 * in memory.
 * The caller is responsible for updating the directory inode.
 */
int hfs_do_create(struct inode * dir, const char * name, int len, int mode,
		  struct inode ** result)
{
	char tmp_name[HFS_NAMELEN];
	struct hfs_brec_key key;
	struct hfs_inode rec;
	int error;
	__u32 fid;

	/* mangle the name */
	len = NAMEOUT(dir, tmp_name, len, name);

	/* allocate an id */
	fid = HFS_SB(dir->i_sb)->s_next_id++;
	dir->i_sb->s_dirt = 1;

	/* init some fields for the file record */
	memset(&rec, 0, sizeof(rec));
	rec.cdrType = HFS_CDR_FIL;
	rec.u.fil.Flags = (mode & S_IWUSR ? 0 : 1);
	rec.u.fil.FlNum = to_l(fid);
	rec.u.fil.CrDat = rec.u.fil.MdDat = hfs_u_to_mtime(CURRENT_TIME);
	rec.u.fil.BkDat = c_to_l(0x80000000);

	/* insert the file record */
	hfs_build_cat_key(HFS_IFID(dir->i_ino), len, tmp_name, &key);
        error = hfs_binsert(HFS_SB(dir->i_sb)->s_cat_tree,
			    &key, &rec, 2+sizeof(struct hfs_fil_inode));
	if (error) {
		return error;
	}

	/* try to create an inode for the new file */
	*result = __hfs_iget(dir->i_sb, &key, HFS_ITYPE_NORM);
	if (!(*result)) {
		/* XXX should remove file record */
		return -EIO;
	}

	/* update counts */
	++HFS_SB(dir->i_sb)->s_file_count;
	if (HFS_IFID(dir->i_ino) == HFS_ROOT_FID) {
		++HFS_SB(dir->i_sb)->s_root_files;
	}
	dir->i_sb->s_dirt = 1;

	return 0;
}

/*
 * hfs_do_mkdir()
 *
 * This function does the main work of creating a new directory by 
 * adding the directory and thread entries to the catalog tree.
 * The caller is responsible for updating the directory inode after
 * this function returns and for ensuring that the name does not
 * conflict with any reserved names.
 */
int hfs_do_mkdir(struct inode * dir, const HFS_CNAME name)
{
	struct hfs_brec_key key;
	struct hfs_inode rec;
	int error;
	__u32 fid;

	/* allocate an id */
	fid = HFS_SB(dir->i_sb)->s_next_id++;
	dir->i_sb->s_dirt = 1;

	/* init some fields for the directory record */
	memset(&rec, 0, sizeof(rec));
	rec.cdrType = HFS_CDR_DIR;
	rec.u.dir.DirID = to_l(fid);
	rec.u.dir.CrDat = rec.u.dir.MdDat = hfs_u_to_mtime(CURRENT_TIME);
	rec.u.dir.BkDat = c_to_l(0x80000000);

	/* insert the directory record */
	hfs_build_cat_key(HFS_IFID(dir->i_ino), name[0], name+1, &key);
	error = hfs_binsert(HFS_SB(dir->i_sb)->s_cat_tree,
			    &key, &rec, 2+sizeof(struct hfs_dir_inode));
	if (error) {
		return error;
	}

	/* init some fields for the thread record */
	memset(&rec, 0, sizeof(rec));
	rec.cdrType = HFS_CDR_THD;
	rec.u.thd.ParID = to_l(HFS_IFID(dir->i_ino));
	rec.u.thd.CName[0] = name[0];
	memcpy(rec.u.thd.CName+1, name+1, name[0]);

	/* insert the thread record */
	hfs_build_cat_key(fid, 0, NULL, &key);
	error = hfs_binsert(HFS_SB(dir->i_sb)->s_cat_tree,
			    &key, &rec, 2+sizeof(struct hfs_thd_inode));
	if (error) {
		/* XXX should remove dir record */
		return error;
	}

	/* update counts */
	++HFS_SB(dir->i_sb)->s_dir_count;
	if (HFS_IFID(dir->i_ino) == HFS_ROOT_FID) {
		++HFS_SB(dir->i_sb)->s_root_dirs;
	}
	dir->i_sb->s_dirt = 1;

	return 0;
}
