/*
 * linux/fs/hfs/dir_cap.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the inode_operations and file_operations
 * structures for HFS directories under the CAP scheme.
 *
 * Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <asm/segment.h>
#include <linux/stat.h>

/*================ Forward declarations ================*/
static int hfs_cap_lookup(struct inode *, const char *, int, struct inode **);
static int hfs_cap_readdir(struct inode *, struct file *, void *, filldir_t);
static int hfs_cap_create(struct inode *, const char *, int, int,
                          struct inode **);
static int hfs_cap_mkdir(struct inode *, const char *, int, int);

/*================ Global variables ================*/

static struct file_operations hfs_cap_dir_operations = {
	NULL,			/* lseek - default */
	hfs_dir_read,		/* read - invalid */
	NULL,			/* write - bad */
	hfs_cap_readdir,	/* readdir */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	NULL			/* fsync */
};

struct inode_operations hfs_cap_dir_inode_operations = {
	&hfs_cap_dir_operations,/* default directory file-ops */
	hfs_cap_create,		/* create */
	hfs_cap_lookup,		/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	hfs_cap_mkdir,		/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

struct inode_operations hfs_cap_alt_dir_inode_operations = {
	&hfs_cap_dir_operations,/* default directory file-ops */
	NULL,			/* create */
	hfs_cap_lookup,		/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

/*================ Function-like macros ================*/

/* Given a (struct inode *) produce a pointer to the search key for
   the corresponding HFS file or directory */
#define KEY(x)			((struct hfs_brec_key *)HFS_I(x)->key)

/* Given a (struct inode *) produce the File ID of its parent */
#define PARENT(x)		from_l(KEY(x)->u.ckr.ParID)

/* Given a (struct inode *) call the s_nameout() function of its filesystem */
#define NAMEOUT(x,a,b,c)	(HFS_SB(x->i_sb)->s_nameout(a,b,c))

/* Given a (struct inode *) call the s_namein() function of its filesystem */
#define NAMEIN(x,a,b)		(HFS_SB(x->i_sb)->s_namein(a,b))

/*================ File-local variables ================*/

static const HFS_CNAME cap_reserved_names[] = {
	"\01.\0",
	"\02..\0",
	"\011.resource\0",
	"\013.finderinfo\0",
};
#define RNAMES (sizeof(cap_reserved_names)/sizeof(HFS_CNAME))

/*================ File-local functions ================*/

/*
 * cap_dupdir()
 *
 * Called when a lookup() call is issued for the ".resource" or
 * ".finderinfo" directories or for ".." in one of those.  The purpose
 * of this function is to generate an inode for the requested
 * directory from the inode for the current directory.
 * Note that we do a iput(dir) before returning.
 */
static struct inode * cap_dupdir(struct inode * dir, ino_t type)
{
	struct inode *tmp;

	if (type == HFS_CAP_BDIR) {
		tmp = NULL;
	} else if (hfs_dupdir(dir, type, &tmp)) {
		if (type == HFS_CAP_NDIR) {
			tmp->i_mode = S_IRWXUGO | S_IFDIR;
			tmp->i_size = HFS_I(dir)->u.dir.files +
				HFS_I(dir)->u.dir.dirs + 4;
			tmp->i_nlink = HFS_I(dir)->u.dir.dirs + 4;
			tmp->i_op = &hfs_cap_dir_inode_operations;
		} else if (type == HFS_CAP_FDIR) {
			tmp->i_mode = S_IRUGO | S_IXUGO | S_IFDIR;
			tmp->i_size = HFS_I(dir)->u.dir.files +
				HFS_I(dir)->u.dir.dirs + 2;
			tmp->i_nlink = 2;
			tmp->i_op = &hfs_cap_alt_dir_inode_operations;
		} else /* if (type == HFS_CAP_RDIR) */ {
			tmp->i_mode = S_IRUGO | S_IXUGO | S_IFDIR;
			tmp->i_size = HFS_I(dir)->u.dir.files + 2;
			tmp->i_nlink = 2;
			tmp->i_op = &hfs_cap_alt_dir_inode_operations;
		}
		tmp->i_mode &= ~HFS_SB(dir->i_sb)->s_umask;
	}
	iput(dir);
	return tmp;
}

/*
 * hfs_cap_lookup()
 *
 * This is the lookup() entry in the inode_operations structure for
 * HFS directories in the CAP scheme.  The purpose is to generate the
 * inode corresponding to an entry in a directory, given the inode for
 * the directory and the name (and its length) of the entry.
 */
static int hfs_cap_lookup(struct inode * dir, const char * name,
			  int len, struct inode ** result)
{
	ino_t type = HFS_ITYPE(dir->i_ino);
	char tmp_name[HFS_NAMELEN];
	
	*result = NULL;
	
	if (!dir || !S_ISDIR(dir->i_mode)) {
		iput(dir);
		return -ENOENT;
	}

	if (len && !name) {
		iput(dir);
		return -EINVAL;
	}

	/* Check for "." */
	if (len==1 && name[0] == '.') {
		/* this little trick skips the iget and iput */
		*result = dir;
		return 0;
	}

	/* Check for "..".  Note that cap_dupdir() and hfs_lookup_parent()
	   each do a iput(dir). */
	if (len==2 && name[0] == '.' && name[1] == '.') {
		if (type != HFS_CAP_NDIR) {
			/* Case for ".." in ".finderinfo" or ".resource" */
			*result = cap_dupdir(dir, HFS_CAP_NDIR);
			return (*result) ? 0 : -ENOENT;
		} else {
			/* Case for ".." in a normal directory */
			int tmp;
			*result = hfs_lookup_parent(dir, &tmp);
			return (*result || tmp) ? tmp : -ENOENT;
		}
	}

	/* Translate 'type' of 'dir' to 'type' of file/dir to get.
	   Also check for special directories if in a normal directory.
	   Note that cap_dupdir() does an iput(dir). */
	if (type==HFS_CAP_NDIR) {
		/* Check for ".resource" or ".finderinfo"
		   Note that cap_dupdir() will call iput(dir). */
		if (len==9 && !strncmp(name, ".resource", 9)) {
			*result = cap_dupdir(dir, HFS_CAP_RDIR);
			return (*result) ? 0 : -ENOENT;
		} else if (len==11 && !strncmp(name, ".finderinfo", 11)) {
			*result = cap_dupdir(dir, HFS_CAP_FDIR);
			return (*result) ? 0 : -ENOENT;
		}
		type = HFS_CAP_NORM;
	} else if (type==HFS_CAP_RDIR) {
		type = HFS_CAP_RSRC;
	} else if (type==HFS_CAP_FDIR) {
		type = HFS_CAP_FNDR;
	} else {
		/* Something is seriously wrong if we get here! */
		iput(dir);
		return -ENOENT;
	}


	/* Perform name-mangling */
	len = NAMEOUT(dir, tmp_name, len, name);

	/* Do an hfs_iget() on the mangled name with the 'type' argument
	   as decoded above. */
	*result = hfs_iget(dir->i_sb, HFS_IFID(dir->i_ino),
			   len, tmp_name, type);
	iput(dir);
	return (*result) ? 0 : -ENOENT;
}

/*
 * hfs_cap_readdir()
 *
 * This is the readdir() entry in the file_operations structure for
 * HFS directories in the CAP scheme.  The purpose is to enumerate the
 * entries in a directory, given the inode of the directory and a
 * (struct file *), the 'f_pos' field of which indicates the location
 * in the directory.  The (struct file *) is updated so that the next
 * call with the same 'dir' and 'filp' arguments will produce the next
 * directory entry.  The entries are returned in 'dirent', which is
 * "filled-in" by calling filldir().  This allows the same readdir()
 * function be used for different dirent formats.  We try to read in
 * as many entries as we can before filldir() refuses to take any more.
 *
 * XXX: In the future it may be a good idea to consider not generating
 * meta-data files for covered directories since the data doesn't
 * correspond to the mounted directory.	 However this requires an
 * iget() for every directory which could be considered an excessive
 * amount of overhead.	Since the inode for a mount point is always
 * in-core this is another argument for a call to get an inode if it
 * is in-core or NULL if it is not.
 * (See the comments with hfs_notify_chage() in hfs/inode.c for the
 * primary argument for this idea.)
 */
static int hfs_cap_readdir(struct inode * dir, struct file * filp,
			   void * dirent, filldir_t filldir)
{
	int res, old_pos, pos, version, skip;
	ino_t fid, ino = dir->i_ino;
	unsigned int len;
	struct hfs_brec *entry;
	char fixed_name[HFS_NAMEMAX];

	if (!dir || !dir->i_sb || !S_ISDIR(dir->i_mode)) {
		return -EBADF;
	}
	
	res = 0;
	while (!res && (filp->f_pos < dir->i_size)) {
		pos = old_pos = filp->f_pos;
		if (!pos) {
			/* Entry 0 is for "." */
			res = filldir(dirent, ".", 1, 0, ino);
		} else if (!(--pos)) {
			/* Entry 1 is for ".." */
			res = filldir(dirent, "..", 2, 1, PARENT(dir));
		} else if ((HFS_ITYPE(ino)==HFS_CAP_NDIR) && !(--pos)) {
			/* In normal dirs entry 2 is for ".finderinfo" */
			res = filldir(dirent, ".finderinfo", 11, old_pos,
				      HFS_IFID(ino) | HFS_CAP_FDIR);
		} else if ((HFS_ITYPE(ino)==HFS_CAP_NDIR) && !(--pos)) {
			/* In normal dirs entry 3 is for ".resource" */
			res = filldir(dirent, ".resource", 9, old_pos,
				      HFS_IFID(ino) | HFS_CAP_RDIR);
		} else {
			/* In regular directories and ".finderinfo"
			   there is a Linux directory entry for each
			   HFS entry, but in ".resource" only files
			   have entries, not directories.  Due to the
			   way we've decremented 'pos' in the tests
			   above, it is now the index of the desired
			   entry. */
			skip = (HFS_ITYPE(ino)==HFS_CAP_RDIR) ?
					HFS_SKIP_DIR : HFS_SKIP_NONE;
			/* The call to filldir() may involve transfers
			   to user-space which can cause a call to
			   schedule().	Therefore before we proceed,
			   we check that nobody changed the directory.
			   If the directory has been changed, then we
			   try again. */
			do {
				version = dir->i_version;
				if ((res=hfs_cat_nth(dir, &entry, pos, skip))) {
					return res;
				}
				fid = HFS_FID(entry->data.raw_inode);
				len = NAMEIN(dir, fixed_name,
					     entry->key->u.ckr.CName);
				hfs_brec_relse(&entry);
				res = filldir(dirent, fixed_name, len,
					      old_pos, fid);
			} while (version != dir->i_version);
		}
		if (!res) {
			++filp->f_pos;
		}
	}
	return 0;
}

/*
 * hfs_cap_create()
 *
 * This is the create() entry in the file_operations structure for
 * HFS directories in the CAP scheme.  The purpose is to create a
 * new file in a directory, given the inode of the directory, the
 * name (and its length) for the new file and its mode.
 * On success *result points to an inode for the new file and 0 is returned.
 * On failure *result is NULL and an error code is returned.
 */
int hfs_cap_create(struct inode * dir, const char * name, int len, int mode,
		   struct inode ** result)
{
	int error;
	ino_t ino;
	struct inode *tmp;

	*result = NULL;
	if (!dir) {
		return -ENOENT;
	}

	error = hfs_do_create(dir, name, len, mode, result);
	if (error) {
		iput(dir);
		return error;
	}

	++dir->i_size;
	++HFS_I(dir)->u.dir.files;
	dir->i_dirt = 1;
	iput(dir);

	ino = HFS_IFID(dir->i_ino) | HFS_CAP_FDIR;
        if ((tmp = iget(dir->i_sb, ino))) {
                if ((tmp->i_sb != dir->i_sb) || (tmp->i_ino != ino)) {
                        iput(tmp);
                } else if (!tmp->i_mode) {
                        clear_inode(tmp);
                } else {
			++tmp->i_size;
			++HFS_I(tmp)->u.dir.files;
                        iput(tmp);
		}
	}
         
	ino = HFS_IFID(dir->i_ino) | HFS_CAP_RDIR;
        if ((tmp = iget(dir->i_sb, ino))) {
                if ((tmp->i_sb != dir->i_sb) || (tmp->i_ino != ino)) {
                        iput(tmp);
                } else if (!tmp->i_mode) {
                        clear_inode(tmp);
                } else {
			++tmp->i_size;
			++HFS_I(tmp)->u.dir.files;
                        iput(tmp);
		}
	}
         
	return 0;
}

/*
 * hfs_cap_mkdir()
 *
 * This is the mkdir() entry in the file_operations structure for
 * HFS directories in the CAP scheme.  The purpose is to create a
 * new directory, given the inode of the parent directory, the
 * name (and its length) for the new directory and its mode.
 * On success 0 is returned.
 * On failure an error code is returned.
 */
int hfs_cap_mkdir(struct inode * dir, const char * name, int len, int mode)
{
	HFS_CNAME tmp_name;
	int error, i;

	if (!dir) {
		return -ENOENT;
	}

	/* mangle the name and check against reserved names */
	tmp_name[0] = NAMEOUT(dir, tmp_name+1, len, name);
	for (i=0; i<RNAMES; ++i) {
		if (!hfs_strcmp(tmp_name, cap_reserved_names[i])) {
			iput(dir);
			return -EEXIST;
		}
	}

	error = hfs_do_mkdir(dir, tmp_name);
	if (error) {
		iput(dir);
		return error;
	}

	++dir->i_size;
	++HFS_I(dir)->u.dir.dirs;
	dir->i_dirt = 1;
	iput(dir);

	return 0;
}
