/*
 * linux/fs/hfs/dir_dbl.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the inode_operations and file_operations
 * structures for HFS directories.
 *
 * Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <asm/segment.h>
#include <linux/stat.h>

/*================ Forward declarations ================*/
static int hfs_dbl_lookup(struct inode *, const char *, int, struct inode **);
static int hfs_dbl_readdir(struct inode *, struct file *, void *, filldir_t);
static int hfs_dbl_create(struct inode *, const char *, int, int,
                          struct inode **);
static int hfs_dbl_mkdir(struct inode *, const char *, int, int);

/*================ Global variables ================*/

static struct file_operations hfs_dbl_dir_operations = {
	NULL,			/* lseek - default */
	hfs_dir_read,		/* read - invalid */
	NULL,			/* write - bad */
	hfs_dbl_readdir,	/* readdir */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* no special open code */
	NULL,			/* no special release code */
	NULL			/* fsync */
};

struct inode_operations hfs_dbl_dir_inode_operations = {
	&hfs_dbl_dir_operations,/* default directory file-ops */
	hfs_dbl_create,		/* create */
	hfs_dbl_lookup,		/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	hfs_dbl_mkdir,		/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

/*================ Function-like macros ================*/

/* Given a (struct inode *) produce a pointer to the search key for
   the corresponding HFS file or directory */
#define KEY(x)			((struct hfs_brec_key *)HFS_I(x)->key)

/* Given a (struct inode *) produce the File ID of its parent */
#define PARENT(x)		from_l(KEY(x)->u.ckr.ParID)

/* Given a (struct inode *) call the s_nameout() function of its filesystem */
#define NAMEOUT(x,a,b,c)	(HFS_SB(x->i_sb)->s_nameout(a,b,c))

/* Given a (struct inode *) call the s_namein() function of its filesystem */
#define NAMEIN(x,a,b)		(HFS_SB(x->i_sb)->s_namein(a,b))

/*================ File-local variables ================*/

static const HFS_CNAME dbl_reserved_names[] = {
        "\01.\0",
        "\02..\0",
};
#define RNAMES (sizeof(dbl_reserved_names)/sizeof(HFS_CNAME))

/*================ File-local functions ================*/

/*
 * hfs_dbl_lookup()
 *
 * This is the lookup() entry in the inode_operations structure for
 * HFS directories in the AppleDouble scheme.  The purpose is to
 * generate the inode corresponding to an entry in a directory, given
 * the inode for the directory and the name (and its length) of the
 * entry.
 */
static int hfs_dbl_lookup(struct inode * dir, const char * name,
			  int len, struct inode ** result)
{
	char tmp_name[HFS_NAMELEN];
	int tmp;
	
	*result = NULL;
	
	if (!dir || !S_ISDIR(dir->i_mode)) {
		iput(dir);
		return -ENOENT;
	}
	
	if (len && !name) {
		iput(dir);
		return -EINVAL;
	}

	/* Check for "." */
	if (len==1 && name[0] == '.') {
		/* this little trick skips the iget and iput */
		*result = dir;
		return 0;
	}

	/* Check for "..".  Note that hfs_lookup_parent() does a iput(dir) */
	if (len==2 && name[0] == '.' && name[1] == '.') {
		*result = hfs_lookup_parent(dir, &tmp);
		return (*result || tmp) ? tmp : -ENOENT;
	}

	/* If name starts with "%" try it as a header file first */
	if (*name=='%') {
		tmp = NAMEOUT(dir, tmp_name, len-1, name+1);
		*result = hfs_iget(dir->i_sb, HFS_IFID(dir->i_ino), tmp,
				   tmp_name, HFS_DBL_HDR);
	}

	/* Try as a normal file/dir unless we found it as a header */
	if (!*result) {
		tmp = NAMEOUT(dir, tmp_name, len, name);
		*result = hfs_iget(dir->i_sb, HFS_IFID(dir->i_ino), tmp,
				   tmp_name, HFS_DBL_NORM);
	}
	
	iput(dir);
	return (*result) ? 0 : -ENOENT;
}

/*
 * hfs_dbl_readdir()
 *
 * This is the readdir() entry in the file_operations structure for
 * HFS directories in the AppleDouble scheme.  The purpose is to
 * enumerate the entries in a directory, given the inode of the
 * directory and a (struct file *), the 'f_pos' field of which
 * indicates the location in the directory.  The (struct file *) is
 * updated so that the next call with the same 'dir' and 'filp'
 * arguments will produce the next directory entry.  The entries are
 * returned in 'dirent', which is "filled-in" by calling filldir().
 * This allows the same readdir() function be used for different
 * formats.  We try to read in as many entries as we can before
 * filldir() refuses to take any more.
 *
 * XXX: In the future it may be a good idea to consider not generating
 * meta-data files for covered directories since the data doesn't
 * correspond to the mounted directory.	 However this requires an
 * iget() for every directory which could be considered an excessive
 * amount of overhead.	Since the inode for a mount point is always
 * in-core this is another argument for a call to get an inode if it
 * is in-core or NULL if it is not.
 * (See the comments with hfs_notify_chage() in hfs/inode.c for the
 * primary argument for this idea.)
 */
static int hfs_dbl_readdir(struct inode * dir, struct file * filp,
			   void * dirent, filldir_t filldir)
{
	int res, pos, version;
	ino_t fid, ino = dir->i_ino;
	unsigned int len;
	struct hfs_brec *entry;
	char fixed_name[HFS_NAMEMAX+1];
	
	if (!dir || !dir->i_sb || !S_ISDIR(dir->i_mode)) {
		return -EBADF;
	}

	res = 0;
	while (!res && (filp->f_pos < dir->i_size)) {
		pos = filp->f_pos;
		if (!pos) {
			/* Entry 0 is for "." */
			res = filldir(dirent, ".", 1, 0, ino);
		} else if (pos==1) {
			/* Entry 1 for ".." */
			res = filldir(dirent, "..", 2, 1, PARENT(dir));
		} else {
			/* There are two Linux directory entries for
			   each HFS file or directory.	The first is
			   the actual file or directory and the second
			   is its header file.	We can simply divide
			   the f_pos by two to get the index of the
			   directory entry.
			   
			   The call to filldir() may involve transfers
			   to user-space which can cause a call to
			   schedule().	Therefore before we proceed,
			   we check that nobody changed the directory.
			   If the directory has been changed, then we
			   try again. */
			do {
				version = dir->i_version;
				if ((res=hfs_cat_nth(dir, &entry, pos>>1,
						     HFS_SKIP_NONE))) {
					return res;
				}
				fid = HFS_FID(entry->data.raw_inode);
				/* The following two statements
				   prepend '%' to the names of header
				   files during name-mangling, while
				   for regular entries the name-
				   mangling takes place normally. */
				fixed_name[0] = '%';
				len = (pos & 1) +
					NAMEIN(dir, fixed_name + (pos & 1),
					       entry->key->u.ckr.CName);
				hfs_brec_relse(&entry);
				res = filldir(dirent, fixed_name, len,
					      pos, fid);
			} while (version != dir->i_version);
		}
		if (!res) {
			++filp->f_pos;
		}
	}
	return 0;
}

/*
 * hfs_dbl_create()
 *
 * This is the create() entry in the file_operations structure for
 * HFS directories in the AppleDouble scheme.  The purpose is to
 * create a new file in a directory, given the inode of the directory,
 * the name (and its length) for the new file and its mode.
 * On success *result points to an inode for the new file and 0 is returned.
 * On failure *result is NULL and an error code is returned.
 */
int hfs_dbl_create(struct inode * dir, const char * name, int len, int mode,
		   struct inode ** result)
{
	int error;

	*result = NULL;
	if (!dir) {
		return -ENOENT;
	}

	/* XXX should try to determine when a header file was intended */

	error = hfs_do_create(dir, name, len, mode, result);
	if (error) {
		iput(dir);
		return error;
	}

	dir->i_size += 2;
	++HFS_I(dir)->u.dir.files;
	dir->i_dirt = 1;
	iput(dir);

	return 0;
}

/*
 * hfs_dbl_mkdir()
 *
 * This is the mkdir() entry in the file_operations structure for
 * HFS directories in the AppleDouble scheme.  The purpose is to
 * create a new directory, given the inode of the parent directory,
 * the name (and its length) for the new directory and its mode.
 * On success 0 is returned.
 * On failure an error code is returned.
 */
int hfs_dbl_mkdir(struct inode * dir, const char * name, int len, int mode)
{
        HFS_CNAME tmp_name;
        int error, i;

        if (!dir) {
                return -ENOENT;
        }

        /* mangle the name and check against reserved names */
        tmp_name[0] = NAMEOUT(dir, tmp_name+1, len, name);
        for (i=0; i<RNAMES; ++i) {
                if (!hfs_strcmp(tmp_name, dbl_reserved_names[i])) {
                        iput(dir);
                        return -EEXIST;
                }
        }

	/* XXX need to check for names clashing with header files */

        error = hfs_do_mkdir(dir, tmp_name);
        if (error) {
                iput(dir);
                return error;
        }

        dir->i_size += 2;
        ++HFS_I(dir)->u.dir.dirs;
        dir->i_dirt = 1;
        iput(dir);

        return 0;
}
