/*
 * linux/fs/hfs/extent.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the functions related to the extents B-tree.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/fs.h>
#include <linux/locks.h>

/*================ File-local functions ================*/

/*
 * dump_ext()
 *
 * prints the content of a extent for debugging purposes.
 */
#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
static void dump_ext(const char *msg, const struct hfs_extent *e) {
	int i, st;

	printk("%s ", msg);
	if (e) {
		st = e->start;
		for (i=0; (i<3) && e->length[i]; ++i) {
			printk("(%d-%d) ", st, st+e->length[i]-1);
			st += e->length[i];
		}
	} else {
		printk("NULL");
	}
	printk("\n");
}
#else
#define dump_ext(A,B) {}
#endif

/*
 * read_extent()
 * 
 * Initializes a (struct hfs_extent) from a (struct hfs_raw_exent) and
 * the number of the starting block for the extent.
 *
 * Note that the callers must check that to,from != NULL
 */
static void read_extent(struct hfs_extent *to,
			const struct hfs_raw_extent *from,
			__u16 start)
{
	to->start = start;
	to->block[0]  = from_w(from->block1);
	to->length[0] = from_w(from->length1);
	to->block[1]  = from_w(from->block2);
	to->length[1] = from_w(from->length2);
	to->block[2]  = from_w(from->block3);
	to->length[2] = from_w(from->length3);
	to->end = start + to->length[0] + to->length[1] + to->length[2] - 1;
	to->next = to->prev = NULL;
	to->count = 0;
}

/*
 * write_extent()
 * 
 * Initializes a (struct hfs_raw_extent) from a (struct hfs_exent).
 *
 * Note that the callers must check that to,from != NULL
 */
static void write_extent(struct hfs_raw_extent *to,
			 const struct hfs_extent *from)
{
	to->block1  = to_w(from->block[0]);
	to->length1 = to_w(from->length[0]);
	to->block2  = to_w(from->block[1]);
	to->length2 = to_w(from->length[1]);
	to->block3  = to_w(from->block[2]);
	to->length3 = to_w(from->length[2]);
}

/*
 * decode_extent()
 *
 * Given an extent record and allocation block offset into the file,
 * return the number of the corresponding allocation block on disk,
 * or -1 if the desired block is not mapped by the given extent.
 *
 * Note that callers must check that extent != NULL
 */
static int decode_extent(const struct hfs_extent * extent, int block)
{
	if (!extent || (block < extent->start) || (block > extent->end)) {
		return -1;
	}
	block -= extent->start;
	if (block < extent->length[0]) {
		return block + extent->block[0];
	}
	block -= extent->length[0];
	if (block < extent->length[1]) {
		return block + extent->block[1];
	}
	return block + extent->block[2] - extent->length[1];
}

/*
 * relse_ext()
 *
 * Reduce the reference count of an in-core extent record by one,
 * removing it from memory if the count falls to zero.
 */
static void relse_ext(struct hfs_extent *ext)
{
	if (--ext->count || !ext->start) {
		return;
	}
	ext->prev->next = ext->next;
	if (ext->next) {
		ext->next->prev = ext->prev;
	}
	KFREE_S(ext, sizeof(*ext));
}

/*
 * find_ext()
 *
 * Given a pointer to a (struct hfs_file) and an allocation block
 * number in the file, find the extent record containing that block.
 * Returns a pointer to the extent record on success or NULL on failure.
 * The 'cache' field of 'fil' also points to the extent so it has a
 * reference count of at least 2.
 *
 * Callers must check that fil != NULL
 */
static struct hfs_extent * find_ext(struct hfs_file *fil, int alloc_block)
{
	struct hfs_btree *tr= HFS_SB(fil->inode->i_sb)->s_ext_tree;
	struct hfs_brec_key target;
	struct hfs_brec *entry;
	struct hfs_extent *ext, *ptr;
	__u32 fid = HFS_IFID(fil->inode->i_ino);
	int tmp;

	if (alloc_block < 0) {
		ext = &fil->first;
		goto found;
	}

	ext = fil->cache;
	if (!ext || (alloc_block < ext->start)) {
		ext = &fil->first;
	}
	while (ext->next && (alloc_block > ext->end)) {
		ext = ext->next;
	}
	if ((alloc_block <= ext->end) && (alloc_block >= ext->start)) {
		goto found;
	}

	/* time to read more extents */
	ext = KMALLOC(sizeof(*ext), GFP_KERNEL);
	if (!ext) {
		goto bail3;
	}

	/*target.KeyLen = ???;	Not needed */
	target.u.xkr.FkType = fil->fork;
	target.u.xkr.FNum = to_l(fid);
	target.u.xkr.FABN = to_w(alloc_block);

	tmp = hfs_bfind(tr, &target, &entry, HFS_BFIND_READ_LE);
	if (tmp < 0) {
		goto bail2;
	}

	if ((entry->key->u.xkr.FNum != target.u.xkr.FNum) ||
	    (entry->key->u.xkr.FkType != fil->fork)) {
		goto bail1;
	}
		
	read_extent(ext, entry->data.raw_extent,
		    from_w(entry->key->u.xkr.FABN));
	hfs_brec_relse(&entry);

	if ((alloc_block > ext->end) && (alloc_block < ext->start)) {
		/* something strange happened */
		goto bail2;
	}

	ptr = fil->cache;
	if (!ptr || (alloc_block < ptr->start)) {
		ptr = &fil->first;
	}
	while (ptr->next && (alloc_block > ptr->end)) {
		ptr = ptr->next;
	}
	if (ext->start == ptr->start) {
		/* somebody beat us to it. */
		KFREE_S(ext, sizeof(*ext));
		ext = ptr;
	} else if (ext->start < ptr->start) {
		/* insert just before ptr */
		ptr->prev->next = ext;
		ext->prev = ptr->prev;
		ext->next = ptr;
		ptr->prev = ext;
	} else {
		/* insert at end */
		ptr->next = ext;
		ext->prev = ptr;
	}
 found:
	ext->count += 2; /* one for return value, one for cache */
	relse_ext(fil->cache);
	fil->cache = ext;
	return ext;

 bail1:
	hfs_brec_relse(&entry);
 bail2:
	KFREE_S(ext, sizeof(*ext));
 bail3:
	return NULL;
}

/*
 * new_extent()
 *
 * Description:
 *   Adds a new extent to a file, extending its physical length.
 * Input Variable(s):
 *   struct hfs_file *fil: the file to extend
 *   struct hfs_extent *ext: the current last extent for 'fil'
 *   __u16 ablock: the number of allocation blocks in 'fil'.
 *   __u16 start: first allocation block to add to 'fil'.
 *   __u16 len: the number of allocation blocks to add to 'fil'.
 *   __u16 ablksz: number of sectors in an allocation block.
 * Output Variable(s):
 *   NONE
 * Returns:
 *   (struct hfs_extent *) the new extent or NULL
 * Preconditions:
 *   'fil' points to a valid (struct hfs_file)
 *   'ext' point to a valid (struct hfs_extent) which is the last in 'fil'
 *   'ablock', 'start', 'len' and 'ablksz' are wht they claim to be.
 * Postconditions:
 *   If NULL is returned then no changes have been made to 'fil.
 *   If the retuen value is non-NULL that it is the extent that has been
 *   added to 'fil' both in memory and on disk.  The 'i_blocks' field of
 *   'fil->inode' has been updated to refect the new physical size.
 * 
 * XXX: currently atomic with respect to itself only because of the
 * locking taking place during writing.  Once truncation is possible
 * the end of the file might move during I/O.  Then it might make sense
 * to inline some or all of this function into find_ext().
 */
static struct hfs_extent *new_extent(struct hfs_file *fil,
				     struct hfs_extent *ext,
				     __u16 ablock, __u16 start, __u16 len,
				     __u16 ablksz)
{
	struct hfs_raw_extent raw;
	struct hfs_brec_key key;
	int error;

	if (fil->inode->i_ino == HFS_EXT_FID) {
		/* Limit extents tree to the record in the MDB */
		return NULL;
	}

	if (!(ext->next = KMALLOC(sizeof(*ext), GFP_KERNEL))) {
		return NULL;
	}
	ext->next->prev = ext;
	ext->next->next = NULL;
	ext = ext->next;
	relse_ext(ext->prev);

	ext->start = ablock;
	ext->block[0] = start;
	ext->length[0] = len;
        ext->block[1] = 0;
	ext->length[1] = 0;
	ext->block[2] = 0;
	ext->length[2] = 0;
	ext->end = ablock + len - 1;
	ext->count = 1;

	write_extent(&raw, ext);
	
	key.KeyLen = 7;
	key.u.xkr.FkType = fil->fork;
	key.u.xkr.FNum = to_l(HFS_IFID(fil->inode->i_ino));
	key.u.xkr.FABN = to_w(ablock);

	error = hfs_binsert(HFS_SB(fil->inode->i_sb)->s_ext_tree, 
			    &key, &raw, sizeof(raw));
	if (error) {
		ext->prev->next = NULL;
		KFREE_S(ext, sizeof(*ext));
		return NULL;
	}
	++ext->count;
	relse_ext(fil->cache);
	fil->cache = ext;
	return ext;
}

/*
 * update_ext()
 *
 * Given a (struct hfs_file) write an extent record back to disk.
 */
static void update_ext(struct hfs_file *fil, struct hfs_extent *ext)
{
	struct hfs_brec_key target;
	struct hfs_brec *entry;

	if (ext->start) {
		/*target.KeyLen = ???;	Not needed */
		target.u.xkr.FkType = fil->fork;
		target.u.xkr.FNum = to_l(HFS_IFID(fil->inode->i_ino));
		target.u.xkr.FABN = to_w(ext->start);

		if (!hfs_bfind(HFS_SB(fil->inode->i_sb)->s_ext_tree,
			       &target, &entry, HFS_BFIND_WRITE)) {
			write_extent(entry->data.raw_extent, ext);
			hfs_brec_relse(&entry);
		}
	}
}

/*
 * zero_blocks()
 * 
 * Zeros-out 'num' alocation blocks beginning with 'start'.
 */
static int zero_blocks(struct super_block *sb, int start, int num) {
	struct buffer_head *bh;
	int end;
	int j;

	start = HFS_SB(sb)->s_fs_start + start * HFS_SB(sb)->s_alloc_blksz;
	end = start + num * HFS_SB(sb)->s_alloc_blksz;

	for (j=start; j<end; ++j) {
        	if (!(bh = getblk(sb->s_dev,j,HFS_SECTOR_SIZE))) {
                	printk("hfs_getblk: cannot get block");
                	return 1;
        	}
        	memset(bh->b_data, 0, HFS_SECTOR_SIZE);
        	mark_buffer_uptodate(bh, 1);
        	mark_buffer_dirty(bh, 1);
        	brelse(bh);
	}
	return 0;
}

/*================ Global functions ================*/

/*
 * hfs_file_extend()
 *
 * Try to add enough allocation blocks to 'fil' so that the total
 * allocation for 'fil' is 'ablock' + 1.
 *
 * XXX: don't yet use clumpsize
 * XXX: should check to see if enough blocks are available before starting
 * XXX: not yet totally race free until new_extent() is redone, but that
 * is not a problem right now since the only way to get here currently is
 * from hfs_do_write() which is called with the inode locked.  new_extent()
 * will need to be fixed before truncation can work.
 */
void hfs_file_extend(struct hfs_file *fil, int ablock)
{
	struct hfs_extent *ext;
	struct super_block *sb = fil->inode->i_sb;
	int i, start, err;
	__u16 need, len=0;
	__u16 ablksz = HFS_SB(sb)->s_alloc_blksz;
	__u32 blocks;

	++ablock;
restart:
	blocks = fil->inode->i_blocks;
	need = ablock - blocks/ablksz;
	if (need < 1) {
		return;
	}

	/* find last extent record and try to extend it */
	if (!(ext = find_ext(fil, blocks/ablksz - 1))) {
		/* somehow we couldn't find the end of the file! */
		return;
	}

	/* determine which is the last used extent in the record */
	for (i=2; (i>=0) && !ext->length[i]; --i) {};
	if (i<0) {
		/* empty extents can't be extended */
		relse_ext(ext);
	} else {
		/* try to extend the last extent */
		start = ext->block[i] + ext->length[i];
		relse_ext(ext);

		err = 0;
		lock_super(sb);
		if ((len = hfs_vbm_count_free(sb, start))) {
			if (need < len) {
				len = need;
			}
			err = hfs_set_vbm_bits(sb, start, len);
		}
		unlock_super(sb);
		if (err) {
			return;
		}
	
		if (len) {
			zero_blocks(sb, start, len);
	
			/* find last extent record */
			if (!(ext = find_ext(fil, blocks/ablksz - 1))) {
				/* somehow we couldn't find end of the file! */
				hfs_clear_vbm_bits(sb, start, len);
				return;
			}
	
			/* determine which is the last used extent */
			for (i=2; (i>=0) && !ext->length[i]; --i) {};

			/* check that nothing important changed during I/O */
			if ((blocks != fil->inode->i_blocks) || (i<0) ||
			    (start != ext->block[i] + ext->length[i])) {
				relse_ext(ext);
				hfs_clear_vbm_bits(sb, start, len);
				goto restart;
			}
		
			ext->length[i] += len;
			ext->end += len;
			blocks = (fil->inode->i_blocks += len * ablksz);
			need -= len;
			update_ext(fil, ext);
			/* ++ext->count; */
			relse_ext(fil->cache);
			fil->cache = ext;
			/* relse_ext(ext); */
		}
	}

	/* add some more extents */
	while (need) {
		len = need;
		err = 0;
		lock_super(sb);
		start = hfs_vbm_search_free(sb, &len);
		if (len) {
			if (need < len) {
				len = need;
			}
			err = hfs_set_vbm_bits(sb, start, len);
		}
		unlock_super(sb);
		if (!len || err) {
			return;
		}
		zero_blocks(sb, start, len);

		/* find last extent record */
		if (!(ext = find_ext(fil, blocks/ablksz - 1))) {
			/* somehow we couldn't find the end of the file! */
			hfs_clear_vbm_bits(sb, start, len);
			return;
		}

		/* check that physical file length didn't change during I/O */
		if (blocks != fil->inode->i_blocks) {
			relse_ext(ext);
			hfs_clear_vbm_bits(sb, start, len);
			goto restart;
		}

		/* determine which is the first free extent in the record */
		for (i=0; (i<3) && ext->length[i]; ++i) {};
		if (i < 3) {
			ext->block[i] = start;
			ext->length[i] = len;
			ext->end += len;
			update_ext(fil, ext);
		} else {
			if (!(ext = new_extent(fil, ext, blocks/ablksz,
					       start, len, ablksz))) {
				hfs_clear_vbm_bits(sb, start, len);
				return;
			}
		}
		blocks = (fil->inode->i_blocks += len * ablksz);
		need -= len;
		/* ++ext->count; */
		relse_ext(fil->cache);
		fil->cache = ext;
		/* relse_ext(ext); */
	}
	return;
}

/*
 * hfs_block_map()
 *
 * Given an hfs_file and a block number within the file, return the
 * number of the corresponding physical block on disk, or zero on
 * error.
 */
int hfs_block_map(struct hfs_file *fil, int block, int create) 
{
	int ablksz, ablock, offset, tmp;
	struct hfs_extent *ext;

	if (!fil || !fil->inode->i_sb) {
		return 0;
	}

#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
	printk("hfs_block_map: ablock %d of file %d, fork %d\n",
	       block, HFS_IFID(fil->inode->i_ino), fil->fork);
#endif

	if (block < 0) {
		printk("hfs_block_map: block < 0\n");
		return 0;
	}
	ablksz = HFS_SB(fil->inode->i_sb)->s_alloc_blksz;
	offset = HFS_SB(fil->inode->i_sb)->s_fs_start + (block % ablksz);
	ablock = block / ablksz;
	if (ablock > HFS_USHRT_MAX) {
		printk("hfs_block_map: block > big; fid=%ld fork=%d\n",
			HFS_IFID(fil->inode->i_ino), fil->fork);
		return 0;
	}
	
	if (block >= fil->inode->i_blocks) {
		if (create) {
			hfs_file_extend(fil, ablock);
		} else {
			return 0;
		}
	}

#if defined(DEBUG_EXTENTS) || defined(DEBUG_ALL)
	printk("(lblock %d offset %d)\n", ablock, offset);
#endif

	if ((ext = find_ext(fil, ablock))) {
		dump_ext("trying new: ", ext);
		tmp = decode_extent(ext, ablock);
		relse_ext(ext);
		if (tmp >= 0) {
			return tmp*ablksz + offset;
		}
	} 

	return 0;
}

/*
 * hfs_extent_out()
 *
 * Copy the first extent record from a (struct hfs_file) to a (struct
 * hfs_raw_extent), record (normaly the one in the catalog entry).
 */
void hfs_extent_out(const struct hfs_file *fil, struct hfs_raw_extent *ext)
{
	if (fil && ext) {
		write_extent(ext, &fil->first);
		dump_ext("extent out: ", &fil->first);
	}
}

/*
 * hfs_extent_in()
 *
 * Copy an hfs_raw_extent to the 'first' and 'cache' fields of an hfs_file.
 */
void hfs_extent_in(struct hfs_file *fil, const struct hfs_raw_extent *ext)
{
	if (fil && ext) {
		read_extent(&fil->first, ext, 0);
		fil->cache = &fil->first;
		fil->first.count = 2;
		dump_ext("extent in: ", &fil->first);
	}
}

/* 
 * hfs_extent_trim()
 *
 * Removes from memory all extents after 'ext' associated with 'fil'.
 */
void hfs_extent_trim(struct hfs_file *fil, struct hfs_extent *ext)
{
	struct hfs_extent *tmp;
	
	if ((fil->cache) && (fil->cache->start > ext->end)) {
		++ext->count;
		fil->cache = ext;
	}

	tmp = ext->next;
	ext->next = NULL;
	while ((ext = tmp)) {
		tmp = ext->next;
		KFREE_S(ext, sizeof(*ext));
	}
}
