/*
 * linux/fs/hfs/file.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the file-related functions which are independent of
 * which scheme is being used to represent forks.
 *
 * Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/sched.h>
#include <linux/locks.h>
#include <linux/malloc.h>
#include <linux/stddef.h>
#include <linux/fcntl.h>
#include <linux/mm.h>
#include <asm/segment.h>
#include <linux/pagemap.h>

/*================ Forward declarations ================*/
static int hfs_file_read(struct inode *, struct file *, char *, int);
static int hfs_file_write(struct inode *, struct file *, const char *, int);
static int hfs_bmap(struct inode *, int);

/*================ Global variables ================*/

static struct file_operations hfs_file_operations = {
	NULL,			/* lseek - default */
	hfs_file_read,		/* read */
	hfs_file_write,		/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	generic_file_mmap,	/* mmap */
	NULL,			/* no special open is needed */
	NULL,			/* release */
	NULL			/* fsync */
};

struct inode_operations hfs_file_inode_operations = {
	&hfs_file_operations,	/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	generic_readpage,
	NULL,
	hfs_bmap,		/* bmap */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

/*================ Variable-like macros ================*/

/* maximum number of blocks to try to read in at once */
#define NBUF 32

/*================ File-local functions ================*/

/*
 * hfs_bmap()
 *
 * This is the bmap() field in the inode_opertions structure for
 * "regular" (non-header) files.  The purpose is to translate an inode
 * and a block number within the corresponding file into a physical
 * block number.  This function just calls hfs_block_map() to do the
 * real work.
 *
 * XXX: Note the reason that there is not bmap() for AppleDouble
 * header files is that the first HFS_DBL_HDR_LEN bytes don't have any
 * corresponding allocation on disk.  I imagine that I could just
 * return 0 for those blocks of the file, but I am not certain about
 * what consequences that might have elsewhere.	 If it actually worked
 * then generic_file_mmap() could be used to do mmap() on any part of
 * an AppleDouble header file that excluded the first HFS_HRD_LEN bytes.
 * I should really look into this.  Of course I should eventually
 * write a complete mmap() for AppleDouble header files anyway.
 */
static int hfs_bmap(struct inode * inode, int block)
{
	return hfs_block_map(&HFS_I(inode)->u.file, block, 0);
}

/*
 * hfs_file_read()
 *
 * This is the read field in the inode_operations structure for
 * "regular" (non-header) files.  The purpose is to transfer upto
 * 'count' bytes from the file corresponding to 'inode', begining at
 * 'filp->offset' bytes into the file.	The data is trasfered to
 * user-space at the address 'buf'.  Returns the number of bytes
 * sucessfully transfered.  This function checks the arguments, does
 * some setup and then calls hfs_do_read() to do the actual transfer.
 */
static int hfs_file_read(struct inode * inode, struct file * filp,
			 char * buf, int count)
{
	int left, read;
	off_t pos, size;

	if (!inode) {
		printk("hfs_file_read: inode = NULL\n");
		return -EINVAL;
	}
	if (!S_ISREG(inode->i_mode)) {
		printk("hfs_file_read: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}
	pos = filp->f_pos;
	size = inode->i_size;
	if (pos > size) {
		left = 0;
	} else {
		left = size - pos;
	}
	if (left > count) {
		left = count;
	}
	if (left <= 0) {
		return 0;
	}
	read = hfs_do_read(&HFS_I(inode)->u.file, pos, buf, left, &filp->f_reada);
	if (read > 0) {
		filp->f_pos += read;
	}
	return read;
}

static int hfs_file_write(struct inode * inode, struct file * filp,
			  const char * buf, int count)
{
	int written;
	off_t pos;

	if (!inode) {
		printk("hfs_file_write: inode = NULL\n");
		return -EINVAL;
	}
	if (!S_ISREG(inode->i_mode)) {
		printk("hfs_file_write: mode = %07o\n", inode->i_mode);
		return -EINVAL;
	}
	if (filp->f_flags & O_APPEND) {
		filp->f_pos = inode->i_size;
	}
	pos = filp->f_pos;
	written = hfs_do_write(&HFS_I(inode)->u.file, pos, buf, count);
	if (written > 0) {
		filp->f_pos += written;
	}
	if (filp->f_pos > inode->i_size) {
		HFS_I(inode)->u.file.size += (filp->f_pos - inode->i_size);
		inode->i_size = filp->f_pos;
	}
	return written;
}

/*================ Global functions ================*/

/*
 * hfs_getblk()
 *
 * Given an hfs_file and a block number return the buffer_head for
 * that block from the file.  If 'create' is non-zero then allocate
 * the necessary block(s) to the file.
 */
struct buffer_head * hfs_getblk(struct hfs_file *fil, int block, int create)
{
	int tmp;
	struct buffer_head *bh = NULL;

	do {
		brelse(bh);

		tmp = hfs_block_map(fil, block, create);
		if (!tmp) {
			return NULL;
		}

		bh = getblk(fil->dev, tmp, HFS_SECTOR_SIZE);

	} while (tmp != hfs_block_map(fil, block, 0));
	return bh;
}

/*
 * hfs_do_read()
 *
 * This function transfers actual data from disk to user-space memory,
 * returning the number of bytes succesfully transfered.  'fil' tells
 * which file on the disk to read from.	 'filp' gives the offset into
 * the Linux file at which to begin the transfer, and is updated to
 * reflect the new position. 'offset' gives the offset into the HFS
 * file at which to begin the trasfer.	Note that this will differ
 * from 'filp->offset' in the case of an AppleDouble header file due
 * to the block of meta-data at the beginning of the file, which has
 * no corresponding place in the HFS file.  'count' tells how many
 * bytes to transfer.  'size' tells how many bytes are actually in the
 * HFS file.  'buf' gives an address in user-space to transfer the
 * data to.
 * 
 * This is just the "guts" of Linus's minix_file_read with
 * "minix"->"hfs" and "BLOCK_SIZE"->"HFS_SECTOR_SIZE" everywhere, and
 * a few other minor changes to accept a (struct hfs_file) as the
 * first argument.
 */
int hfs_do_read(struct hfs_file * fil, off_t pos, char * buf, int count,
		off_t * reada)
{
	int size;
	int read, chars, offset;
	int block, blocks;
	int bhrequest, uptodate;
	struct buffer_head ** bhb, ** bhe;
	struct buffer_head * bhreq[NBUF];
	struct buffer_head * buflist[NBUF];

	read = 0;

	/* split 'pos' in to block and (byte) offset components */
	block = pos >> HFS_SECTOR_SIZE_BITS;
	offset = pos & (HFS_SECTOR_SIZE-1);

	/* compute the physical size of the file in blocks */
	size = (fil->size + (HFS_SECTOR_SIZE-1)) >> HFS_SECTOR_SIZE_BITS;

	/* compute the number of physical blocks to be transferred */
	blocks = (count+offset+HFS_SECTOR_SIZE-1) >> HFS_SECTOR_SIZE_BITS;

	bhb = bhe = buflist;
	if (*reada) {
		if (blocks < read_ahead[MAJOR(fil->dev)] /
		    (HFS_SECTOR_SIZE >> 9)) {
			blocks = read_ahead[MAJOR(fil->dev)] /
				(HFS_SECTOR_SIZE >> 9);
		}
		if (block + blocks > size) {
			blocks = size - block;
		}
	}

	/* We do this in a two stage process.  We first try and
	   request as many blocks as we can, then we wait for the
	   first one to complete, and then we try and wrap up as many
	   as are actually done.  This routine is rather generic, in
	   that it can be used in a filesystem by substituting the
	   appropriate function in for getblk.
	   
	   This routine is optimized to make maximum use of the
	   various buffers and caches. */

	do {
		bhrequest = 0;
		uptodate = 1;
		while (blocks) {
			--blocks;
			*bhb = hfs_getblk(fil, block++, 0);
			if (*bhb && !buffer_uptodate(*bhb)) {
				uptodate = 0;
				bhreq[bhrequest++] = *bhb;
			}

			if (++bhb == &buflist[NBUF]) {
				bhb = buflist;
			}

			/* If the block we have on hand is uptodate,
			   go ahead and complete processing. */
			if (uptodate) {
				break;
			}
			if (bhb == bhe) {
				break;
			}
		}

		/* Now request them all */
		if (bhrequest) {
			ll_rw_block(READ, bhrequest, bhreq);
		}

		do {  /* Finish off all I/O that has actually completed */
			if (*bhe) {
				wait_on_buffer(*bhe);
				if (!buffer_uptodate(*bhe)) { /* read error? */
					brelse(*bhe);
					if (++bhe == &buflist[NBUF]) {
						bhe = buflist;
					}
					count = 0;
					break;
				}
			}
			if (count < HFS_SECTOR_SIZE - offset) {
				chars = count;
			}
			else
				chars = HFS_SECTOR_SIZE - offset;
			count -= chars;
			read += chars;
			if (*bhe) {
				memcpy_tofs(buf, offset+(*bhe)->b_data, chars);
				brelse(*bhe);
				buf += chars;
			} else {
				while (chars-->0) {
					put_fs_byte(0,buf++);
				}
			}
			offset = 0;
			if (++bhe == &buflist[NBUF]) {
				bhe = buflist;
			}
		} while (count > 0 && bhe != bhb &&
			 (!*bhe || !buffer_locked(*bhe)));
	} while (count > 0);

	/* Release the read-ahead blocks */
	while (bhe != bhb) {
		brelse(*bhe);
		if (++bhe == &buflist[NBUF]) {
			bhe = buflist;
		}
	};
	if (!read) {
		return -EIO;
	}
	*reada = 1;
	fil->inode->i_atime = CURRENT_TIME;
	return read;
}
 
int hfs_do_write(struct hfs_file * fil, off_t pos, const char * buf, int count)
{
	int written, c;
	struct buffer_head * bh;
	char * p;

	written = 0;
	while (written < count) {
		bh = hfs_getblk(fil, pos/HFS_SECTOR_SIZE, 1);
		if (!bh) {
			if (!written) {
				written = -ENOSPC;
			}
			break;
		}
		c = HFS_SECTOR_SIZE - (pos % HFS_SECTOR_SIZE);
		if (c > count - written) {
			c = count - written;
		}
		if (c != HFS_SECTOR_SIZE && !buffer_uptodate(bh)) {
			ll_rw_block(READ, 1, &bh);
			wait_on_buffer(bh);
			if (!buffer_uptodate(bh)) {
				brelse(bh);
				if (!written) {
					written = -EIO;
				}
				break;
			}
		}
		p = (pos % HFS_SECTOR_SIZE) + bh->b_data;
		memcpy_fromfs(p, buf, c);
		update_vm_cache(fil->inode,pos,p,c);
		pos += c;
		written += c;
		buf += c;
		mark_buffer_uptodate(bh, 1);
		mark_buffer_dirty(bh, 0);
		brelse(bh);
	}
	fil->inode->i_mtime = fil->inode->i_ctime = CURRENT_TIME;
	fil->inode->i_dirt = 1;
	return written;
}
