/*
 * linux/fs/hfs/file_cap.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the file_ops and inode_ops for the meta-data
 * files under the CAP representation.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <asm/segment.h>

/*================ Forward declarations ================*/

static int cap_info_read(struct inode *, struct file *, char *, int);
static int cap_open(struct inode *, struct file *);

/*================ Global variables ================*/

static struct file_operations hfs_cap_info_operations = {
	NULL,			/* lseek - default */
	cap_info_read,		/* read */
	NULL,			/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap - not yet */
	cap_open,		/* open */
	NULL,			/* release */
	NULL			/* fsync */
};

struct inode_operations hfs_cap_info_inode_operations = {
	&hfs_cap_info_operations,	/* default file operations */
	NULL,				/* create */
	NULL,				/* lookup */
	NULL,				/* link */
	NULL,				/* unlink */
	NULL,				/* symlink */
	NULL,				/* mkdir */
	NULL,				/* rmdir */
	NULL,				/* mknod */
	NULL,				/* rename */
	NULL,				/* readlink */
	NULL,				/* follow_link */
	NULL,				/* bmap - none */
	NULL,				/* truncate */
	NULL,				/* permission */
	NULL				/* smap */
};

/*================ Variable-like macros ================*/

/* to keep some of the code from getting to verbose */
#define RAW	(entry->data.raw_inode)

/*================ Function-like macros ================*/

/* Given an inode produce the catalog key of the correspoding HFS file */
#define KEY(i) ((struct hfs_brec_key *)(HFS_I(i)->key))

/*================ File-local functions ================*/

/*
 * cap_info_read()
 *
 * This is the read() entry in the file_operations structure for CAP
 * meta-data files.  The purpose is to transfer upto 'count' bytes
 * from the file corresponding to 'inode' beginnning at offset
 * 'file->f_pos' to user-space at the address 'buf'.  The return value
 * is the number of bytes actually trasferred.
 */
static int cap_info_read(struct inode *inode, struct file *file,
			char *buf, int count)
{
	off_t pos = file->f_pos;
	off_t size = inode->i_size;
	char * ptr;

	if (count < 0) {
		return -EINVAL;
	}
	if (!(ptr = (char *)(HFS_I(inode)->u.cap_info.meta))) {
		printk("HFS: NULL finderinfo structure\n");
		return -EIO;
	}
	if (pos >= size) {
		return 0;
	}
	if (count > size - pos) {
		count = size - pos;
	}

	memcpy_tofs(buf, ptr+pos, count);
	file->f_pos += count;
	return count;
}

/*
 * cap_open()
 *
 * This is the open() field in the file_operations structure for CAP
 * info files.	It is called when the file is opened and reads in the
 * information from the catalog entry if neccessary.
 */
static int cap_open(struct inode * inode, struct file *filp)
{
	struct hfs_brec *entry;
	struct hfs_cap_info *res;
	HFS_BYTE *info1, *info2;

	if (HFS_I(inode)->u.cap_info.meta) {
		return 0;
	}

	if (hfs_cat_find(inode->i_sb, KEY(inode), &entry, 0)) {
		return 1;
	}
	if (!(res = KMALLOC(sizeof(struct hfs_cap_info),GFP_KERNEL))) {
		hfs_brec_relse(&entry);
		return 1;
	}
	memset(res, 0, sizeof(struct hfs_cap_info));
	if (RAW->cdrType==HFS_CDR_FIL) {
		info1 = RAW->u.fil.UsrWds;
		info2 = RAW->u.fil.FndrInfo;
	} else {
		info1 = RAW->u.dir.UsrInfo;
		info2 = RAW->u.dir.FndrInfo;
	}
	memcpy(res->fi_fndr, info1, 16);
	memcpy(res->fi_fndr+16, info2, 16);
	res->fi_magic1 = HFS_CAP_MAGIC1;
	res->fi_version = HFS_CAP_VERSION;
	res->fi_magic = HFS_CAP_MAGIC;
	res->fi_bitmap = 0x02;
	memcpy(res->fi_macfilename, entry->key->u.ckr.CName+1,
	       entry->key->u.ckr.CName[0]);
	HFS_I(inode)->u.cap_info.meta = res;
	hfs_brec_relse(&entry);
	return 0;
}

