/*
 * linux/fs/hfs/file_nat.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the file_ops and inode_ops for the meta-data
 * files under the Netatalk representation.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/stat.h>
#include <linux/malloc.h>
#include <linux/stddef.h>
#include <linux/mm.h>
#include <asm/segment.h>

/*================ Forward declarations ================*/
static int nat_hdr_read(struct inode *, struct file *, char *, int);
static int nat_open(struct inode *, struct file *);

/*================ Global variables ================*/

static struct file_operations hfs_nat_hdr_operations = {
	NULL,			/* lseek - default */
	nat_hdr_read,		/* read */
	NULL,			/* write */
	NULL,			/* readdir - bad */
	NULL,			/* select - default */
	NULL,			/* ioctl - default */
	NULL,			/* mmap - XXX: not yet since default
				   relies upon bmap() */
	nat_open,		/* open */
	NULL,			/* release */
	NULL			/* fsync */
};

struct inode_operations hfs_nat_hdr_inode_operations = {
	&hfs_nat_hdr_operations,/* default file operations */
	NULL,			/* create */
	NULL,			/* lookup */
	NULL,			/* link */
	NULL,			/* unlink */
	NULL,			/* symlink */
	NULL,			/* mkdir */
	NULL,			/* rmdir */
	NULL,			/* mknod */
	NULL,			/* rename */
	NULL,			/* readlink */
	NULL,			/* follow_link */
	NULL,			/* bmap - XXX: not avaiable since
				   header part has no disk block */
	NULL,			/* truncate */
	NULL,			/* permission */
	NULL			/* smap */
};

/*================ Variable-like macros ================*/

/* to keep some of the code from getting to verbose */
#define RAW	(entry->data.raw_inode)

/*================ Function-like macros ================*/

/* get catalog tree key for a given inode */
#define KEY(i) ((struct hfs_brec_key *)(HFS_I(i)->key))

/*================ File-local variables ================*/

/*
 * When a AppleDouble header is created for a Netatalk file this template is
 * used to initialize the descriptor table.
 */
static const struct hfs_dbl_descr nat_hdr_template[5] = {
	/* descriptor for resource fork: */
	{c_to_l(2),
	 c_to_l(HFS_NAT_HDR_LEN),
	 c_to_l(0)},
	/* descriptor for original filename: */
	{c_to_l(3),
	 c_to_l(offsetof(struct hfs_nat_hdr, real_name)),
	 c_to_l(0)},
	/* descriptor for comment: */
	{c_to_l(4),
	 c_to_l(offsetof(struct hfs_nat_hdr, comment)),
	 c_to_l(0)},
	/* descriptor for mac times and file attribute word: */
	{c_to_l(7),
	 c_to_l(offsetof(struct hfs_nat_hdr, create_time)),
	 c_to_l(16)},
	/* descriptor for finder info: */
	{c_to_l(9),
	 c_to_l(offsetof(struct hfs_nat_hdr, finderinfo)),
	 c_to_l(32)},
};

/*================ File-local functions ================*/

/*
 * nat_hdr_read()
 *
 * This is the read field in the inode_operations structure for
 * Netatalk header files.  The purpose is to transfer upto 'count'
 * bytes from the file corresponding to 'inode', begining at
 * 'filp->offset' bytes into the file.	The data is trasfered to
 * user-space at the address 'buf'.  Returns the number of bytes
 * sucessfully transfered.  This function checks the arguments, does
 * some setup, transfers any required portion of the first HFS_DBL_HDR_LEN
 * bytes (which don't actually reside anywhere on disk) and then calls
 * hfs_do_read() to transfer any required data from the rest of the file.
 */
static int nat_hdr_read(struct inode * inode, struct file * filp,
			char * buf, int count)
{
	off_t size, pos, mem_count;
	int left, read;
	struct hfs_inode_info * hfsi;

	if (!inode) {
		printk("hfs_file_read: inode = NULL\n");
		return -EINVAL;
	}
	if (!S_ISREG(inode->i_mode)) {
		printk("hfs_file_read: mode = %07o\n",inode->i_mode);
		return -EINVAL;
	}
	hfsi = HFS_I(inode);
	pos = filp->f_pos;
	size = inode->i_size;
	if (pos > size) {
		left = 0;
	} else {
		left = size - pos;
	}
	if (left > count) {
		left = count;
	}
	if (left <= 0) {
		return 0;
	}

	/* do the header if applicable */
	mem_count = 0;
	if (pos < HFS_NAT_HDR_LEN) {
		mem_count = HFS_NAT_HDR_LEN - pos;
		if (left < mem_count) {
			mem_count = left;
		}
		memcpy_tofs(buf, (char *)hfsi->u.nat_hdr.meta+pos, mem_count);
		filp->f_pos += mem_count;
		buf += mem_count;
		left -= mem_count;
		if (!left) {
			return mem_count;
		}
	}

	pos -= (HFS_NAT_HDR_LEN - mem_count);

	read = hfs_do_read(&hfsi->u.nat_hdr.rsrc, pos, buf, left, &filp->f_reada);
	if (read > 0) {
                filp->f_pos += read;
		return mem_count + read;
        } else {
		return mem_count;
	}
}

/*
 * nat_open()
 *
 * This is the open() field in the file_operations structure for
 * Netatalk header files.  It is called when the file is opened and
 * reads in the information from the catalog entry if neccessary.
 */
static int nat_open(struct inode * inode, struct file *filp)
{
	struct hfs_brec *entry;
	struct hfs_nat_hdr *res;
	HFS_LWORD mtime, ctime, btime;
	HFS_BYTE *info1, *info2;

	if (HFS_I(inode)->u.nat_hdr.meta) {
	        return 0;
	}

	if (hfs_cat_find(inode->i_sb, KEY(inode), &entry, 0)) {
		return 1;
	}
	
	if (!(res = KMALLOC(HFS_NAT_HDR_LEN, GFP_KERNEL))) {
		hfs_brec_relse(&entry);
		return 1;
	}
	
	memset(res, 0x00, HFS_NAT_HDR_LEN);
	res->magic = to_l(HFS_DBL_MAGIC);
	res->version = to_l(HFS_NAT_VERSION);
	res->entries = to_w(5);
	memcpy(res->descr, &nat_hdr_template, sizeof(nat_hdr_template));
	if (RAW->cdrType == HFS_CDR_FIL) {
		res->descr[0].length = RAW->u.fil.RLgLen;
		res->attr = byte_to_word(RAW->u.fil.Flags);
		mtime = RAW->u.fil.MdDat;
		ctime = RAW->u.fil.CrDat;
		btime = RAW->u.fil.BkDat;
		info1 = RAW->u.fil.UsrWds;
		info2 = RAW->u.fil.FndrInfo;
	} else {
		res->attr = RAW->u.dir.Flags;
		mtime = RAW->u.dir.MdDat;
		ctime = RAW->u.dir.CrDat;
		btime = RAW->u.dir.BkDat;
		info1 = RAW->u.dir.UsrInfo;
		info2 = RAW->u.dir.FndrInfo;
	}
	res->create_time = hfs_m_to_htime(ctime);
	res->modify_time = hfs_m_to_htime(mtime);
	res->backup_time = hfs_m_to_htime(btime);
	res->descr[1].length = byte_to_long(entry->key->u.ckr.CName[0]);
	memcpy(res->real_name, entry->key->u.ckr.CName+1,
	       entry->key->u.ckr.CName[0]);
	memcpy(res->finderinfo, info1, 16);
	memcpy(res->finderinfo+16, info2, 16);
	HFS_I(inode)->u.nat_hdr.meta = res;
	hfs_brec_relse(&entry);
	return 0;
}
