/*
 * linux/fs/hfs/hfs_btree.h
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the declarations of the B-tree structures.
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 */

/*================ Variable-like macros ================*/

/* Flags (or'ed together) to bfind(), described with the bfind() function */
#define HFS_BFIND_RDONLY	0x01
#define HFS_BFIND_EXACT		0x02
#define HFS_BFIND_INSERT	0x04
#define HFS_BFIND_DELETE	0x08
#define HFS_BFIND_LOCK		0x10

/* Flags field of the hfs_brec */
#define HFS_BREC_FIRST		0x01
#define HFS_BREC_OVERFLOW	0x02
#define HFS_BREC_UNDERFLOW	0x04
#define HFS_BREC_RIGHT		0x08
#define HFS_BREC_UNINITIALIZED	0x80

/* The stickyness of a (struct hfs_bnode) */
#define HFS_NOT_STICKY	0
#define HFS_STICKY		1

/* The number of hash buckets in a B-tree's bnode cache */
#define HFS_CACHELEN	17	/* primes are best? */

/*
 * Legal values for the 'ndType' field of a (struct NodeDescriptor)
 *
 * Reference: _Inside Macintosh: Files_ p. 2-65
 */
#define ndIndxNode	0x00	/* An internal (index) node */
#define ndHdrNode	0x01	/* The tree header node (node 0) */
#define ndMapNode	0x02	/* Holds part of the bitmap of used nodes */
#define ndLeafNode	0xFF	/* A leaf (ndNHeight==1) node */

/*================ Function-like macros ================*/

/* Accesss the cache slot which should contain the desired node */
#define bhash(tree, node) ((tree)->cache[(node) % HFS_CACHELEN])

/* round up to multiple of sizeof(HFS_WORD) */
#define ROUND(X) ((X + sizeof(HFS_WORD) - 1) & ~(sizeof(HFS_WORD)-1))

/* Refer to the (base-1) array of offsets in a bnode */
#define RECTBL(X,N) (((HFS_WORD *)((X)->bh->b_data + HFS_BNODE_SIZE))[-(N)])

/*================ File-local data types ================*/

/*
 * typdef cmpfn
 *
 * The type 'cmpfn' is a comparison function taking 2 keys and
 * returning a positive, negative or zero integer according to the
 * ordering of the two keys (just like strcmp() does for strings).
 */
typedef int (*cmpfn)(const struct hfs_brec_key *, const struct hfs_brec_key *);

/*
 * enum hfs_lock_t
 *
 * Access types used when requesting access to a B-node
 */
enum hfs_lock_t {
	HFS_LOCK_NONE = 0,	/* Illegal */
	HFS_LOCK_READ,		/* read-only access */
	HFS_LOCK_RESRV,		/* might potentially modify */
	HFS_LOCK_WRITE,		/* will modify now (exclusive access) */
	HFS_LOCK_BLIND,		/* assume access arbitrated elsewhere */
};

/*
 * struct BThdrRec
 *
 * The B-tree header record
 *
 * This data structure is stored in the first node (512-byte block) of
 * each B-tree file.  It contains important information about the
 * B-tree.  Most fields vary over the life of the tree and are
 * indicated by a 'V' in the comments.	The other fields are fixed for
 * the life of the tree and are indicated by a 'F'.
 *
 * Reference: _Inside Macintosh: Files_ pp. 2-68 through 2-69 */
struct BTHdrRec {
	HFS_WORD	bthDepth;	 /* (V) The number of levels
					    in this B-tree */
	HFS_LWORD	bthRoot;	 /* (V) The node number of the
					    root node */
	HFS_LWORD	bthNRecs;	 /* (V) The number of leaf
					    records */
	HFS_LWORD	bthFNode;	 /* (V) The number of the
					    first leaf node */
	HFS_LWORD	bthLNode;	 /* (V) The number of the last
					    leaf node */
	HFS_WORD	bthNodeSize;	 /* (F) The number of bytes in
					    a node (always 512) */
	HFS_WORD	bthKeyLen;	 /* (F) The maximum allowable
					    length of a key */
	HFS_LWORD	bthNNodes;	 /* (V) The total number of
					    nodes */
	HFS_LWORD	bthFree;	 /* (V) The number of unused
					    nodes */
	HFS_BYTE	bthResv[76];	 /* Reserved */
} DONT_ALIGN;

/*
 * struct NodeDesciptor
 *
 * The B-tree node descriptor.
 *
 * This structure begins each node in the B-tree file.	It contains
 * important information about the node's contents.  'V' and 'F' in
 * the comments indicate fields that are variable or fixed over the
 * life of a node, where the 'life' of a node is defined as the period
 * between leaving and reentering the free pool.
 *
 * Reference: _Inside Macintosh: Files_ p. 2-64
 */
struct NodeDescriptor {
	HFS_LWORD	ndFLink;	 /* (V) Node number of the
					    next node at the same level */
	HFS_LWORD	ndBLink;	 /* (V) Node number of the
					    previous node at the same
					    level */
	HFS_BYTE	ndType;		 /* (F) The type of node */
	HFS_BYTE	ndNHeight;	 /* (F) The level of this node
					    where leaves are at level 1 */
	HFS_WORD	ndNRecs;	 /* (V) The number of records
					    in this node */
	HFS_WORD	ndResv2;	 /* Reseved */
} DONT_ALIGN;

/*
 * struct hfs_btree
 *
 * An in-core B-tree.
 *
 * This structure holds information from the BTHdrRec, MDB
 * (superblock) and other information needed to work with the B-tree.
 * 
 * Both the superblock and the BTHdrRec are kept in-core, so this
 * information is mostly redundant.  However, on a little-endian
 * processor like the x86 it is convenient to have it all in the
 * proper byte-order.  It is also nice to have it all in one place.
 */
struct hfs_btree {
	int		    magic;	       /* Magic number to
						  guard against wild
						  pointers */
	cmpfn		    compare;	       /* Comparison function
						  for this tree */
	struct hfs_bnode    *head;	       /* Pointer to the in-
						  core copy of node 0 */
	struct hfs_bnode    *root;	       /* Pointer to the
						  in-core copy of the
						  root node */
	struct hfs_raw_extent		       /* Extent in the MDB */
			    *raw_extent;
	HFS_LWORD	    *size;	       /* Size in the MDB */
	dev_t		    dev;	       /* The device
						  containing this
						  filesystem */
	int		    reserved;	       /* bnodes claimed but
						  not yet used */
	struct hfs_bnode		       /* The bnode cache */
			    *cache[HFS_CACHELEN];
	struct hfs_file	    file;	       /* Location of the
						  B-tree on disk */
	/* Fields from the BTHdrRec in native byte-order: */
	__u16		    bthKeyLen;
	__u32		    bthRoot;
};

/*
 * struct hfs_bnode
 *
 * An in-core B-tree node
 *
 * This structure holds information from the NodeDescriptor in native
 * byte-order, a pointer to the buffer_head which contains the actual
 * node and fields neccesary for locking access to the node durring
 * updates.  The use of the locking fields is explained with the
 * locking functions.
 */
struct hfs_bnode {
	int		    magic;   /* Magic number to guard against
					wild pointers */
	struct buffer_head  *bh;     /* The buffer_head containing the
					actual node */
	struct hfs_btree    *tree;   /* The tree to which this node
					belongs */
	struct hfs_bnode    *prev;   /* Next node in this hash bucket */
	struct hfs_bnode    *next;   /* Previous node in this hash
					bucket */
	int		    sticky;  /* Boolean: non-zero means keep
					this node in-core (set for
					root and head) */
	__u32		    node;    /* Node number */
	/* locking related fields: */
	struct wait_queue   *wqueue; /* Wait queue for write access */
	struct wait_queue   *rqueue; /* Wait queue for read or reserve
					access */
	int		    count;   /* Number of processes accessing
					this node */
	int		    resrv;   /* Boolean, true means a process
					had placed a 'reservation' on
					this node */
	int		    lock;    /* Boolean, true means some
					process has exclusive access,
					so KEEP OUT */
};

struct hfs_bnode_ref {
	struct hfs_bnode *bn;
	int lock_type;
};

/* Convert a (struct hfs_tree *) to a pointer to the b-tree header */
extern inline struct BTHdrRec *btree_head(const struct hfs_btree *tree)
{ return (struct BTHdrRec *)((char *)tree->head->bh->b_data +
			sizeof(struct NodeDescriptor)); }

/* Convert a (struct hfs_bnode *) to a pointer to the header (node
   descriptor) of the bnode */
extern inline struct NodeDescriptor *bnode_head(const struct hfs_bnode *bnode)
{ return (struct NodeDescriptor *)bnode->bh->b_data; }

/* Convert a (struct hfs_bnode *) and an index to the value of the
   n-th offset in the bnode (N >= 1) to the offset */
extern inline __u16 bnode_offset(const struct hfs_bnode *bnode, int n)
{ return from_w(RECTBL(bnode,n)); }

/* Convert a (struct hfs_bnode *) and an index to the size of the
   n-th record in the bnode (N >= 1) */
extern inline __u16 bnode_rsize(const struct hfs_bnode *bnode, int n)
{ return bnode_offset(bnode, n+1) - bnode_offset(bnode, n); }

/* Convert a (struct hfs_bnode *) to the offset of the empty part */
extern inline __u16 bnode_end(const struct hfs_bnode *bnode)
{ return bnode_offset(bnode, from_w(bnode_head(bnode)->ndNRecs) + 1); }

/* Convert a (struct hfs_bnode *) to the number of free bytes it contains */
extern inline __u16 bnode_freespace(const struct hfs_bnode *bnode)
{ return HFS_SECTOR_SIZE - bnode_end(bnode)
	  - (from_w(bnode_head(bnode)->ndNRecs) + 1)*sizeof(HFS_WORD); }

/* Convert a (struct hfs_bnode *) X and an index N to
   the address of the record N in the bnode (N >= 1) */
extern inline void *bnode_datastart(const struct hfs_bnode *bnode)
{ return (void *)(bnode->bh->b_data + sizeof(struct NodeDescriptor)); }

/* Convert a (struct hfs_bnode *) X and an index N to
   the address of the record N in the bnode (N >= 1) */
extern inline void *bnode_address(const struct hfs_bnode *bnode, int n)
{ return (void *)(bnode->bh->b_data + bnode_offset(bnode, n)); }

/* Convert a (struct hfs_bnode *) X and an index N to
   the address of the key for record N in the bnode (N >= 1) */
extern inline struct hfs_brec_key *bnode_key(const struct hfs_bnode *bnode, int n)
{ return (struct hfs_brec_key *)bnode_address(bnode, n); }

/* Convert a (struct hfs_brec *) X to the address of the record's key */
extern inline void *brec_address(const struct hfs_brec *brec)
{ return bnode_address(brec->bnode->bn, brec->record); }

/* Convert a (struct hfs_brec *) X to the address of the record's key */
extern inline struct hfs_brec_key *brec_key(const struct hfs_brec *brec)
{ return bnode_key(brec->bnode->bn, brec->record); }

/* Convert a (struct hfs_brec_key *) X to the address of the record */
extern inline void *bkey_address(const struct hfs_brec_key *key)
{ return (void *)key + ROUND(key->KeyLen+1); }

/* Dirty the buffer holding the b-tree header */
extern inline void btree_dirty(const struct hfs_btree *tree)
{ mark_buffer_dirty(tree->head->bh, 1); }

/* Dirty the buffer holding the bnode */
extern inline void bnode_dirty(const struct hfs_bnode *bn)
{ mark_buffer_dirty(bn->bh, 1); }

extern struct hfs_bnode * hfs_bnode_read(struct hfs_btree *, int, int);
extern void hfs_bnode_relse(struct hfs_bnode_ref *);
extern struct hfs_bnode_ref * hfs_bnode_find(struct hfs_btree *, __u32, int);
extern void hfs_bnode_lock(struct hfs_bnode_ref *, int);

extern int hfs_bnode_bitop(struct hfs_btree *, __u32, int);
extern struct hfs_bnode * hfs_bnode_alloc(struct hfs_btree *);
extern struct hfs_bnode * hfs_bnode_init(struct hfs_btree *, __u32);
extern void hfs_btree_extend(struct hfs_btree *);

extern void hfs_brec_lock(struct hfs_brec *);
extern struct hfs_brec *hfs_brec_get_root(struct hfs_btree *, int, int);
extern struct hfs_brec *hfs_brec_get_child(struct hfs_btree *,
					   struct hfs_brec *, int);
extern int hfs_brec_find(struct hfs_btree *, const struct hfs_brec_key *,
			 struct hfs_brec **, int);
