/* 
 *  linux/include/linux/hfs_fs.h
 *
 *  Copyright (C) 1995, 1996  Paul H. Hargrove
 *  This file may be distributed under the terms of the GNU Public License
 */

#ifndef _HFS_FS_H
#define _HFS_FS_H

#include <linux/version.h>
#include <linux/fs.h>
#include <linux/dirent.h>
#include <linux/errno.h>
#include <asm/types.h>
#include <asm/byteorder.h>

#include "hfs_fs_i.h"
#include "hfs_fs_sb.h"

#define HFS_I(inode)		((struct hfs_inode_info *)(&inode->u))
#define HFS_SB(super)		((struct hfs_sb_info *)(&super->u))
#define HFS_FID(X)		(from_l((X->cdrType==HFS_CDR_FIL) ? \
					(X)->u.fil.FlNum : (X)->u.dir.DirID))

/* offsets to various blocks */
#define HFS_PMAP_BLK		1 /* First block of partition map */
#define HFS_MDB_BLK		2 /* Block (w/i partition) of MDB */

/* magic numbers for various disk blocks */
#define HFS_DRVR_DESC_MAGIC	0x4552 /* "ER": driver descriptor map */
#define HFS_OLD_PMAP_MAGIC	0x5453 /* "TS": old-type partition map */
#define HFS_NEW_PMAP_MAGIC	0x504D /* "PM": new-type partition map */
#define HFS_SUPER_MAGIC		0x4244 /* "BD": HFS MDB (super block) */
#define HFS_MFS_SUPER_MAGIC	0xD2D7 /* MFS MDB (super block) */

/* magic numbers for CAP structures */
#define HFS_CAP_MAGIC		0xFF
#define HFS_CAP_MAGIC1		0xDA
#define HFS_CAP_VERSION		0x10

/* magic numbers for Apple Double header files */
#define HFS_DBL_MAGIC		0x00051607
#define HFS_DBL_VERSION		0x00020000
#define HFS_NAT_VERSION		0x00010000

/* magic numbers for various internal structures */
#define HFS_FILE_MAGIC		0x4801
#define HFS_BREC_MAGIC		0x4802
#define HFS_BTREE_MAGIC		0x4803
#define HFS_BNODE_MAGIC		0x4804

/* various FIXED size parameters */
#define HFS_SECTOR_SIZE		512    /* size of an HFS sector */
#define HFS_SECTOR_SIZE_BITS	9      /* log_2(HFS_SECTOR_SIZE) */
#define HFS_BNODE_SIZE		512    /* size of an HFS sector */
#define HFS_NAMELEN		31     /* maximum length of an HFS filename */
#define HFS_NAMEMAX		(3*31) /* max size of ENCODED filename */
#define HFS_BM_MAXBLOCKS	(16)   /* max number of bitmap blocks */
#define HFS_BM_BPB (8*HFS_SECTOR_SIZE) /* number of bits per bitmap block */

/* Meanings of the arAtrb field of the MDB,
 * Reference: _Inside Macintosh: Files_ p. 2-61
 */
#define HFS_SB_ATTRIB_HLOCK 0x0080
#define HFS_SB_ATTRIB_CLEAN 0x0100
#define HFS_SB_ATTRIB_SPARED 0x0200
#define HFS_SB_ATTRIB_SLOCK 0x8000

/* The space used for the AppleDouble or AppleSingle headers */
#define HFS_DBL_HDR_LEN		1024

/* The space used for the Netatalk header */
#define HFS_NAT_HDR_LEN		1024  /* 589 for an exact match */

/* 2**16 - 1 */
#define HFS_USHRT_MAX	65535

/* Some special File ID numbers */
#define HFS_POR_FID	1	/* Parent Of the Root */
#define HFS_ROOT_FID	2	/* ROOT directory */
#define HFS_EXT_FID	3	/* EXTents B-tree */
#define HFS_CAT_FID	4	/* CATalog B-tree */
#define HFS_BAD_FID	5	/* BAD blocks file */

/* Macros to extract file ID and file "type" from the Linux inode number */
#define HFS_IFID(X)	((X) & 0x3FFFFFFF)
#define HFS_ITYPE(X)	((X) & 0xC0000000)

/* generic ITYPEs */
#define HFS_ITYPE_0	0x00000000
#define HFS_ITYPE_1	0x40000000
#define HFS_ITYPE_2	0x80000000
#define HFS_ITYPE_3	0xC0000000
#define HFS_ITYPE_NORM	HFS_ITYPE_0	/* "normal" directory or file */

/* ITYPEs for CAP */
#define HFS_CAP_NORM	HFS_ITYPE_0	/* data fork or normal directory */
#define HFS_CAP_DATA	HFS_ITYPE_0	/* data fork of file */
#define HFS_CAP_NDIR	HFS_ITYPE_0	/* normal directory */
#define HFS_CAP_FNDR	HFS_ITYPE_1	/* finder info for file or dir */
#define HFS_CAP_BDIR	HFS_ITYPE_1	/* invalid value for a directory */
#define HFS_CAP_RSRC	HFS_ITYPE_2	/* resource fork of file */
#define HFS_CAP_RDIR	HFS_ITYPE_2	/* .resource directory */
#define HFS_CAP_FDIR	HFS_ITYPE_3	/* .finderinfo directory */

/* ITYPEs for Apple Double */
#define HFS_DBL_NORM	HFS_ITYPE_0	/* data fork or directory */
#define HFS_DBL_DATA	HFS_ITYPE_0	/* data fork of file */
#define HFS_DBL_DIR	HFS_ITYPE_0	/* directory */
#define HFS_DBL_HDR	HFS_ITYPE_1	/* AD header of file or dir */

/* ITYPEs for Apple Single */
#define HFS_SGL_NORM	HFS_ITYPE_0	/* data fork or directory */
#define HFS_SGL_SNGL	HFS_ITYPE_0	/* AppleSingle file */
#define HFS_SGL_DIR	HFS_ITYPE_0	/* directory */
#define HFS_SGL_DINF	HFS_ITYPE_1	/* %DirInfo for directory */

/* ITYPEs for netatalk */
#define HFS_NAT_NORM	HFS_ITYPE_0	/* data fork or directory */
#define HFS_NAT_DATA	HFS_ITYPE_0	/* data fork of file */
#define HFS_NAT_NDIR	HFS_ITYPE_0	/* normal directory */
#define HFS_NAT_HDR	HFS_ITYPE_1	/* AD header of file or dir */
#define HFS_NAT_HDIR	HFS_ITYPE_2	/* directory holding AD headers */

/* Flags to OR together for 'skip' argument to hfs_cat_next() */
#define HFS_SKIP_NONE	0x00
#define HFS_SKIP_DIR	0x01	/* exclude directory records */
#define HFS_SKIP_FIL	0x02	/* exclude file records */
#define HFS_SKIP_FTH	0x04	/* exclude file thread records */

/* Mode for hfs_bfind */
enum hfs_bfind_mode {
	HFS_BFIND_READ_EQ,	/* find the exact entry */
	HFS_BFIND_READ_LE,	/* find largest entry <= key */
	HFS_BFIND_WRITE		/* find the exact entry, locked for writing */
};

/* Typedefs for the 68000's data types */
typedef __u8		HFS_BYTE;
typedef __u16		HFS_WORD;
typedef __u32		HFS_LWORD;
typedef HFS_BYTE	HFS_CNAME[HFS_NAMELEN+1];

/* Macro to prevent GCC from playing with alignment of HFS structures */
#define DONT_ALIGN	__attribute__ ((packed))

/* Byte-order conversions for 16-bit "word" */
#define from_w(X)	(ntohs(X))
#define to_w(X)		((HFS_WORD)htons(X))
#define c_to_w(X)	((HFS_WORD)__constant_htons(X))

/* Byte-order conversions for 32-bit "long word" */
#define from_l(X)	(ntohl(X))
#define to_l(X)		((HFS_LWORD)htonl(X))
#define c_to_l(X)	((HFS_LWORD)__constant_htonl(X))

/* get the host-endian bit number of a big-endian bitfield element */
#if   defined(__BIG_ENDIAN_BITFIELD)
#define HFS_BE_BIT(X)	(X)
#elif defined(__LITTLE_ENDIAN_BITFIELD)
#define HFS_BE_BIT(X)	((X)^7)
#else
#error "Don't know if bitfields are big- or little-endian!"
#endif

/*
 * There are three time systems.  All three are based on seconds since
 * a particular time/date.
 *	unix:	unsigned lil-endian since 00:00 GMT, Jan. 1, 1970
 *	mac:	unsigned big-endian since 00:00 GMT, Jan. 1, 1904
 *	header:	  SIGNED big-endian since 00:00 GMT, Jan. 1, 2000
 *
 */
#define hfs_h_to_mtime(ARG)	to_l((signed long)from_l(ARG)+3029529600U)
#define hfs_m_to_htime(ARG)	((signed long)to_l(from_l(ARG)-3029529600U))
#define hfs_h_to_utime(ARG)	((signed long)from_l(ARG)+946684800U)
#define hfs_u_to_htime(ARG)	((signed long)to_l((ARG)-946684800U))
#define hfs_u_to_mtime(ARG)	to_l((ARG)+2082844800U)
#define hfs_m_to_utime(ARG)	(from_l(ARG)-2082844800U)

/* 
 * host byte-order independent type promotions
 */
extern __inline__ HFS_LWORD word_to_long(HFS_WORD w) {
	union {
		HFS_WORD	words[2]	DONT_ALIGN;
		HFS_LWORD	lword		DONT_ALIGN;
	} x;
	x.lword = 0;
	x.words[1] = w;
	return x.lword;
}

extern __inline__ HFS_LWORD byte_to_long(HFS_BYTE b) {
	union {
		HFS_BYTE	bytes[4]	DONT_ALIGN;
		HFS_LWORD	lword		DONT_ALIGN;
	} x;
	x.lword = 0;
	x.bytes[3] = b;
	return x.lword;
}

extern __inline__ HFS_WORD byte_to_word(HFS_BYTE b) {
	union {
		HFS_BYTE	bytes[2]	DONT_ALIGN;
		HFS_WORD	word		DONT_ALIGN;
	} x;
	x.word = 0;
	x.bytes[1] = b;
	return x.word;
}

/* add to an HFS_WORD returning the result in native byte order */
extern __inline__ __u16 add_w(HFS_WORD *x, __s16 y)
{
        __u16 tmp = y + from_w(*x);
        *x = to_w(tmp);
        return tmp;
}

/* add to an HFS_LWORD returning the result in native byte order */
extern __inline__ __u32 add_l(HFS_LWORD *x, __s32 y)
{
        __u32 tmp = y + from_l(*x);
        *x = to_l(tmp);
        return tmp;
}


/*======== HFS structures as they appear on the disk ========*/

/* An extent record on disk*/
struct hfs_raw_extent{
	HFS_WORD	block1;
	HFS_WORD	length1;
	HFS_WORD	block2;
	HFS_WORD	length2;
	HFS_WORD	block3;
	HFS_WORD	length3;
} DONT_ALIGN;

 /* The key used in the catalog b-tree */
struct hfs_cat_key {
	HFS_BYTE	Resrv1; /* padding */
	HFS_LWORD	ParID;	/* File ID of the dir containing the entry */
	HFS_CNAME	CName;	/* The filename of the entry */
} DONT_ALIGN;

 /* The key used in the extents b-tree */
struct hfs_ext_key {
	HFS_BYTE	FkType; /* fork type (HFS_FK_DATA or HFS_FK_RSRC) */
	HFS_LWORD	FNum;	/* The File ID of the file */
	HFS_WORD	FABN;	/* offset (in alloc blocks) into the file */
} DONT_ALIGN;

/* generic B-tree key */
union hfs_gen_key {
	struct hfs_cat_key ckr; /* The key used in the catalog b-tree */
	struct hfs_ext_key xkr; /* The key used in the extents b-tree */
};

/* A btree record key on disk */
struct hfs_brec_key{
	HFS_BYTE	KeyLen; /* The number of bytes in the key */
	union hfs_gen_key u;
} DONT_ALIGN;

/* The catalog entry for a file */
struct hfs_fil_inode {
	HFS_BYTE		Flags;		/* Flags such as read-only */
	HFS_BYTE		Typ;		/* file version number = 0 */
	HFS_BYTE		UsrWds[16];	/* data used by FInder */
	HFS_LWORD		FlNum;		/* The file ID */
	HFS_WORD		StBlk;		/* obsolete */
	HFS_LWORD		LgLen;		/* The logical EOF of
						   the data fork*/
	HFS_LWORD		PyLen;		/* The physical EOF of
						   the data fork */
	HFS_WORD		RStBlk;		/* obsolete */
	HFS_LWORD		RLgLen;		/* The logical EOF of
						   the resource fork */
	HFS_LWORD		RPyLen;		/* The physical EOF of
						   the resource fork */
	HFS_LWORD		CrDat;		/* The creation date */
	HFS_LWORD		MdDat;		/* The modified date */
	HFS_LWORD		BkDat;		/* The last backup date */
	HFS_BYTE		FndrInfo[16];	/* more data for the
						   Finder */
	HFS_WORD		ClpSize;	/* number of bytes to
						   allocate when
						   extending files */
	struct hfs_raw_extent	ExtRec;		/* first extent record
						   for the data fork */
	struct hfs_raw_extent	RExtRec;	/* first extent record
						   for the resource
						   fork */
	HFS_LWORD		Resrv;		/* reserved for future
						   use by Apple */
} DONT_ALIGN;

/* the inode for a directory */
struct hfs_dir_inode {
	HFS_WORD	Flags;		/* flags */
	HFS_WORD	Val;		/* The number of files in the
					   directory (Valence) */
	HFS_LWORD	DirID;		/* The file ID of this directory */
	HFS_LWORD	CrDat;		/* The creation date */
	HFS_LWORD	MdDat;		/* The modification date */
	HFS_LWORD	BkDat;		/* The last backup date */
	HFS_BYTE	UsrInfo[16];	/* data used by the Finder */
	HFS_BYTE	FndrInfo[16];	/* more data use by the Finder */
	HFS_BYTE	Resrv[16];	/* reserved for future
					   use by Apple */
} DONT_ALIGN;

/* entry for a thread 
 * A thread marks the beginning of the entries in a given directory,
 * serving the same purpose as a '..' directory entry.
 */
struct hfs_thd_inode {
	HFS_BYTE	Reserv[8];	/* reserved for future
					   use by Apple */
	HFS_LWORD	ParID;		/* file ID of the parent
					   of this directory */
	HFS_CNAME	CName;		/* The name of this directory */
} DONT_ALIGN;

/* generic inode entry */
union hfs_gen_inode{
	struct hfs_fil_inode fil;
	struct hfs_dir_inode dir;
	struct hfs_thd_inode thd;
};

/* A catalog tree entry (equivalent to an inode) */
struct hfs_inode {
	HFS_BYTE	cdrType;	/* The type of entry
					   (see hfs_cdr_type, below) */
	HFS_BYTE	cdrResrv2;	/* padding */
	union hfs_gen_inode u;		/* the actual entry */
} DONT_ALIGN;

/* legal values for hfs_inode.cdrType */
#define HFS_CDR_DIR	1
#define HFS_CDR_FIL	2
#define HFS_CDR_THD	3
#define HFS_CDR_FTH	4

/* legal values for hfs_brec_key.u.xkr.FkType and hfs_file.fork */
#define HFS_FK_DATA	0x00
#define HFS_FK_RSRC	0xFF

/*======== Data structures kept in memory ========*/

/*
 * struct hfs_brec
 *
 * An in-core reference to a record in a (struct hfs_bnode)
 *
 * This structure holds information about a single record within a
 * bnode, possibly including the paths from the root to the record,
 * and from the record to a leaf.
 */
struct hfs_brec {
	int			magic;		/* A magic number */
	struct hfs_bnode_ref	*bnode;		/* The node containing
						   this record */
	int			record;		/* The index of this
						   record within the
						   node */
	int			flags;		/* used internally */
	struct hfs_brec_key	*key;		/* A pointer to the
						   key within the node */
	struct hfs_brec		*up, *down;	/* A doubly-linked
						   list leading from
						   the root to some
						   leaf */
	union {			/* The data associated with the key */
		/* To allow copying without knowing what type */
		void			*generic;
		/* In a leaf of the extents b-tree: */
		struct hfs_inode	*raw_inode;
		/* In a leaf of the catalog b-tree: */
		struct hfs_raw_extent	*raw_extent;
		/* In an internal node of either b-tree: */
		HFS_LWORD		*child; 
	} data;
};

/*
 * A descriptor for a single entry within the header of an
 * Apple Double of Apple Single header file.
 * An array of these make up a table of contents for the file.
 */
struct hfs_dbl_descr {
        HFS_LWORD       id;     /* The Apple assigned ID for the entry type */
        HFS_LWORD       offset; /* The offset to reach the entry */
        HFS_LWORD       length; /* The length of the entry */
} DONT_ALIGN;

/* 
 * The header for a version 1 Apple Double header file.
 * There is space reserved for all the fields we need for Netatalk
 */
struct hfs_nat_hdr {
	HFS_LWORD	magic;
	HFS_LWORD	version;
	HFS_BYTE	homefs[16];
	HFS_WORD	entries;
	struct hfs_dbl_descr descr[5];
	char		real_name[255];	/* id=3 */
	char		comment[200];	/* id=4 XXX: not done yet */
	HFS_LWORD	create_time;	/* \	   */
	HFS_LWORD	modify_time;	/*  |      */
	HFS_LWORD	backup_time;	/*  | id=7 */
	HFS_WORD	filler;		/*  |      */
	HFS_WORD	attr;		/* /	   */
	HFS_BYTE	finderinfo[32]; /* id=9 */
} DONT_ALIGN;

/* 
 * The header for a version 2 Apple Double header file.
 * There is space reserved for all the fields we can deal with.
 */
struct hfs_dbl_hdr {
	HFS_LWORD	magic;
	HFS_LWORD	version;
	HFS_BYTE	filler[16];
	HFS_WORD	entries;
	struct hfs_dbl_descr descr[16];
	HFS_LWORD	create_time;	/* \	   */
	HFS_LWORD	modify_time;	/*  | id=8 */
	HFS_LWORD	backup_time;	/*  |	   */
	HFS_LWORD	access_time;	/* /	   */
	HFS_BYTE	finderinfo[32]; /* id=9 */
	HFS_LWORD	fileinfo;	/* id=10 */
	HFS_CNAME	real_name;	/* id=3 */
	char		comment[200];	/* id=4 XXX: not done yet */
} DONT_ALIGN;

/* finder metadata for CAP */
struct hfs_cap_info {
	HFS_BYTE	fi_fndr[32];
	HFS_WORD	fi_attr;
	HFS_BYTE	fi_magic1;
	HFS_BYTE	fi_version;
	HFS_BYTE	fi_magic;
	HFS_BYTE	fi_bitmap;
	HFS_BYTE	fi_shortfilename[12+1];
	HFS_BYTE	fi_macfilename[32+1];
	HFS_BYTE	fi_comln;
	HFS_BYTE	fi_comnt[200];
} DONT_ALIGN;

/* string.c */
extern hfs_strcmp(const unsigned char *, const unsigned char *);

/* bitops.c */
extern __u32 hfs_find_first_zero_bit(const void *, __u32);
extern __u32 hfs_find_next_zero_bit(const void *, __u32, __u32);
extern __u32 hfs_count_zero_bits(const void *, __u32, __u32, __u32);

/* bitmap.c */
extern __u16 hfs_vbm_count_free(const struct super_block *, __u16);
extern __u16 hfs_vbm_search_free(const struct super_block *, __u16 *);
extern int hfs_set_vbm_bits(struct super_block *, __u16, __u16);
extern int hfs_clear_vbm_bits(struct super_block *, __u16, __u16);

/* btree.c */
extern void hfs_brec_relse(struct hfs_brec **);
extern void hfs_brec_relse_one(struct hfs_brec **);
extern int hfs_bsucc(struct hfs_brec **);
extern int hfs_bfind(struct hfs_btree *, const struct hfs_brec_key *,
		     struct hfs_brec **, int);
extern int hfs_binsert(struct hfs_btree *, const struct hfs_brec_key *,
		       const void *, __u16);
extern struct hfs_btree *hfs_btree_init(struct super_block *, ino_t,
					struct hfs_raw_extent *,
					HFS_LWORD *, HFS_LWORD);
extern void hfs_btree_free(struct hfs_btree *);
extern void hfs_bnode_add(struct hfs_btree *, int);

/* extent.c */
extern void hfs_extent_in(struct hfs_file *, const struct hfs_raw_extent *);
extern void hfs_extent_out(const struct hfs_file *, struct hfs_raw_extent *);
extern int hfs_block_map(struct hfs_file *, int, int);
extern void hfs_file_extend(struct hfs_file *, int);
extern void hfs_extent_trim(struct hfs_file *, struct hfs_extent *);

/* catalog.c */
extern void hfs_build_cat_key(ino_t, __u8, const char *, struct hfs_brec_key *);
extern int hfs_cat_find(struct super_block *, const struct hfs_brec_key *,
			struct hfs_brec **, int);
extern int hfs_cat_next(struct hfs_brec **, unsigned int, int);
extern int hfs_cat_nth(struct inode *, struct hfs_brec **, __u32, int);

/* inode.c */
extern struct inode * __hfs_iget(struct super_block *,
				 const struct hfs_brec_key *, ino_t);
extern int hfs_count_entries(struct inode *);
extern void hfs_read_file_inode(struct hfs_file *, struct inode *,
				struct hfs_inode *, __u8, ino_t);
extern void hfs_write_file_inode(struct hfs_file *, struct inode *,
				 struct hfs_inode *);
extern void hfs_write_inode(struct inode *, struct hfs_brec *);
extern int hfs_notify_change(struct inode *, struct iattr *);

/* inode_cap.c */
extern int hfs_cap_ifill(struct inode *, void *, ino_t, ino_t);
extern void hfs_cap_put_inode(struct inode *);
extern void hfs_cap_write_inode(struct inode *);

/* inode_dbl.c */
extern int hfs_dbl_ifill(struct inode *, void *, ino_t, ino_t);
extern void hfs_dbl_put_inode(struct inode *);
extern void hfs_dbl_write_inode(struct inode *);

/* inode_nat.c */
extern int hfs_nat_ifill(struct inode *, void *, ino_t, ino_t);
extern void hfs_nat_put_inode(struct inode *);
extern void hfs_nat_write_inode(struct inode *);

#if 0
/* inode_sngl.c */
extern int hfs_sngl_ifill(struct inode *, struct hfs_brec *, ino_t, ino_t);
extern void hfs_sngl_put_inode(struct inode *);
extern void hfs_sngl_write_inode(struct inode *);
#endif

/* super.c */
extern struct super_block *hfs_read_super(struct super_block *,void *,int);

/* file.c */
extern struct buffer_head * hfs_getblk(struct hfs_file *, int, int);
extern int hfs_do_read(struct hfs_file *, off_t, char *, int, off_t *);
extern int hfs_do_write(struct hfs_file *, off_t, const char *, int);
extern int hfs_do_create(struct inode *, const char *, int, int,
			 struct inode **);
extern int hfs_do_mkdir(struct inode *, const HFS_CNAME);
extern struct inode_operations hfs_file_inode_operations;

/* file_cap.c */
extern struct inode_operations hfs_cap_info_inode_operations;

/* file_dbl.c */
extern struct inode_operations hfs_dbl_hdr_inode_operations;

/* file_nat.c */
extern struct inode_operations hfs_nat_hdr_inode_operations;

#if 0
/* file_sngl.c */
extern struct inode_operations hfs_sngl_file_inode_operations;
#endif

/* dir.c */
extern int hfs_dir_read(struct inode *, struct file *, char *, int);
extern int hfs_dupdir(struct inode *, ino_t, struct inode **);
extern struct inode * hfs_lookup_parent(struct inode *, int *);

/* dir_cap.c */
extern struct inode_operations hfs_cap_dir_inode_operations;

/* dir_dbl.c */
extern struct inode_operations hfs_dbl_dir_inode_operations;

/* dir_nat.c */
extern struct inode_operations hfs_nat_dir_inode_operations;

#if 0
/* dir_sngl.c */
extern struct inode_operations hfs_sngl_dir_inode_operations;
#endif

/* trans.c */
extern int hfs_cap2mac(char *, int, const char *);
extern int hfs_apl2mac(char *, int, const char *);
extern int hfs_triv2mac(char *, int, const char *);
extern int hfs_latin2mac(char *, int, const char *);
extern int hfs_mac2cap(char *, const unsigned char *);
extern int hfs_mac2latin(char *, const unsigned char *);
extern int hfs_mac2seven(char *, const unsigned char *);
extern int hfs_mac2eight(char *, const unsigned char *);
extern int hfs_mac2alpha(char *, const unsigned char *);
extern int hfs_mac2triv(char *, const unsigned char *);

extern __inline__ struct inode *hfs_iget(struct super_block *sb, ino_t parent,
					 __u8 len, const char *name, ino_t type)
{
	struct hfs_brec_key key;

	hfs_build_cat_key(parent, len, name, &key);
	return __hfs_iget(sb, &key, type);
}

#if defined(DEBUG_ALL) || defined(DEBUG_MEM)
#	include <linux/malloc.h>
#	include <linux/kernel.h>
extern long int hfs_alloc;
#	define KMALLOC(A,B) \
		(printk("HFS: %ld bytes allocation at %s:%u\n", \
	 	(hfs_alloc += (A)), __FILE__, __LINE__), kmalloc(A,B))
#	define KFREE_S(A,B) \
		(printk("HFS: %ld bytes allocation at %s:%u\n", \
	 	(hfs_alloc -= (B)), __FILE__, __LINE__), kfree_s(A,B))
#else
#	define KMALLOC kmalloc
#	define KFREE_S kfree_s
#endif

#endif
