/* 
 * linux/include/linux/hfs_fs_i.h
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file defines the type (struct hfs_inode_info) and the two
 * subordinate types (struct hfs_extent) and (struct hfs_file).
 */

#ifndef _HFS_FS_I
#define _HFS_FS_I

/*
 * struct hfs_extent
 *
 * The offset to allocation block mapping for a given file is
 * contained in a series of these structures.  Each (struct
 * hfs_extent) records upto three runs of contiguous allocation
 * blocks.  An allocation block is a contiguous group of physical
 * blocks.
 */
struct hfs_extent {
	unsigned short	   start;     /* Where in the file this record
				         begins (in allocation blocks) */
	unsigned short     end;       /* Where in the file this record
				         ends (in allocation blocks) */
	unsigned short	   block[3];  /* The allocation block on disk which
				         begins this extent */
	unsigned short	   length[3]; /* The number of allocation blocks
				         in this extent */
	struct hfs_extent  *next;     /* Next extent record for this file */
	struct hfs_extent  *prev;     /* Previous extent record for this file */
	int		   count;     /* Number of times it is used */
};

/*
 * struct hfs_file 
 *
 * This structure holds the information necessary to locate a single
 * Linux file in an HFS filesystem.  Since a file has two forks, there
 * can be two different (struct hfs_file)s corresponding to the same
 * file on disk.
 */
struct hfs_file {
	int		   magic;     /* A magic number */
	struct inode	   *inode;    /* The inode corresponding to
					 this file. */
	struct hfs_extent  first;     /* The first extent record for
					 this file */
	struct hfs_extent  *cache;     /* The most-recently accessed
					 extent record for this file */
	off_t		   size;      /* The actual size of the file on disk */
	dev_t		   dev;	      /* The device holding this file */
	__u16		   clumpsize; /* The number of bytes to
					 allocate when extending this file */
	__u8		   fork;      /* The fork (HFS_FK_DATA or
					 HFS_FK_RSRC) of this file */
};

/*
 * struct hfs_inode_info
 *
 * The HFS-specific part of a Linux (struct inode)
 */
struct hfs_inode_info {
	/* The search key for the catalog tree entry for this file.
	   Actually had type (struct hfs_brec_key), defined in hfs_fs.h. */
	unsigned char	key[38];

	union {
		/* data for a directory under all schemes */
		struct {		
				       
			__u32			dirs;
			__u32			files;
		} dir;

		/* data for a file under the CAP scheme
		   and for the data fork under AppleDouble */
		struct hfs_file file;
					  
		/* data for the metadata file under the CAP scheme */
		struct {
			void			*meta;
		} cap_info;

		/* data for the header file under AppleDouble */
		struct {		
			struct hfs_dbl_hdr	*meta;
			struct hfs_file		rsrc;
		} dbl_hdr;

		/* data for the header file under Netatalk */
		struct {		
			struct hfs_nat_hdr	*meta;
			struct hfs_file		rsrc;
		} nat_hdr;
#if 0
		/* data for the file under AppleSingle */
		struct {		
			void 			*meta;
			struct hfs_file		rsrc;
			struct hfs_file		data;
		} sngl;
#endif
	} u;
};

#endif
