/*
 * linux/fs/hfs/inode.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains inode-related functions which do not depend on
 * which scheme is being used to represent forks.
 *
 * Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/stat.h>
#include <linux/stddef.h>
#include <linux/malloc.h>
#include <linux/mm.h>

/*================ Variable-like macros ================*/

/* to keep some of the code from getting to verbose */
#define RAW	(entry->data.raw_inode)

#define HFS_VALID_MODE_BITS  (S_IFREG | S_IFDIR | S_IRWXU | S_IRWXG | S_IRWXO)

/*================ File-local functions ================*/

/*
 * propogate_change()
 *
 * Given an HFS superblock, the inode number of a HFS file and a
 * (struct iattr) describing changes to an inode, propogate thoes
 * changes to the inode for which we were given the inode number
 * if it is in-core.
 *
 * XXX: It would be helpful if instead of calling iget(), there was a
 * way to get the inode if it were in-core, or NULL if not.  As it
 * currently stands it is neccesary to call iget() which will return
 * the inode if it is in-core.	However, if the inode is not in-core,
 * iget() will allocate an empty inode (possibly by removing an inode
 * from memory) and insert it into the inode hash tables.  When iget()
 * returns this inode to us we proceed to call clear_inode(), removing
 * it from the hash tables and making most of the work done by iget()
 * a waste of time.  The other reason to want such a function is
 * explained in hfs/dir.c with the hfs_*_readdir() functions. In that
 * case it would allow the hfs_*_readdir() function to exhibit a
 * particularly desirable property without the need to over burden
 * iget() with requests for inodes that will simply be cleared.
 */
static void propogate_change(struct super_block *sb, ino_t ino,
				    struct iattr * attr)
{
	struct inode *inode;
	int old_dirt;

	if ((inode = iget(sb, ino))) {
		if ((inode->i_sb != sb) || (inode->i_ino != ino)) {
			iput(inode);
		} else if (!inode->i_mode) {
			clear_inode(inode);
		} else {
			/* We don't dirty the inode here since notify_change()
			   makes sure that one of the in-core inodes which
			   correspond to a given file is dirtied. */
			old_dirt = inode->i_dirt;
			inode_setattr(inode, attr);
			inode->i_dirt = old_dirt;
			iput(inode);
		}
	}
	return;
}

/*================ Global functions ================*/

/*
 * hfs_count_entries()
 *
 * Given the inode of an HFS directory, count the number of files
 * and directories in the directory.
 * Return 0 on success, non-zero on error.
 */
int hfs_count_entries(struct inode *dir)
{
	struct hfs_brec *entry;
	struct hfs_brec_key key;
	int error;

	HFS_I(dir)->u.dir.dirs = HFS_I(dir)->u.dir.files = 0;

	hfs_build_cat_key(HFS_IFID(dir->i_ino), 0, NULL, &key);
	if (hfs_cat_find(dir->i_sb, &key, &entry, 0)) {
		return -1;
	}

	do {
		if (!(error=hfs_cat_next(&entry, 1, HFS_SKIP_NONE))) {
			if (RAW->cdrType == HFS_CDR_FIL) {
				++HFS_I(dir)->u.dir.files;
			} else if (RAW->cdrType == HFS_CDR_DIR) {
				++HFS_I(dir)->u.dir.dirs;
			}
		}
	} while (!error); /* -ENOENT is normal termination */
	
	return (error == -ENOENT) ? 0 : error;
}

/*
 * hfs_read_file_inode()
 *
 * Given an hfs_file and coresponding inode, initialize them from the
 * given HFS catalog entry (hfs_inode), the given fork type
 * (HFS_FK_DATA or HFS_FK_RSRC) and given file id number.
 */
void hfs_read_file_inode(struct hfs_file *file, struct inode *inode,
			    struct hfs_inode *raw, __u8 fork, ino_t fid)
{
	HFS_LWORD lsize, psize;
	struct hfs_raw_extent *ext;

	file->magic = HFS_FILE_MAGIC;
	file->inode = inode;
	file->clumpsize = from_w(raw->u.fil.ClpSize);
	file->dev = inode->i_dev;
	if ((file->fork = fork) == HFS_FK_DATA) {
		lsize = raw->u.fil.LgLen;
		psize = raw->u.fil.PyLen;
		ext = &raw->u.fil.ExtRec;
	} else {
		lsize = raw->u.fil.RLgLen;
		psize = raw->u.fil.RPyLen;
		ext = &raw->u.fil.RExtRec;
	}
	hfs_extent_in(file, ext);
	if (!IS_NOEXEC(inode) && (fork == HFS_FK_DATA)) {
		inode->i_mode = ((raw->u.fil.Flags & 1) ? 0 : S_IWUGO) |
				S_IRUGO | S_IXUGO | S_IFREG;
	} else {
		inode->i_mode = ((raw->u.fil.Flags & 1) ? 0 : S_IWUGO) |
				S_IRUGO | S_IFREG;
	}
	inode->i_nlink = 1;
	file->size = inode->i_size = from_l(lsize);
	inode->i_blocks = from_l(psize) >> HFS_SECTOR_SIZE_BITS;
}

/*
 * hfs_write_file_inode()
 *
 * This function is called by the hfs_*_write_inode() functions.
 * Its purpose is to copy the relevent fields from the inode and an
 * hfs_file to the catalog B-tree record for a regular file.
 */
void hfs_write_file_inode(struct hfs_file *file, struct inode *inode,
			     struct hfs_inode *raw)
{
	if (file->magic == HFS_FILE_MAGIC) {
		if (file->fork == HFS_FK_DATA) {
			raw->u.fil.LgLen = to_l(file->size);
			raw->u.fil.PyLen = to_l(inode->i_blocks <<
						HFS_SECTOR_SIZE_BITS);
			hfs_extent_out(file, &raw->u.fil.ExtRec);
		} else {
			raw->u.fil.RLgLen = to_l(file->size);
			raw->u.fil.RPyLen = to_l(inode->i_blocks <<
						 HFS_SECTOR_SIZE_BITS);
			hfs_extent_out(file, &raw->u.fil.RExtRec);
		}
	} else {
		printk("hfs_write_inode: corrupted hfs_file structure!\n");
	}
}

/*
 * hfs_write_inode()
 *
 * This function is called by the hfs_*_write_inode() functions.
 * Its purpose is to copy the relevent fields from the inode to the
 * catalog B-tree record.
 */
void hfs_write_inode(struct inode * inode, struct hfs_brec *entry)
{
        HFS_LWORD mtime, ctime;

        mtime = hfs_u_to_mtime(inode->i_mtime);
        ctime = hfs_u_to_mtime(inode->i_ctime);
        if (S_ISREG(inode->i_mode)) {
                RAW->u.fil.Flags &= ~1;
                RAW->u.fil.Flags |= (inode->i_mode & S_IWUSR ? 0 : 1);
                RAW->u.fil.MdDat = mtime;
                RAW->u.fil.CrDat = ctime;
        } else if (S_ISDIR(inode->i_mode)) {
                RAW->u.dir.Val = to_w(HFS_I(inode)->u.dir.dirs +
                                      HFS_I(inode)->u.dir.files);
                RAW->u.dir.MdDat = mtime;
                RAW->u.dir.CrDat = ctime;
        }
}

/*
 * hfs_notify_change()
 *
 * Based very closely on fs/msdos/inode.c by Werner Almesberger
 *
 * This is the notify_change() field in the super_operations structure
 * for HFS file systems.  The purpose is to take that changes made to
 * an inode and apply then in a filesystem-dependent manner.  In this
 * case the process has a few of tasks to do:
 *  1) prevent changes to i_size until we have a truncate() function.
 *  2) prevent changes to the i_uig and i_gid fields.
 *  3) map file permisions to the closest allowable permissions
 *  4) This very significant task arises since multiple Linux files
 *  can share the same on-disk inode under HFS (for instance the data
 *  and resource forks of a file).  Because of this a change to 
 *  permissions must be applied to any other in-core inodes which 
 *  correspond to the same HFS file.
 */
int hfs_notify_change(struct inode * inode, struct iattr * attr)
{
	int error;
	ino_t fid;
	umode_t owner_bits;

	error = inode_change_ok(inode, attr);
	if (error) {
		return error;
	}

	/* XXX until we have a valid i_op->truncate() */
	if ((attr->ia_valid & ATTR_SIZE) &&
	    (attr->ia_size != inode->i_size)) {
		return -EPERM;
	}

	if (((attr->ia_valid & ATTR_UID) && 
	     (attr->ia_uid != HFS_SB(inode->i_sb)->s_uid)) ||
	    ((attr->ia_valid & ATTR_GID) && 
	     (attr->ia_gid != HFS_SB(inode->i_sb)->s_gid)) ||
	    ((attr->ia_valid & ATTR_MODE) &&
	     (S_ISDIR(inode->i_mode) ||
	      (attr->ia_mode & ~HFS_VALID_MODE_BITS)))) {
		error = -EPERM;
	}

	if (error) {
		return HFS_SB(inode->i_sb)->s_quiet ? 0 : error;
	}
	
	/* make the mode bits legal for the file's owner: */
	if (!IS_NOEXEC(inode)) {
		owner_bits = S_IXUSR | S_IRUSR | (attr->ia_mode & S_IWUSR);
	} else {
		owner_bits = S_IRUSR | (attr->ia_mode & S_IWUSR);
	}
	owner_bits &= ~HFS_SB(inode->i_sb)->s_umask;

	attr->ia_mode = ((attr->ia_mode & S_IFMT) | owner_bits |
			 (owner_bits >> 3) | (owner_bits >> 6)) &
			~HFS_SB(inode->i_sb)->s_umask;

	inode_setattr(inode, attr);

	/* We wouldn't want to mess with the sizes of the other fork
	   in the case of a file, and directory sizes will be kept
	   current by the create and unlink operations. */
	attr->ia_valid &= ~ATTR_SIZE;

	fid = HFS_IFID(inode->i_ino);
	/* We must change all in-core inodes correspong to this file. */
	if ((fid | HFS_ITYPE_0) != inode->i_ino) {
		propogate_change(inode->i_sb, fid | HFS_ITYPE_0, attr);
	}
	if ((fid | HFS_ITYPE_1) != inode->i_ino) {
		propogate_change(inode->i_sb, fid | HFS_ITYPE_1, attr);
	}
	if ((fid | HFS_ITYPE_2) != inode->i_ino) {
		propogate_change(inode->i_sb, fid | HFS_ITYPE_2, attr);
	}
	if ((fid | HFS_ITYPE_3) != inode->i_ino) {
		propogate_change(inode->i_sb, fid | HFS_ITYPE_3, attr);
	}

	return 0;
}

/*
 * __hfs_iget()
 *
 * Given the super_block for a HFS filesystem, a 'key' for an entry in
 * the catalog B-tree and the 'type' of the desired file return the
 * inode for that file/directory or NULL.  Note that 'type' indicates
 * whether we want the actual file or directory, or the corresponding
 * meta-data (AppleDouble header file or CAP meta-data file).  Note
 * also that this function is never called for the special directories
 * such as ".finderinfo" or ".AppleDouble", because those inodes are
 * generated by the *_dupdir() functions in dir.c.
 *
 * In an ideal world we could call iget() and would not need this
 * function.  However, since there is no way to even know the inode
 * number until we've found the file/directory in the catalog B-tree
 * that simply won't happen.
 *
 * The main idea here is to look in the catalog B-tree to get the
 * vital info about the file or directory (including the file id which
 * becomes the inode number) and then to call iget() and return the
 * inode if it is complete.  If it is not then we use the catalog
 * entry to fill in the missing info, by calling the appropriate
 * 'fillin' function.  Note that these fillin functions are
 * essentially hfs_*_read_inode() functions, but since there is no way
 * to pass the catalog entry through iget() to such a read_inode()
 * function, we have to call them after iget() returns an incomplete
 * inode to us.	 This is pretty much the same problem faced in the NFS
 * code, and pretty much the same solution. The SMB filesystem deals
 * with this in a different way: by using the address of the
 * KMALLOC()'d space which holds the data as the inode number.
 *
 * XXX: Both this function and NFS's corresponding nfs_fhget() would
 * benifit from a way to pass an additional (void *) through iget() to
 * the VFS read_inode() function.
 *
 * XXX: See the comment preceeding hfs_notify_change() for a
 * discussion of an alternate kernel change that would make this a
 * little bit cleaner.
 *
 * XXX: One last alternative is to cache the catalog entry in a place
 * that the VFS read_inode function could find before calling iget()
 * and free removing it from that cache after iget() returns, so that
 * the VFS read_inode function can actually do its work.  All that
 * would do is shift responsibility around.  */
struct inode * __hfs_iget(struct super_block *sb, 
			  const struct hfs_brec_key *key, ino_t type)
{
	struct inode * inode;
	ino_t ino;
	struct hfs_brec *entry;
	HFS_LWORD mtime, ctime;

	if (!sb || hfs_cat_find(sb, key, &entry, 0)) {
		goto bail3;
	}
	if (RAW->cdrType == HFS_CDR_FIL) {
		ino = from_l(RAW->u.fil.FlNum);
	} else if (RAW->cdrType == HFS_CDR_DIR) {
		ino = from_l(RAW->u.dir.DirID);
	} else {
		printk("hfs_iget: entry is neither file nor directory\n");
		goto bail2;
	}
	if ((inode = iget(sb, ino | type)) &&
	    (inode->i_dev == sb->s_dev) && (!inode->i_mode)) {
		memcpy(HFS_I(inode)->key, key, sizeof(*key));
		if (RAW->cdrType==HFS_CDR_FIL) {
			mtime = RAW->u.fil.MdDat;
			ctime = RAW->u.fil.CrDat;
		} else {
			mtime = RAW->u.dir.MdDat;
			ctime = RAW->u.dir.CrDat;
			inode->i_blocks = 0;
		}
		inode->i_rdev = 0;
		inode->i_ctime = hfs_m_to_utime(ctime);
		inode->i_atime = inode->i_mtime = hfs_m_to_utime(mtime);
		inode->i_uid = HFS_SB(sb)->s_uid;
		inode->i_gid = HFS_SB(sb)->s_gid;
		inode->i_blksize = HFS_SECTOR_SIZE;
		if (HFS_SB(sb)->s_ifill(inode, entry, ino, type)) {
			goto bail1;
		}
		inode->i_mode &= ~HFS_SB(sb)->s_umask;
	}
	hfs_brec_relse(&entry);
	return inode;

bail1:
	clear_inode(inode);
bail2:
	hfs_brec_relse(&entry);
bail3:
	return NULL;
}
