/*
 * linux/fs/hfs/inode_nat.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains the inode-related functions specific to Netatalk.
 *
 * Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 */

#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/stat.h>
#include <linux/stddef.h>
#include <linux/malloc.h>
#include <linux/mm.h>

/*================ Variable-like macros ================*/

/* to keep some of the code from getting to verbose */
#define RAW	(entry->data.raw_inode)

/*================ Function-like macros ================*/

/* Given an inode produce the catalog key for the corresponding file */
#define KEY(i) ((struct hfs_brec_key *)(HFS_I(i)->key))

/*================ Global functions ================*/

/* 
 * hfs_nat_ifill()
 *
 * This function serves the same purpose as a read_inode() function
 * does in other filesystems.  It is called by __hfs_iget() to fill in
 * the missing fields of an uninitialized inode under the Netatalk
 * scheme.
 *
 * Note that the hfs_brec is passed as a (void *) so that we can
 * define the s_ifill field in the file hfs_fs_sb.h.
 */
int hfs_nat_ifill(struct inode * inode, void *dummy, ino_t ino, ino_t type)
{
	struct hfs_brec *entry = (struct hfs_brec *)dummy;

	if (type == HFS_NAT_HDR) {
		if (RAW->cdrType == HFS_CDR_FIL) {
			hfs_read_file_inode(&(HFS_I(inode)->u.nat_hdr.rsrc),
					inode, RAW, HFS_FK_RSRC, ino);
			inode->i_size += HFS_NAT_HDR_LEN;
		} else {
			inode->i_size = HFS_NAT_HDR_LEN;
			memset(&(HFS_I(inode)->u.nat_hdr.rsrc), 0,
				sizeof(struct hfs_file));
			inode->i_mode = ((RAW->u.fil.Flags & 1)?0:S_IWUGO) |
				S_IRUGO | S_IFREG;
			inode->i_nlink = 1;
		}
		HFS_I(inode)->u.nat_hdr.meta = NULL;
		inode->i_op = &hfs_nat_hdr_inode_operations;
	} else if (RAW->cdrType == HFS_CDR_FIL) {
		hfs_read_file_inode(&(HFS_I(inode)->u.file), inode,
				RAW, HFS_FK_DATA, ino);
		inode->i_op = &hfs_file_inode_operations;
	} else { /* Directory */
		if (hfs_count_entries(inode)) {
			return 1;
		}
		inode->i_size = 3 + from_w(RAW->u.dir.Val);
		inode->i_nlink = 3 + HFS_I(inode)->u.dir.dirs;
		inode->i_mode = S_IRWXUGO | S_IFDIR;
		inode->i_op = &hfs_nat_dir_inode_operations;
	}
	return 0;
}

/*
 * hfs_nat_put_inode()
 *
 * This is the put_inode() entry in the super_operations for HFS
 * filesystems in the Netatalk scheme.  The purpose is to perform
 * any filesystem-dependent cleanup necessary when the use-count of an
 * inode falls to zero.
 *
 * Once the filesystem is writable this function will need to take
 * care of returning the blocks belonging to deleted files to the free
 * pool.
 */
void hfs_nat_put_inode(struct inode * inode)
{
	if (S_ISREG(inode->i_mode)) {
		switch (HFS_ITYPE(inode->i_ino)) {
		case HFS_NAT_HDR:
			if (HFS_I(inode)->u.nat_hdr.meta) {
				KFREE_S(HFS_I(inode)->u.nat_hdr.meta,
					HFS_NAT_HDR_LEN);
				HFS_I(inode)->u.nat_hdr.meta = NULL;
			}
			hfs_extent_trim(&HFS_I(inode)->u.nat_hdr.rsrc,
					&HFS_I(inode)->u.nat_hdr.rsrc.first);
			break;
			
		case HFS_NAT_DATA:
			hfs_extent_trim(&HFS_I(inode)->u.file,
					&HFS_I(inode)->u.file.first);
			break;
			
		default:
		}
	}
}

/*
 * hfs_nat_write_inode()
 *
 * This is the write_inode() entry in the super_operations for HFS
 * filesystems in the Netatalk scheme.  The purpose is to propogate
 * changes made to the in-core inode to the on-disk inode.
 */
void hfs_nat_write_inode(struct inode * inode)
{
	struct hfs_brec *entry;

	if (!hfs_cat_find(inode->i_sb, KEY(inode), &entry, 1)) {
		hfs_write_inode(inode, entry);
		if (S_ISREG(inode->i_mode))
			if (HFS_ITYPE(inode->i_ino) == HFS_NAT_DATA) {
				hfs_write_file_inode(&HFS_I(inode)->u.file, inode,
						 RAW);
			} else {
				hfs_write_file_inode(&HFS_I(inode)->u.nat_hdr.rsrc,
						 inode, RAW);
			}
		hfs_brec_relse(&entry);
	} else {
		printk("Unable to read inode %ld\n", inode->i_ino);
	}
	inode->i_dirt = 0;
}
