/*
 * linux/fs/hfs/super.c
 *
 * Copyright (C) 1995, 1996  Paul H. Hargrove
 * This file may be distributed under the terms of the GNU Public License.
 *
 * This file contains hfs_read_super() some of the the super_ops and
 * module_init() and module_cleanup().	The remaining super_ops are in
 * inode.c since they deal with inodes.
 *
 * Code in hfs_read_super() to handle the Mac partition table based on
 * a patch contributed by Holger Schemel (aeglos@valinor.owl.de).
 *
 * Based on the minix file system code, (C) 1991, 1992 by Linus Torvalds
 *
 * "XXX" in a comment is a note to myself to consider changing something.
 *
 * In function preconditions the term "valid" applied to a pointer to
 * a structure means that the pointer is non-NULL and the structure it
 * points to has all fields initialized to consistent values.
 *
 * The code in this file initilizes some structures by calling
 * memset(&foo, 0, sizeof(foo)).  This produces the desired behavior
 * only due to the non-ANSI assumption that the machine representation
 * of NULL is all zeros.
 */

#include <linux/config.h>
#define __NO_VERSION__ /* so things work the same with new and old kernels */
#include <linux/module.h>
#include "hfs_fs.h"
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/locks.h>
#include <linux/stddef.h>
#include <linux/stat.h>

/*================ Forward declarations ================*/

static void hfs_put_super(struct super_block *);
static void hfs_statfs(struct super_block *, struct statfs *, int);
static void hfs_write_super(struct super_block *);

/*================ Global variables ================*/

/* A counter used to track brec allocation when debugging */
#if defined(DEBUG_BRECS) || defined(DEBUG_ALL)
int hfs_bcount = 0;
#endif

static struct super_operations hfs_cap_super_operations = { 
	NULL,			/* read_inode */
	hfs_notify_change,	/* notify_change - in inode.c */
	hfs_cap_write_inode,	/* write_inode	 - in inode_cap.c */
	hfs_cap_put_inode,	/* put_inode	 - in inode_cap.c */
	hfs_put_super,		/* put_super */
	hfs_write_super,	/* write_super */
	hfs_statfs,		/* statfs */
	NULL			/* remount_fs */
};

static struct super_operations hfs_dbl_super_operations = { 
	NULL,			/* read_inode */
	hfs_notify_change,	/* notify_change - in inode.c */
	hfs_dbl_write_inode,	/* write_inode	 - in inode_dbl.c */
	hfs_dbl_put_inode,	/* put_inode	 - in inode_dbl.c */
	hfs_put_super,		/* put_super */
	hfs_write_super,	/* write_super */
	hfs_statfs,		/* statfs */
	NULL			/* remount_fs */
};

static struct super_operations hfs_nat_super_operations = { 
	NULL,			/* read_inode */
	hfs_notify_change,	/* notify_change - in inode.c */
	hfs_nat_write_inode,	/* write_inode	 - in inode_nat.c */
	hfs_nat_put_inode,	/* put_inode	 - in inode_nat.c */
	hfs_put_super,		/* put_super */
	hfs_write_super,	/* write_super */
	hfs_statfs,		/* statfs */
	NULL			/* remount_fs */
};

/*================ File-local data types ================*/

/*
 * The Macintosh Driver Descriptor Block
 *
 * On partitioned Macintosh media this is block 0.
 * We really only need the "magic number" to check for partitioned media.
 */
struct hfs_drvr_desc {
	HFS_WORD	ddSig;	/* The signature bytes to recognize
				   partitioned media */
	/* a bunch more stuff we don't need */
} DONT_ALIGN;

/* 
 * The Mac partiton map
 *
 * For each partition on the media there is a physical block (512-byte
 * block) containing one of these structures.  These blocks are
 * contiguous starting at block 1.
 */
struct hfs_partition_map {
	HFS_WORD	pmSig;		/* Signature bytes to verify
					   that this is a partition
					   map block */
	HFS_WORD	reSigPad;	/* padding */
	HFS_LWORD	pmMapBlkCnt;	/* (At least in block 1) this
					   is the number of partition
					   map blocks */
	HFS_LWORD	pmPyPartStart;	/* The physical block number
					   of the first block in this
					   partition */
	HFS_LWORD	pmPartBlkCnt;	/* The number of physical
					   blocks in this partition */
	HFS_BYTE	pmPartName[32]; /* (null terminated?) string
					   giving the name of this
					   partition */
	HFS_BYTE	pmPartType[32]; /* (null terminated?) string
					   giving the type of this
					   partition */
	/* a bunch more stuff we don't need */
} DONT_ALIGN;

/* 
 * The HFS Master Directory Block (MDB).
 *
 * Also known as the Volume Information Block (VIB), this structure is
 * the HFS equivalent of a superblock.
 *
 * Reference: _Inside Macintosh: Files_ pages 2-59 through 2-62
 */
struct hfs_super {
	HFS_WORD		drSigWord;	/* Signature word to
						   indicate fs type */
	HFS_LWORD		drCrDate;	/* fs creation date/time */
	HFS_LWORD		drLsMod;	/* fs modification date/time */
	HFS_WORD		drAtrb;		/* fs attributes */
	HFS_WORD		drNmFls;	/* number of files in
						   root directory */
	HFS_WORD		drVBMSt;	/* location (in
						   512-byte blocks) of
						   the volume bitmap */
	HFS_WORD		drAllocPtr;	/* location (in
						   allocation blocks)
						   to begin next
						   allocation search */
	HFS_WORD		drNmAlBlks;	/* number of
						   allocation blocks
						   in the fs */
	HFS_LWORD		drAlBlkSiz;	/* size (in bytes) of
						   an allocation block */
	HFS_LWORD		drClpSiz;	/* clumpsize, the size
						   (in bytes) to
						   allocate when
						   extending a file */
	HFS_WORD		drAlBlSt;	/* location (in
						   512-byte blocks) of
						   the first
						   allocation block */
	HFS_LWORD		drNxtCNID;	/* File ID to assign
						   to the next file or
						   directory created */
	HFS_WORD		drFreeBks;	/* number of unused
						   allocation blocks */
	HFS_BYTE		drVNlen;	/* number of bytes in
						   the volume label */
	HFS_BYTE		drVN[27];	/* the volume label */
	HFS_LWORD		drVolBkUp;	/* fs backup date/time */
	HFS_WORD		drVSeqNum;	/* backup sequence number */
	HFS_LWORD		drWrCnt;	/* fs write count
						   XXX: not yet updated */
	HFS_LWORD		drXTClpSiz;	/* clumpsize for the
						   extents B-tree */
	HFS_LWORD		drCTClpSiz;	/* clumpsize for the
						   catalog B-tree */
	HFS_WORD		drNmRtDirs;	/* number of
						   directories in the
						   root directory */
	HFS_LWORD		drFilCnt;	/* number of files in
						   the fs */
	HFS_LWORD		drDirCnt;	/* number of
						   directories in the fs */
	HFS_BYTE		drFndrInfo[32]; /* data used by the Finder */
	HFS_WORD		drVCSize;	/* MacOS caching parameter */
	HFS_WORD		drVCBMSize;	/* MacOS caching parameter */
	HFS_WORD		drCtlCSize;	/* MacOS caching parameter */
	HFS_LWORD		drXTFlSize;	/* size (in bytes) of
						   the extents B-tree */
	struct hfs_raw_extent	drXTExtRec;	/* first extent record
						   for the extents B-tree */
	HFS_LWORD		drCTFlSize;	/* size (in bytes) of
						   the catalog B-tree */
	struct hfs_raw_extent	drCTExtRec;	/* first extent record
						   for the catalog B-tree */
} DONT_ALIGN;

/*================ Global-local variables ================*/

#ifdef MODULE
char kernel_version[]=UTS_RELEASE;
#endif

/*================ File-local variables ================*/

#ifdef MODULE
static struct file_system_type hfs_fs = {hfs_read_super,"hfs",1,NULL};
#endif

/*================ File-local functions ================*/

/*
 * hfs_commit_super()
 *
 * Description:
 *   This updates the MDB on disk (look also at hfs_write_super()).
 *   It does not check, if the superblock has been modified, or
 *   if the filesystem has been mounted read-only. It is mainly
 *   called by hfs_write_super.
 * Input Variable(s):
 *   struct super_block *sb: Pointer to the hfs superblock
 * Output Variable(s):
 *   NONE
 * Returns:
 *   void
 * Preconditions:
 *   'sb' points to a "valid" (struct super_block).
 * Postconditions:
 *   The HFS MDB in memory and on disk will be updated, by copying the
 *   possibly modified fields from the super block (in native byte
 *   order) to the disk block buffer.
 */
static void hfs_commit_super(struct super_block *sb)
{
	struct hfs_sb_info *hsb = HFS_SB(sb);

	/* These parameters may have been modified, so write them back */
	hsb->s_hs->drFreeBks  = to_w(hsb->s_free_ablocks);
	hsb->s_hs->drNxtCNID  = to_l(hsb->s_next_id);
	hsb->s_hs->drWrCnt    = to_l(hsb->s_write_count);
	hsb->s_hs->drNmFls    = to_w(hsb->s_root_files);
	hsb->s_hs->drNmRtDirs = to_w(hsb->s_root_dirs);
	hsb->s_hs->drFilCnt   = to_l(hsb->s_file_count);
	hsb->s_hs->drDirCnt   = to_l(hsb->s_dir_count);

	/* update volume mtime */
	hsb->s_hs->drLsMod = hfs_u_to_mtime(CURRENT_TIME);

	/* write MDB to disk */
	mark_buffer_dirty(hsb->s_sbbh, 1);

	sb->s_dirt = 0;
}

/*
 * hfs_write_super()
 *
 * Description:
 *   This function is called by the VFS only. When the filesystem
 *   is mounted r/w it updates the MDB on disk.
 * Input Variable(s):
 *   struct super_block *sb: Pointer to the hfs superblock
 * Output Variable(s):
 *   NONE
 * Returns:
 *   void
 * Preconditions:
 *   'sb' points to a "valid" (struct super_block).
 * Postconditions:
 *   The MDB is marked 'unsuccessfully unmouted' by clearing bit 8 of drAtrb
 *   (hfs_put_super() must set this flag!). Some MDB fields are updated
 *   and the MDB buffer is written to disk by calling hfs_commit_super().
 */
static void hfs_write_super(struct super_block *sb)
{
	struct hfs_sb_info *hsb = HFS_SB(sb);
	/* is this a valid hfs superblock? */
	if (!sb || sb->s_magic != HFS_SUPER_MAGIC) {
		return;
	}

	if (!(sb->s_flags & MS_RDONLY)) {
		if (from_w(hsb->s_hs->drAtrb) & HFS_SB_ATTRIB_CLEAN) {
			hsb->s_hs->drAtrb &= to_w(~HFS_SB_ATTRIB_CLEAN);
		}
		hfs_commit_super(sb);
	}
	sb->s_dirt = 0;
}

/*
 * do_put_super()
 *
 * Called by hfs_put_super() to do the real work of releasing the
 * resources associated with the superblock.  This function is also
 * called by hfs_read_super() to cleanup a failed attempt to read the
 * superblock.	Always called with the super block locked.
 */
static void do_put_super(struct super_block *sb)
{
	int lcv;
	dev_t dev = sb->s_dev;
	struct hfs_sb_info *hsb = HFS_SB(sb);

	/* If the filesystem is mounted r/w, set the
	 * 'successfully unmounted' flag of the drAtrb 
	 * field of the MDB and write the MDB to disk
	 */
	if (!(sb->s_flags & MS_RDONLY)) {
		hsb->s_hs->drAtrb = to_w(from_w(hsb->s_hs->drAtrb) |
					 HFS_SB_ATTRIB_CLEAN);
		hfs_commit_super(sb);
	}

	sb->s_dev = 0;
	if (hsb->s_ext_tree) {
		hfs_btree_free(hsb->s_ext_tree);
	}
	if (hsb->s_cat_tree) {
		hfs_btree_free(hsb->s_cat_tree);
	}
	if (hsb->s_sbbh) {
		brelse(hsb->s_sbbh);
	}
	for (lcv = 0; hsb->s_bitmap[lcv] && (lcv < HFS_BM_MAXBLOCKS); ++lcv) {
		brelse(hsb->s_bitmap[lcv]);
	}

	set_blocksize(dev,BLOCK_SIZE);
	MOD_DEC_USE_COUNT;
	return;
}

/*
 * hfs_put_super()
 *
 * This is the put_super() entry in the super_operations structure for
 * HFS filesystems.  The purpose is to release the resources
 * associated with the superblock sb.  This function simply locks sb,
 * calls do_put_super() and unlocks sb.
 */
static void hfs_put_super(struct super_block *sb)
{
	lock_super(sb);
	do_put_super(sb);
	unlock_super(sb);
	return;
}

/*
 * hfs_statfs()
 *
 * This is the statfs() entry in the super_operations structure for
 * HFS filesystems.  The purpose is to return various data about the
 * filesystem.
 */
static void hfs_statfs(struct super_block *sb, struct statfs *buf, int len)
{
	struct statfs tmp;

	tmp.f_type = HFS_SUPER_MAGIC;
	tmp.f_bsize = HFS_SECTOR_SIZE;
	tmp.f_blocks = HFS_SB(sb)->s_alloc_blksz * HFS_SB(sb)->s_fs_ablocks;
	tmp.f_bfree = HFS_SB(sb)->s_alloc_blksz * HFS_SB(sb)->s_free_ablocks;
	tmp.f_bavail = tmp.f_bfree;
	tmp.f_files = -1; /* According to the statfs manual page, -1 is the  */
	tmp.f_ffree = -1; /* correct value when the meaning is undefined.    */
	tmp.f_namelen = HFS_NAMELEN;

	memcpy_tofs(buf, &tmp, len);
}

/*
 * parse_options()
 * 
 * adapted from linux/fs/msdos/inode.c written 1992,93 by Werner Almesberger
 * This function is called by hfs_read_super() to parse the mount options.
 */
static int parse_options(char *options, uid_t *uid, gid_t *gid, int *umask,
			 char *names, char *fork, int *part, int *quiet)
{
	char *this_char,*value;

	*uid = current->uid;
	*gid = current->gid;
	*umask = current->fs->umask;
	*names = 'c';
	*fork = 'c';
	*part = 0;
	*quiet = 0;
	if (!options) {
		return 1;
	}
	for (this_char = strtok(options,","); this_char;
	     this_char = strtok(NULL,",")) {
		if ((value = strchr(this_char,'=')) != NULL) {
			*value++ = 0;
		}
		if (!strcmp(this_char,"uid")) {
			if (!value || !*value) {
				return 0;
			}
			*uid = simple_strtoul(value,&value,0);
			if (*value) {
				return 0;
			}
		} else if (!strcmp(this_char,"gid")) {
			if (!value || !*value) {
				return 0;
			}
			*gid = simple_strtoul(value,&value,0);
			if (*value) {
				return 0;
			}
		} else if (!strcmp(this_char,"umask")) {
			if (!value || !*value) {
				return 0;
			}
			*umask = simple_strtoul(value,&value,8);
			if (*value) {
				return 0;
			}
		} else if (!strcmp(this_char,"part")) {
			if (!value || !*value) {
				return 0;
			}
			*part = simple_strtoul(value,&value,8);
			if (*value) {
				return 0;
			}
		} else if (!strcmp(this_char,"names") && value) {
			if (value[0] && !value[1] && strchr("ac78tl",*value)) {
				*names = *value;
			} else if (!strcmp(value,"alpha")) {
				*names = 'a';
			} else if (!strcmp(value,"cap")) {
				*names = 'c';
			} else if (!strcmp(value,"7bit")) {
				*names = '7';
			} else if (!strcmp(value,"8bit")) {
				*names = '8';
			} else if (!strcmp(value,"trivial")) {
				*names = 't';
			} else if (!strcmp(value,"latin")) {
				*names = 'l';
			} else {
				return 0;
			}
		} else if (!strcmp(this_char,"fork") && value) {
			if (value[0] && !value[1] && strchr("cdsn",*value)) {
				*fork = *value;
			} else if (!strcmp(value,"cap")) {
				*fork = 'c';
			} else if (!strcmp(value,"double")) {
				*fork = 'd';
			} else if (!strcmp(value,"single")) {
				*fork = 's';
			} else if (!strcmp(value,"netatalk")) {
				*fork = 'n';
			} else {
				return 0;
			}
                } else if (!strcmp(this_char,"quiet")) {
                        if (value) {
				return 0;
			}
                        *quiet = 1;
		} else {
			return 0;
		}
	}
	return 1;
}

/*================ Global functions ================*/

/*
 * hfs_read_super()
 *
 * This is the function that is responsible for mounting an HFS
 * filesystem.	It performs all the tasks necessary to get enough data
 * from the disk to read the root inode.  This includes parsing the
 * mount options, dealing with Macintosh partitons, reading the
 * superblock and the allocation bitmap blocks, calling
 * hfs_btree_init() to get the necessary data about the extents and
 * catalog B-trees and, finally, reading the root inode into memory.
 */
struct super_block *hfs_read_super(struct super_block *s, void *data,
				   int silent)
{
	struct hfs_super *hs;
	int lcv, part, quiet, umask, dev=s->s_dev;
	uid_t uid;
	gid_t gid;
	unsigned int bs, block;
	unsigned int partition_start = 0;
	char names, fork;
	struct buffer_head *bh, **bmbh;
	struct hfs_partition_map *pm;


	if (!parse_options((char *)data, &uid, &gid, &umask,
			   &names, &fork, &part, &quiet)) {
		printk("hfs_read_super: unable to parse mount options.\n");
		goto bail2;
	}

	/* in case someone tries to unload the module while we wait on I/O */
	MOD_INC_USE_COUNT;

	lock_super(s);

	memset(HFS_SB(s), 0, sizeof(struct hfs_sb_info));
	HFS_SB(s)->s_uid = uid;
	HFS_SB(s)->s_gid = gid;
	HFS_SB(s)->s_umask = umask;
	HFS_SB(s)->s_quiet = quiet;
	if (names=='c') {
		HFS_SB(s)->s_nameout = hfs_cap2mac;
		HFS_SB(s)->s_namein = hfs_mac2cap;
	} else if (names=='t') {
		HFS_SB(s)->s_nameout = hfs_triv2mac;
		HFS_SB(s)->s_namein = hfs_mac2triv;
	} else if (names=='7') {
		HFS_SB(s)->s_namein = hfs_mac2seven;
		HFS_SB(s)->s_nameout = hfs_apl2mac;
	} else if (names=='8') {
		HFS_SB(s)->s_namein = hfs_mac2eight;
		HFS_SB(s)->s_nameout = hfs_apl2mac;
	} else if (names=='l') {
		HFS_SB(s)->s_nameout = hfs_latin2mac;
		HFS_SB(s)->s_namein = hfs_mac2latin;
	} else {/* names=='a' */
		HFS_SB(s)->s_namein = hfs_mac2alpha;
		HFS_SB(s)->s_nameout = hfs_apl2mac;
	}

	set_blocksize(dev,HFS_SECTOR_SIZE);

	if (!(bh = bread(dev, HFS_PMAP_BLK, HFS_SECTOR_SIZE))) {
		printk("hfs_read_super: unable to read partition map.\n");
		goto bail1;
	}

	pm = (struct hfs_partition_map *)(bh->b_data);
	if (from_w(pm->pmSig) == HFS_OLD_PMAP_MAGIC) {
		printk("hfs_read_super: This disk has an unsupported "
		       "old-style partition map.\n");
		printk("hfs_read_super: If you'd like to see support "
		       "for this please contact the author.");
		goto bail0;
	} else if (from_w(pm->pmSig) == HFS_NEW_PMAP_MAGIC) {
		int i, pmap_entries;

		pmap_entries = from_l(pm->pmMapBlkCnt);
		for (i=0; i<pmap_entries && !partition_start; ++i) {
			/* read the entry if we don't have it yet */
			if (i != 0) {
				brelse(bh);
				if (!(bh = bread(dev, HFS_PMAP_BLK + i,
						 HFS_SECTOR_SIZE))) {
					printk("hfs_read_super: unable to "
					       "read partition map.\n");
					goto bail1;
				}
				pm = (struct hfs_partition_map *)(bh->b_data);
				if (from_w(pm->pmSig) != HFS_NEW_PMAP_MAGIC) {
					printk("hfs_read_super: invalid "
					       "entry in partition map\n");
					goto bail0;
				}
			}
			/* look for an HFS partition */
			if (!memcmp(pm->pmPartType, "Apple_HFS", 9) &&
			    ((--part) <= 0)) {
				partition_start = from_l(pm->pmPyPartStart);
			}
		}
		if (part > 0) {
			printk("hfs_read_super: unable to locate the "
			       "requested parition\n");
			goto bail0;
		}
		if (!partition_start) {
			printk("hfs_read_super: no HFS parition was found\n");
			goto bail0;
		}
	}
	brelse(bh);
				
	/* if we get here we are on unpartitioned media or
	   we found an HFS partition */
	if (!(HFS_SB(s)->s_sbbh =
	      bread(dev, HFS_MDB_BLK + partition_start, HFS_SECTOR_SIZE))) {
		printk("hfs_read_super: unable to read superblock.\n");
		goto bail1;
	}

	HFS_SB(s)->s_hs = hs = (struct hfs_super *)HFS_SB(s)->s_sbbh->b_data;

	if ((s->s_magic = from_w(hs->drSigWord)) != HFS_SUPER_MAGIC) {
		if (!silent) {
			printk("VFS: Can't find a HFS filesystem on dev "
			       "0x%04x.\n", dev);
		}
		goto bail1;
	}

	bs = from_l(hs->drAlBlkSiz);
	if (!bs || bs > HFS_USHRT_MAX || (bs & (HFS_SECTOR_SIZE-1))) {
		printk("hfs_read_super: bad allocation block size\n");
		goto bail1;
	}
	s->s_blocksize_bits = HFS_SECTOR_SIZE_BITS;
	HFS_SB(s)->s_alloc_blksz = bs >> s->s_blocksize_bits;

	/* These parameters are read from the MDB , and never written */
	HFS_SB(s)->s_fs_ablocks = from_w(hs->drNmAlBlks);
	HFS_SB(s)->s_fs_start = from_w(hs->drAlBlSt) + partition_start;
	HFS_SB(s)->s_clump_size = from_l(hs->drClpSiz);

	/* These parameters are read from and written to the MDB */
	HFS_SB(s)->s_free_ablocks = from_w(hs->drFreeBks);
	HFS_SB(s)->s_next_id = from_l(hs->drNxtCNID);
	HFS_SB(s)->s_write_count = from_l(hs->drWrCnt);
	HFS_SB(s)->s_root_files = from_w(hs->drNmFls);
	HFS_SB(s)->s_root_dirs = from_w(hs->drNmRtDirs);
	HFS_SB(s)->s_file_count = from_l(hs->drFilCnt);
	HFS_SB(s)->s_dir_count = from_l(hs->drDirCnt);
	
	/* read in the bitmap */
	block = from_w(hs->drVBMSt) + partition_start;
	bmbh = HFS_SB(s)->s_bitmap;
	lcv = (HFS_SB(s)->s_fs_ablocks + 4095) / 4096;
	for ( ; lcv; --lcv, ++bmbh, ++block) {
		if (!(*bmbh = bread(dev, block, HFS_SECTOR_SIZE))) {
			printk("hfs_read_super: unable to read bitmap\n");
			goto bail1;
		}
	}

	if (HFS_ITYPE(HFS_SB(s)->s_next_id) != 0) {
		printk("hfs_read_super: too many files.\n");
		goto bail1;
	}

	if (!(HFS_SB(s)->s_ext_tree =
	      hfs_btree_init(s, HFS_EXT_FID, &hs->drXTExtRec,
			     &hs->drXTFlSize, hs->drXTClpSiz)) ||
	    !(HFS_SB(s)->s_cat_tree =
	      hfs_btree_init(s, HFS_CAT_FID, &hs->drCTExtRec,
			     &hs->drCTFlSize, hs->drCTClpSiz))) {
		printk("hfs_read_super: unable to initialize "
		       "data structures\n");
		goto bail1;
	}

	s->s_blocksize = HFS_SECTOR_SIZE;

	switch (fork) {
	case 's':
		printk("hfs_read_super: AppleSingle not yet implemented.\n");
		goto bail1;
		/* break; */
	
	case 'd':
		HFS_SB(s)->s_ifill = hfs_dbl_ifill;
		s->s_op = &hfs_dbl_super_operations;
		break;
	
	case 'c':
		HFS_SB(s)->s_ifill = hfs_cap_ifill;
		s->s_op = &hfs_cap_super_operations;
		break;

	case 'n':
		HFS_SB(s)->s_ifill = hfs_nat_ifill;
		s->s_op = &hfs_nat_super_operations;
		break;
	}
	s->s_mounted = hfs_iget(s, HFS_POR_FID, hs->drVNlen,
				hs->drVN, HFS_ITYPE_NORM);
	if (!s->s_mounted) {
		printk("hfs_read_super: get root inode failed.\n");
		goto bail1;
	}
	unlock_super(s);
	return s;

bail0:
	brelse(bh);
bail1:
	do_put_super(s);
	unlock_super(s);
bail2:
	s->s_dev = 0;
	return NULL;	
}

#ifdef MODULE
int init_module(void) {
#ifdef PARANOID
	if ((sizeof(struct hfs_sb_info)>sizeof(struct ext2_sb_info)) ||
	    (sizeof(struct hfs_inode_info)>sizeof(struct ext2_inode_info))) {
		printk("Filesystem structures too big!!\n");
		return -ENOMEM; /* well sort of */
	}
#endif
/*	register_symtab(NULL); */
	return register_filesystem(&hfs_fs);
}

void cleanup_module(void) {
#if defined(DEBUF_BRECS) || defined(DEBUG_ALL)
	printk("hfs_bcount = %d\n", hfs_bcount);
#endif
	unregister_filesystem(&hfs_fs);
}
#endif

#if defined(DEBUG_ALL) || defined(DEBUG_MEM)
long int hfs_alloc = 0;
#endif
